/**************************************************************************
 * Copyright (c) 2004-2006, Andrew Klyachkin <andrewk@region73.ru>        *
 * All rights reserved.                                                   *
 *                                                                        *
 * Redistribution and use in source and binary forms, with or without     *
 * modification, are permitted provided that the following conditions     *
 * are met:                                                               *
 *                                                                        *
 * 1. Redistributions of source code must retain the above copyright      *
 *    notice, this list of conditions and the following disclaimer.       *
 * 2. Redistributions in binary form must reproduce the above copyright   *
 *    notice, this list of conditions and the following disclaimer in the *
 *    documentation and/or other materials provided with the distribution.*
 * 3. Neither the name of the Andrew Klyachkin nor the names of his       *
 *    contributors may be used to endorse or promote products derived     *
 *    from this software without specific prior written permission.       *
 *                                                                        *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS    *
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT      *
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR  *
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT   *
 * AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, *
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       *
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  *
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  *
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    *
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  *
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   *
 **************************************************************************/

#include <sys/types.h>
#include <sys/socket.h>
#include <stdlib.h>
#include <stdio.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <string.h>
#include <ctype.h>

#include "libnnfc.h"

void insertcommand(char *command, struct command **commands)
{
    struct command  *c = *commands;

    if (!c) {
        (*commands) = calloc(1, sizeof(struct command));
        (*commands)->option = calloc(1, strlen(command));
        strncpy((*commands)->option, command, strlen(command));
    } else {
        while (c->next) {
            c = c->next;
        }
        c->next = calloc(1, sizeof(struct command));
        c->next->option = calloc(1, strlen(command)+1);
        strncpy(c->next->option, command, strlen(command));
    }
}

void addvalue(char *command, char *value, struct command **commands)
{
    struct command  *c = *commands;
    struct value    *v;

    if (!c) return;

    while (c) {
        if (!strncmp(c->option, command, strlen(command))) break;
        c = c->next;
    }

    if (!c) return;

    v = c->values;
    if (!v) {
        c->values = calloc(1, sizeof(struct value));
        c->values->value = calloc(1, strlen(value)+1);
        strncpy(c->values->value, value, strlen(value));
    } else {
        while (v->next) {
            v = v->next;
        }
        v->next = calloc(1, sizeof(struct value));
        v->next->value = calloc(1, strlen(value)+1);
        strncpy(v->next->value, value, strlen(value));
    }
}

void printconfig(struct command *commands)
{
    struct command  *c = commands;
    struct value    *v;

    while (c) {
        printf("%s:", c->option);
        v = c->values;
        while (v) {
            printf("{%s}", v->value);
            v = v->next;
        }
        printf("\n");
        c = c->next;
    }
}

char *getvalue(char *command, struct command *commands)
{
    struct command  *c = commands;

    while (c) {
        if (!strncmp(c->option, command, strlen(command))) {
            return c->values->value;
        }
        c = c->next;
    }

    return NULL;
}

struct value *getallvalues(char *command, struct command *commands)
{
    struct command	*c = commands;

    while (c) {
        if (!strncmp(c->option, command, strlen(command))) {
            return c->values;
        }
        c = c->next;
    }

    return NULL;
}

int readcfg(char *fname, struct command **commands)
{
    FILE    *fp;
    int     c, i, line = 1;
    char    command[256];
    char    value[256];

    fp = fopen(fname, "r");
    if (!fp) {
        return 1;
    }

    while ((c = fgetc(fp)) != EOF) {
        if (c == '\n') {
            line++;
            continue;
        }
        if (isspace(c))
            continue;
        if (c == '#' || c == ';') {
            while ((c = fgetc(fp)) != '\n');
            line++;
            continue;
        }
        memset(&command, 0, 256);
        i = 0;
        while (c != '=' && !isspace(c)) {
            if (i< 256)
                command[i] = c;
            c = fgetc(fp);
            i++;
        }
        insertcommand(command, &(*commands));

        while (isspace(c)) {
            c = fgetc(fp);
        }
        if (c != '=') {
            fclose(fp);
            return 1;
        }
        c = fgetc(fp);
        if (isspace(c)) 
            while (isspace(c)) {
                c = fgetc(fp);
            }
        memset(&value, 0, 256);
        i = 0;
        if (c == '{') {
            while (c != '}') {
                do {
                    c = fgetc(fp);
                    if (c == '\n') line++;
                } while (c != '"' && c != '\'' && c != '}');

                if (c == '}') break;

                c = fgetc(fp);
                while (c != '"' && c != '\'') {
                    if (c == '\n') line++;
                    if (i < 256)
                        value[i] = c;
                    i++;
                    c = fgetc(fp);
                }

                addvalue(command, value, &(*commands));
                memset(&value, 0, 256);
                i = 0;
            }
        } else
        if (c == '"' || c == '\'') {
            /* value in quotes */
            c = fgetc(fp);
            while (c != '"' && c != '\'') {
                if (c == '\n') line++;
                if (i < 256)
                    value[i] = c;
                i++;
                c = fgetc(fp);
            }
        } else {
            while (!isspace(c) && c != ';' && c != '#') {
                if (i < 256)
                    value[i] = c;
                i++;
                c = fgetc(fp);
            }
        }
        if (value[0] != 0) {
            addvalue(command, value, &(*commands));
        }
        if (c == '\n') line++;
    }
    fclose(fp);
    return 0;
}

int freecmds(struct command *commands)
{
    struct command  *c;
    struct value    *v;

    if (!commands) return -1;

    while (commands) {
        c = commands->next;
        while (commands->values) {
            v = commands->values->next;
            free(commands->values);
            commands->values = v;
        }
        free(commands);
        commands = c;
    }

    return 0;
}
