/**************************************************************************
 * Copyright (c) 2004-2006, Andrew Klyachkin <andrewk@region73.ru>        *
 * All rights reserved.                                                   *
 *                                                                        *
 * Redistribution and use in source and binary forms, with or without     *
 * modification, are permitted provided that the following conditions     *
 * are met:                                                               *
 *                                                                        *
 * 1. Redistributions of source code must retain the above copyright      *
 *    notice, this list of conditions and the following disclaimer.       *
 * 2. Redistributions in binary form must reproduce the above copyright   *
 *    notice, this list of conditions and the following disclaimer in the *
 *    documentation and/or other materials provided with the distribution.*
 * 3. Neither the name of the Andrew Klyachkin nor the names of his       *
 *    contributors may be used to endorse or promote products derived     *
 *    from this software without specific prior written permission.       *
 *                                                                        *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS    *
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT      *
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR  *
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT   *
 * AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, *
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       *
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  *
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  *
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    *
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  *
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   *
 **************************************************************************/

#ifndef LIBNNFC_H
#define LIBNNFC_H

#include <config.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/param.h>
#include <netinet/in.h>
#include <stdio.h>
#include <sys/time.h>
#include <unistd.h>

#include <pthread.h>

#include <libconfig.h>

#define MAXMSG  4096

#define LOADFUNC(func, fnname)  *(void **)(&func) = dlsym(m->handle, fnname); \
        if ((error = (char *)dlerror()) != NULL) { \
            syslog(LOG_ERR, "module_open: %s\n", error); \
            if (m->handle != NULL) \
                dlclose(m->handle); \
            m->handle = NULL; \
            return -1; \
        }

#define CFG_GETI(var, name, cmds)   { \
        if (getvalue(name, cmds)) { \
            var = atoi(getvalue(name, cmds)); \
        } \
    }

#define CFG_GETS(var, name, cmds)   { \
        if (getvalue(name, cmds)) { \
            n = strlen(getvalue(name, cmds)); \
            var = calloc(n+1, sizeof(char)); \
            strncpy(var, getvalue(name, cmds), n); \
        } \
    }

/* Constants for sampling mode field in header */
#define SMPL_NONE   0       /* No sampling mode is configured */
#define SMPL_PACKET 1       /* One of every x packet is placed in a flow */
#define	SMPL_RESV1  2       /* Reserved */
#define	SMPL_RESV2  3       /* Reserved */

/* netflow v5 structure */
struct flowhdr5_t {
    short   version;        /* netflow version */
    short   count;          /* count of flows in packet */
    int     uptime;         /* msecs since router startup */
    int     secs;           /* secs since Epoch */
    int     nsecs;          /* nanosecs since Epoch */
    int     flowseq;        /* seq counter of total flows sent */
    char    engine;         /* type of flow engine */
    char    eng_id;         /* id of flow engine */
    short   sampling;       /* sampling mode and interval */
};

struct flow5_t {
    unsigned char   srcaddr[4]; /* source ip address */
    unsigned char   dstaddr[4]; /* destination ip address */
    unsigned char   nexthop[4]; /* nexthop address */
    short           input_if;   /* SNMP index of input interface */
    short           output_if;  /* SNMP index of output interface */
    int             dpkts;      /* packets in flow */
    int             doctets;    /* octets in flow */
    int             fuptime;    /* uptime at start of flow */
    int             luptime;    /* uptime at end of flow */
    short           srcport;    /* source port */
    short           dstport;    /* destination port */
    char            resv1;      /* reserved */
    char            flags;      /* TCP flags */
    char            prot;       /* IP protocol */
    char            tos;        /* IP ToS */
    short           src_as;     /* AS of source address */
    short           dst_as;     /* AS of destination address */
    char            srcmask;    /* Source address mask bits */
    char            dstmask;    /* Destination address mask bits */
    short           resv2;      /* Reserved */
};

/* netflow v7 structure */
struct flowhdr7_t {
    short   version;        /* netflow version */
    short   count;          /* flows count in packet */
    int     uptime;         /* msecs since router startup */
    int     secs;           /* secs since Epoch */
    int     nsecs;          /* nanosecs since Epoch */
    int     flowseq;        /* seq counter of total flows sent */
    int     resv;           /* reserved */
};

struct flow7_t {
    unsigned char   srcaddr[4]; /* source ip address */
    unsigned char   dstaddr[4]; /* destination ip address */
    unsigned char   nexthop[4]; /* nexthop address */
    short           input_if;   /* SNMP index of input interface */
    short           output_if;  /* SNMP index of output interface */
    int             dpkts;      /* packets in flow */
    int             doctets;    /* octets in flow */
    int             fuptime;    /* uptime at start of flow */
    int             luptime;    /* uptime at end of flow */
    short           srcport;    /* source port */
    short           dstport;    /* destination port */
    char            mode;       /* shortcut mode */
    char            flags;      /* TCP flags */
    char            prot;       /* IP protocol */
    char            tos;        /* IP ToS */
    short           src_as;     /* AS of source address */
    short           dst_as;     /* AS of destination address */
    char            srcmask;    /* Source address mask bits */
    char            dstmask;    /* Destination address mask bits */
    short           resv2;      /* Reserved */
    unsigned char   router[4];  /* router which is shortcut by switch */
};

/* netflow v8 format */

struct flowhdr8_t {
    short   version;        /* netflow version */
    short   count;          /* flows count in packet */
    int     uptime;         /* msecs since router startup */
    int     secs;           /* secs since Epoch */
    int     nsecs;          /* nanosecs since Epoch */
    int     flowseq;        /* seq counter of total flows sent */
    char    engine;         /* type of flow engine */
    char    eng_id;         /* id of flow engine */
    char    aggregation;    /* type of agg.scheme being used */
    char    aggver;         /* agg. subformat version number */
    short   sampling;       /* sampling mode and interval */
    short   resv;           /* reserved */
};

struct flow8_asmatrix_t {
    int     flows;          /* number of flows */
    int     dpkts;          /* packets in flow */
    int     doctets;        /* octets in flow */
    int     fuptime;        /* Uptime at start of flow */
    int     luptime;        /* Uptime at end of flow */
    short   src_as;         /* AS of source address */
    short   dst_as;         /* AS of destination address */
    short   input_if;       /* SNMP index of input interface */
    short   output_if;      /* SNMP index of output interface */
};

struct flow8_protmatrix_t {
    int     flows;          /* number of flows */
    int     dpkts;          /* packets in flow */
    int     doctets;        /* octets in flow */
    int     fuptime;        /* Uptime at start of flow */
    int     luptime;        /* Uptime at end of flow */
    char    prot;           /* IP protocol */
    char    resv1;          /* reserved */
    short   resv2;          /* reserved */
    short   srcport;        /* source port */
    short   dstport;        /* destination port */
};

struct flow8_srcmatrix_t {
    int             flows;          /* number of flows */
    int             dpkts;          /* packets in flow */
    int             doctets;        /* octets in flow */
    int             fuptime;        /* Uptime at start of flow */
    int             luptime;        /* Uptime at end of flow */
    unsigned char   srcprefix[4];   /* source prefix */
    char            srcmask;        /* Source address mask bits */
    char            resv1;          /* reserved */
    short           src_as;         /* AS of source address */
    short           input_if;       /* SNMP index of input interface */
    short           resv2;          /* reserved */
};

struct flow8_dstmatrix_t {
    int             flows;          /* number of flows */
    int             dpkts;          /* packets in flow */
    int             doctets;        /* octets in flow */
    int             fuptime;        /* Uptime at start of flow */
    int             luptime;        /* Uptime at end of flow */
    unsigned char   dstprefix[4];   /* destination prefix */
    char            dstmask;        /* Destination address mask bits */
    char            resv1;          /* reserved */
    short           dst_as;         /* AS of destination address */
    short           output_if;      /* SNMP index of output interface */
    short           resv2;          /* reserved */
};

struct flow8_prefixmatrix_t {
    int             flows;          /* number of flows */
    int             dpkts;          /* packets in flow */
    int             doctets;        /* octets in flow */
    int             fuptime;        /* Uptime at start of flow */
    int             luptime;        /* Uptime at end of flow */
    unsigned char   srcprefix[4];   /* source prefix */
    unsigned char   dstprefix[4];   /* destination prefix */
    char            dstmask;        /* Destination address mask bits */
    char            srcmask;        /* Source address mask bits */
    char            resv;           /* reserved */
    short           src_as;         /* AS of source address */
    short           dst_as;         /* AS of destination address */
    short           input_if;       /* SNMP index of input interface */
    short           output_if;      /* SNMP index of output interface */
};

struct flow8_astosmatrix_t {
    int     flows;          /* number of flows */
    int     dpkts;          /* packets in flow */
    int     doctets;        /* octets in flow */
    int     fuptime;        /* Uptime at start of flow */
    int     luptime;        /* Uptime at end of flow */
    short   src_as;         /* AS of source address */
    short   dst_as;         /* AS of destination address */
    short   input_if;       /* SNMP index of input interface */
    short   output_if;      /* SNMP index of output interface */
    char    tos;            /* IP ToS */
    char    resv1;          /* reserved */
    short   resv2;          /* reserved */
};

struct flow8_dsttosmatrix_t {
    int             flows;          /* number of flows */
    int             dpkts;          /* packets in flow */
    int             doctets;        /* octets in flow */
    int             fuptime;        /* Uptime at start of flow */
    int             luptime;        /* Uptime at end of flow */
    unsigned char   dstprefix[4];   /* destination prefix */
    char            dstmask;        /* Destination address mask bits */
    char            tos;            /* IP ToS */
    short           dst_as;         /* AS of destination address */
    short           output_if;      /* SNMP index of output interface */
    short           resv;           /* reserved */
};

struct flow8_prefixtosmatrix_t {
    int             flows;          /* number of flows */
    int             dpkts;          /* packets in flow */
    int             doctets;        /* octets in flow */
    int             fuptime;        /* Uptime at start of flow */
    int             luptime;        /* Uptime at end of flow */
    unsigned char   srcprefix[4];   /* source prefix */
    unsigned char   dstprefix[4];   /* destination prefix */
    char            dstmask;        /* Destination address mask bits */
    char            srcmask;        /* Source address mask bits */
    char            tos;            /* IP ToS */
    char            resv;           /* reserved */
    short           src_as;         /* AS of source address */
    short           dst_as;         /* AS of destination address */
    short           input_if;       /* SNMP index of input interface */
    short           output_if;      /* SNMP index of output interface */
};

struct flow8_prottosmatrix_t {
    int     flows;          /* number of flows */
    int     dpkts;          /* packets in flow */
    int     doctets;        /* octets in flow */
    int     fuptime;        /* Uptime at start of flow */
    int     luptime;        /* Uptime at end of flow */
    char    prot;           /* IP protocol */
    char    tos;            /* IP ToS */
    short   resv;           /* reserved */
    short   srcport;        /* source port */
    short   dstport;        /* destination port */
    short   input_if;       /* SNMP index of input interface */
    short   output_if;      /* SNMP index of output interface */
};

struct flow8_srctosmatrix_t {
    int             flows;          /* number of flows */
    int             dpkts;          /* packets in flow */
    int             doctets;        /* octets in flow */
    int             fuptime;        /* Uptime at start of flow */
    int             luptime;        /* Uptime at end of flow */
    unsigned char   srcprefix[4];   /* source prefix */
    char            srcmask;        /* Source address mask bits */
    char            tos;            /* IP ToS */
    short           src_as;         /* AS of source address */
    short           input_if;       /* SNMP index of input interface */
    short           resv;           /* reserved */
};

struct flow8_prefixportmatrix_t {
    int             flows;          /* number of flows */
    int             dpkts;          /* packets in flow */
    int             doctets;        /* octets in flow */
    int             fuptime;        /* Uptime at start of flow */
    int             luptime;        /* Uptime at end of flow */
    unsigned char   srcprefix[4];   /* source prefix */
    unsigned char   dstprefix[4];   /* destination prefix */
    char            dstmask;        /* Destination address mask bits */
    char            srcmask;        /* Source address mask bits */
    char            tos;            /* IP ToS */
    char            prot;           /* IP protocol */
    short           srcport;        /* source port */
    short           dstport;        /* destination port */
    short           input_if;       /* SNMP index of input interface */
    short           output_if;      /* SNMP index of output interface */
};

/* netflow v9 structure */
struct flowhdr9_t {
    short   version;        /* netflow version */
    short   count;          /* count of flows in packet */
    int     uptime;         /* msecs since router startup */
    int     secs;           /* secs since Epoch */
    int     flowseq;        /* seq counter of total flows sent */
    int     srcid;          /* vendor-specific */
};

/* field types */
#define FT9_IN_BYTES                    1
#define FT9_IN_PKTS                     2
#define FT9_FLOWS                       3
#define FT9_PROTOCOL                    4
#define FT9_SRC_TOS                     5
#define FT9_TCP_FLAGS                   6
#define FT9_L4_SRC_PORT                 7
#define FT9_IPV4_SRC_ADDR               8
#define FT9_SRC_MASK                    9
#define FT9_INPUT_SNMP                  10
#define FT9_L4_DST_PORT                 11
#define FT9_IPV4_DST_ADDR               12
#define FT9_DST_MASK                    13
#define FT9_OUTPUT_SNMP                 14
#define FT9_IPV4_NEXT_HOP               15
#define FT9_SRC_AS                      16
#define FT9_DST_AS                      17
#define FT9_BGP_IPV4_NEXT_HOP           18
#define FT9_MUL_DST_PKTS                19
#define FT9_MUL_DST_BYTES               20
#define FT9_LAST_SWITCHED               21
#define FT9_FIRST_SWITCHED              22
#define FT9_OUT_BYTES                   23
#define FT9_OUT_PKTS                    24
#define FT9_MIN_PKT_LNGTH               25
#define FT9_MAX_PKT_LNGTH               26
#define FT9_IPV6_SRC_ADDR               27
#define FT9_IPV6_DST_ADDR               28
#define FT9_IPV6_SRC_MASK               29
#define FT9_IPV6_DST_MASK               30
#define FT9_IPV6_FLOW_LABEL             31
#define FT9_ICMP_TYPE                   32
#define FT9_MUL_IGMP_TYPE               33
#define FT9_SAMPLING_INTERVAL           34
#define FT9_SAMPLING_ALGORITHM          35
#define FT9_FLOW_ACTIVE_TIMEOUT         36
#define FT9_FLOW_INACTIVE_TIMEOUT       37
#define FT9_ENGINE_TYPE                 38
#define FT9_ENGINE_ID                   39
#define FT9_TOTAL_BYTES_EXP             40
#define FT9_TOTAL_PKTS_EXP              41
#define FT9_TOTAL_FLOWS_EXP             42
#define FT9_VENDOR_SPECIFIC             43
#define FT9_IPV4_SRC_PREFIX             44
#define FT9_IPV4_DST_PREFIX             45
#define FT9_MPLS_TOP_LABEL_TYPE         46
#define FT9_MPLS_TOP_LABEL_IP_ADDR      47
#define FT9_FLOW_SAMPLER_ID             48
#define FT9_FLOW_SAMPLER_MODE           49
#define FT9_FLOW_SAMPLER_RANDOM_INTVL   50
#define FT9_VENDOR_SPECIFIC_2           51
#define FT9_MIN_TTL                     52
#define FT9_MAX_TTL                     53
#define FT9_IPV4_IDENT                  54
#define FT9_DST_TOS                     55
#define FT9_IN_SRC_MAC                  56
#define FT9_OUT_DST_MAC                 57
#define FT9_SRC_VLAN                    58
#define FT9_DST_VLAN                    59
#define FT9_IP_PROTOCOL_VERSION         60
#define FT9_DIRECTION                   61
#define FT9_IPV6_NEXT_HOP               62
#define FT9_BGP_IPV6_NEXT_HOP           63
#define FT9_IPV6_OPTION_HEADERS         64
#define FT9_VENDOR_SPECIFIC_3           65
#define FT9_VENDOR_SPECIFIC_4           66
#define FT9_VENDOR_SPECIFIC_5           67
#define FT9_VENDOR_SPECIFIC_6           68
#define FT9_VENDOR_SPECIFIC_7           69
#define FT9_MPLS_LABEL_1                70
#define FT9_MPLS_LABEL_2                71
#define FT9_MPLS_LABEL_3                72
#define FT9_MPLS_LABEL_4                73
#define FT9_MPLS_LABEL_5                74
#define FT9_MPLS_LABEL_6                75
#define FT9_MPLS_LABEL_7                76
#define FT9_MPLS_LABEL_8                77
#define FT9_MPLS_LABEL_9                78
#define FT9_MPLS_LABEL_10               79
#define FT9_IN_DST_MAC                  80
#define FT9_OUT_SRC_MAC                 81
#define FT9_IF_NAME                     82
#define FT9_IF_DESC                     83
#define FT9_SAMPLER_NAME                84
#define FT9_IN_PERMANENT_BYTES          85
#define FT9_IN_PERMANENT_PKTS           86
#define FT9_VENDOR_SPECIFIC_8           87

struct flowtemplate9_t {
    short   id;             /* flowset id */
    short   length;         /* length of template */
    struct {
        short   tmpl_id;    /* template id */
        short   fldcnt;     /* field count */
        struct {
            short   type;   /* field type */
            short   len;    /* field length */
        }       *fields;    /* fields in template */
    }       *tmpl;          /* templates */
};

struct flowdata9_t {
    short   tmpl_id;        /* template id */
    short   length;         /* length of record */
    void    *fields;        /* fields */
    char    *resv;          /* padding to 32-bit boundary */
};

/* netflow v1 structure */
struct flowhdr1_t {
    short   version;        /* netflow version */
    short   count;          /* count of flows in packet */
    int     uptime;         /* msecs since router startup */
    int     secs;           /* secs since Epoch */
    int     nsecs;          /* nanosecs since Epoch */
};

struct flow1_t {
    unsigned char   srcaddr[4];     /* source address */
    unsigned char   dstaddr[4];     /* destination address */
    unsigned char   nexthop[4];     /* nexthop address */
    short           input_if;       /* SNMP index of input interface */
    short           output_if;      /* SNMP index of output interface */
    int             dpkts;          /* packets in flow */
    int             doctets;        /* octets in flow */
    int             fuptime;        /* Uptime at start of flow */
    int             luptime;        /* Uptime at end of flow */
    short           srcport;        /* source port */
    short           dstport;        /* destination port */
    short           resv1;          /* reserved */
    char            prot;           /* IP protocol */
    char            tos;            /* IP ToS */
    char            flags;          /* TCP flags */
    char            tcp_retx_cnt;   /* number of mis-sequenced packets */
    char            tcp_retx_sec;   /* cumulative secs between */
                                    /* mis-sequenced packets */
    char            tcp_misseq_cnt; /* number of mis-sequenced packets sent */
    int             resv2;          /* reserved */
};

/* Internal nnfc structures */

struct nnfc_module {
    char    name[MAXPATHLEN];       /* module name */
    void    *handle;                /* module handle */
    void    (* connect) (char *, void *);   
                                    /* connection function */
    void    (* close) (void *);     /* close db function */
    void    (* insertpkt) (uint32_t, void *, void *);
                                    /* insert netflow packet */
    void    (* begin) (void *);     /* begin transaction */
    void    (* commit) (void *);    /* commit transaction */
    void*   (* create_thread_context) ();
                                    /* create library variables in thread context */
    struct nnfc_module *next;       /* next module */
};

struct value {
    char            *value;
    struct value    *next;
};

struct command {
    char            *option;
    struct value    *values;
    struct command  *next;
};

/* functions in nnfc_lib.c */
int n_socket(int, int, int);
void n_bind(int, struct sockaddr *, socklen_t);
ssize_t n_recvfrom(int, void *, size_t, int, struct sockaddr *, socklen_t *);
int n_select(int, fd_set *, fd_set *, fd_set *, struct timeval *);
void n_shutdown(int, int);
void n_daemon();
char *sock_ntop(const struct sockaddr *, socklen_t);
char *n_sock_ntop(const struct sockaddr *, socklen_t);

/* function in config.c */
void insertcommand(char *, struct command **);
void addvalue(char *, char *, struct command **);
void printconfig(struct command *);
char *getvalue(char *, struct command *);
struct value *getallvalues(char *, struct command *);
int readcfg(char *, struct command **);
int freecmds(struct command *);

/* functions in module.c */
void module_close(struct nnfc_module *);
int module_open(struct nnfc_module *);

#endif	/* LIBNNFC_H */
