/**************************************************************************
 * Copyright (c) 2004-2006, Andrew Klyachkin <andrewk@region73.ru>        *
 * All rights reserved.                                                   *
 *                                                                        *
 * Redistribution and use in source and binary forms, with or without     *
 * modification, are permitted provided that the following conditions     *
 * are met:                                                               *
 *                                                                        *
 * 1. Redistributions of source code must retain the above copyright      *
 *    notice, this list of conditions and the following disclaimer.       *
 * 2. Redistributions in binary form must reproduce the above copyright   *
 *    notice, this list of conditions and the following disclaimer in the *
 *    documentation and/or other materials provided with the distribution.*
 * 3. Neither the name of the Andrew Klyachkin nor the names of his       *
 *    contributors may be used to endorse or promote products derived     *
 *    from this software without specific prior written permission.       *
 *                                                                        *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS    *
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT      *
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR  *
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT   *
 * AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, *
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       *
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  *
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  *
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    *
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  *
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   *
 **************************************************************************/

#include <stdio.h>
#include <dlfcn.h>
#include <syslog.h>
#include <stdlib.h>
#include <pthread.h>

#include "libnnfc.h"

void module_close(struct nnfc_module *m)
{
    if (m->handle != NULL) {
        dlclose(m->handle);
        m->handle = NULL;
    }
}

int module_open(struct nnfc_module *m)
{
    char    *error;

    if (!m->name) return -1;
    m->handle = dlopen(m->name, RTLD_NOW);
    if (!m->handle) {
        syslog(LOG_ERR, "module_open: cannot open module %s\n", m->name);
        syslog(LOG_ERR, "module_open: %s\n", dlerror());
        return -1;
    }
    dlerror();
    LOADFUNC(m->connect, "_connect");
    LOADFUNC(m->close, "_close");
    LOADFUNC(m->insertpkt, "insertpkt");
    LOADFUNC(m->begin, "begin");
    LOADFUNC(m->commit, "commit");
    LOADFUNC(m->create_thread_context, "_create_thread_context");

    return 0;
}
