/**************************************************************************
 * Copyright (c) 2004-2006, Andrew Klyachkin <andrewk@region73.ru>        *
 * All rights reserved.                                                   *
 *                                                                        *
 * Redistribution and use in source and binary forms, with or without     *
 * modification, are permitted provided that the following conditions     *
 * are met:                                                               *
 *                                                                        *
 * 1. Redistributions of source code must retain the above copyright      *
 *    notice, this list of conditions and the following disclaimer.       *
 * 2. Redistributions in binary form must reproduce the above copyright   *
 *    notice, this list of conditions and the following disclaimer in the *
 *    documentation and/or other materials provided with the distribution.*
 * 3. Neither the name of the Andrew Klyachkin nor the names of his       *
 *    contributors may be used to endorse or promote products derived     *
 *    from this software without specific prior written permission.       *
 *                                                                        *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS    *
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT      *
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR  *
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT   *
 * AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, *
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       *
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  *
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  *
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    *
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  *
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   *
 **************************************************************************/

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <syslog.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <sys/time.h>
#include <time.h>
#include <string.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>

#include <libnnfc.h>

/*
 * Parameters for this plugins are:
 * filename - file name template to store netflow dump (as for strftime)
 */

static int  fflow = -1;
char        *ftmpl;

void * _create_thread_context()
{
}

/* creating file */
void _connect(char *fname, void *param)
{
    struct command  *commands = 0;
    int             n = 0;

    if (!fname) 
        return;

    if (readcfg(fname, &commands)) {
        syslog(LOG_ERR, "Parse error!\n");
        return;
    }

    CFG_GETS(ftmpl, "filename", commands);

    freecmds(commands);
}

void begin(void *param)
{
    char        fnname[MAXPATHLEN];
	time_t      ttm;
    struct tm   *stm;

    ttm = time(NULL);
    stm = localtime(&ttm);
    strftime(fnname, MAXPATHLEN, ftmpl, stm);
    fflow = open(fnname, O_WRONLY | O_CREAT | O_APPEND, S_IRUSR | S_IWUSR);
    if (fflow < 0) {
        syslog(LOG_ERR, "Cannot open file %s for writing: %s!",
            fnname, strerror(errno));
        return;
    }
    return;
}

void commit(void *param)
{
    close(fflow);
    fflow = -1;
    return;
}

/* closes connection to database */
void _close(void *param)
{
    if (fflow >= 0) close(fflow);
    fflow = -1;

    if (ftmpl) free(ftmpl);
}

void insertpkt(uint32_t address, void *packet, void *param)
{
    struct flowhdr1_t   *hdr = (struct flowhdr1_t *)packet;
    short               version = ntohs(*((short *)packet));
    int                 i = 0, n = 0, len = 0;

    if (!packet)
        return;

    if (fflow < 0)
        return;

    switch (version) {
    case 1:
        len = sizeof(struct flowhdr1_t) +
            ntohs(hdr->count) * sizeof(struct flow1_t);
        break;
    case 5:
        len = sizeof(struct flowhdr5_t) +
            ntohs(hdr->count) * sizeof(struct flow5_t);
        break;
    case 7:
        len = sizeof(struct flowhdr7_t) +
            ntohs(hdr->count) * sizeof(struct flow7_t);
        break;
    default:
        syslog(LOG_ERR, "dump: unknown netflow version %d", version);
        return;
    }

    n = write(fflow, packet, len);
    if (n < len) {
        syslog(LOG_ERR, "dump: %s", strerror(errno));
    } 
    if (fsync(fflow) < 0) {
        syslog(LOG_ERR, "dump: %s", strerror(errno));
    }
}

