/**************************************************************************
 * Copyright (c) 2004-2006, Andrew Klyachkin <andrewk@region73.ru>        *
 * All rights reserved.                                                   *
 *                                                                        *
 * Redistribution and use in source and binary forms, with or without     *
 * modification, are permitted provided that the following conditions     *
 * are met:                                                               *
 *                                                                        *
 * 1. Redistributions of source code must retain the above copyright      *
 *    notice, this list of conditions and the following disclaimer.       *
 * 2. Redistributions in binary form must reproduce the above copyright   *
 *    notice, this list of conditions and the following disclaimer in the *
 *    documentation and/or other materials provided with the distribution.*
 * 3. Neither the name of the Andrew Klyachkin nor the names of his       *
 *    contributors may be used to endorse or promote products derived     *
 *    from this software without specific prior written permission.       *
 *                                                                        *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS    *
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT      *
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR  *
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT   *
 * AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, *
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       *
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  *
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  *
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    *
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  *
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   *
 **************************************************************************/

#include "mysql_module.h"

void * _create_thread_context()
{
    thread_context *ctx = (thread_context*) malloc(sizeof(thread_context));
    ctx->connected = 0;
    ctx->ltime = 1;
    ctx->dbhost = 0;
    ctx->dbport;
    ctx->dbuser = 0;
    ctx->dbpass = 0;
    ctx->dbname = 0;
    ctx->dbheader = 0;
    ctx->dbflows = 0;
    return ctx;
}

/* establishes connection to database */
void _connect(char *fname, void *dbparam)
{
    struct command  *commands = 0;
    int             n = 0;
    char            check_for_transactions[1024], current_engine[50];
    MYSQL_RES       *trans_result;
    MYSQL_ROW       trans_result_row;
    thread_context *ctx = (thread_context *) dbparam;

    if (!fname) return;

    if (!ctx->config[0]) {
        memset(&ctx->config, 0, MAXPATHLEN);
        strncpy(ctx->config, fname, MAXPATHLEN);
    }

    if (readcfg(ctx->config, &commands)) {
        syslog(LOG_ERR, "mysql: parse error!");
        return;
    }

    CFG_GETI(ctx->ltime, "localtime", commands);
    CFG_GETI(ctx->dbport, "dbport", commands);

    CFG_GETS(ctx->dbhost, "dbhost", commands);
    CFG_GETS(ctx->dbuser, "dbuser", commands);
    CFG_GETS(ctx->dbpass, "dbpass", commands);
    CFG_GETS(ctx->dbname, "dbname", commands);
    CFG_GETS(ctx->dbheader, "dbheader", commands);
    CFG_GETS(ctx->dbflows, "dbflows", commands);

    freecmds(commands);

    if (!ctx->dbhost || !ctx->dbuser || !ctx->dbpass || !ctx->dbname || !ctx->dbheader || !ctx->dbflows) {
        syslog(LOG_ERR, "mysql: required parameter missing ;)");
        return;
    }

    mysql_init(&ctx->conn);
    mysql_options(&ctx->conn, MYSQL_READ_DEFAULT_GROUP, "nnfc");
    if (!mysql_real_connect(&ctx->conn, ctx->dbhost, ctx->dbuser, ctx->dbpass, 
        ctx->dbname, ctx->dbport, NULL, 0)) {

        syslog(LOG_ERR, "Couldn't connect to mysql database");
        syslog(LOG_ERR, mysql_error(&ctx->conn));
        return;
    }

    /* Check whether MySQL engine supports transactions for this table */
    sprintf(check_for_transactions, "select engine from information_schema.tables "
                "where table_name = \'%s\'", ctx->dbflows);

    if (!mysql_real_query(&ctx->conn, check_for_transactions, strlen(check_for_transactions)))
    {
        trans_result = mysql_store_result(&ctx->conn);
        trans_result_row = mysql_fetch_row(trans_result);
        // printf("Current engine is %s\n", trans_result_row[0]);
        strcpy(current_engine, trans_result_row[0]);
        if ((strcmp(current_engine, "InnoDB") == 0) || (strcmp(current_engine, "BDB") == 0))
        {
            ctx->trans_enabled = 1;
            printf("Current engine is %s. Transaction support is enabled.\n", current_engine);
        }
        else 
        {
            ctx->trans_enabled = 0;
            printf("Current engine is %s. Transaction support is disabled.\n", current_engine);
        }
        mysql_free_result(trans_result);
    }

    /* Create SQL-insert statements */
    sprintf(ctx->flow1hdr_t_insert, "insert into %s (address, version_id, "
            "count_id, uptime, secs, nsecs) values (?, ?, ?, ?, ?, ?)",
            ctx->dbheader);

    sprintf(ctx->flow5hdr_t_insert, "insert into %s (address, version_id, "
            "count_id, uptime, secs, nsecs, flowseq, engine, eng_id, sampling) "
            "values (?, ?, ?, ?, ?, ?, ?, ?, ?, ?)", ctx->dbheader);

    sprintf(ctx->flow7hdr_t_insert, "insert into %s (address, version_id, "
            "count_id, uptime, secs, nsecs, flowseq, resv) values "
            "(?, ?, ?, ?, ?, ?, ?, ?)", ctx->dbheader);

    sprintf(ctx->flow1_t_insert, "insert into %s (address, version_id, number_id, "
                "srcaddr, dstaddr, nexthop, input_if, output_if, dpkts, doctets, "
                "fuptime, luptime, srcport, dstport, resv1, flags, prot, "
                "tos, tcp_retx_cnt, tcp_retx_sec, tcp_misseq_cnt, resv2) "
                "values (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, "
                        "?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)", ctx->dbflows);

    sprintf(ctx->flow5_t_insert, "insert into %s (address, version_id, number_id, "
                "srcaddr, dstaddr, nexthop, input_if, output_if, dpkts, doctets, "
                "fuptime, luptime, srcport, dstport, resv1, flags, prot, tos, "
                "srcmask, dstmask) "
                "values (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, "
                        "?, ?, ?, ?, ?, ?, ?, ?, ?, ?)", ctx->dbflows);

    sprintf(ctx->flow7_t_insert, "insert into %s (address, version_id, number_id, "
                "srcaddr, dstaddr, nexthop, input_if, output_if, dpkts, doctets, "
                "fuptime, luptime, srcport, dstport, mode, flags, prot, "
                "tos, src_as, dst_as, srcmask, dstmask, resv2, router) "
                "values (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, "
                        "?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)", ctx->dbflows);

    /* initialize statement handlers */
    ctx->flow1hdr_t_stmt = mysql_stmt_init(&ctx->conn);
    if (!ctx->flow1hdr_t_stmt)
    {
        syslog(LOG_ERR, "mysql_stmt_init(flow1hdr_t), out of memory");
        return;
    }

    ctx->flow5hdr_t_stmt = mysql_stmt_init(&ctx->conn);
    if (!ctx->flow5hdr_t_stmt)
    {
        syslog(LOG_ERR, "mysql_stmt_init(flow5hdr_t), out of memory");
        return;
    }

    ctx->flow7hdr_t_stmt = mysql_stmt_init(&ctx->conn);
    if (!ctx->flow7hdr_t_stmt)
    {
        syslog(LOG_ERR, "mysql_stmt_init(flow7hdr_t), out of memory");
        return;
    }

    ctx->flow1_t_stmt = mysql_stmt_init(&ctx->conn);
    if (!ctx->flow1_t_stmt)
    {
        syslog(LOG_ERR, "mysql_stmt_init(flow1_t), out of memory");
        return;
    }

    ctx->flow5_t_stmt = mysql_stmt_init(&ctx->conn);
    if (!ctx->flow5_t_stmt)
    {
        syslog(LOG_ERR, "mysql_stmt_init(flow5_t), out of memory");
        return;
    }

    ctx->flow7_t_stmt = mysql_stmt_init(&ctx->conn);
    if (!ctx->flow7_t_stmt)
    {
        syslog(LOG_ERR, "mysql_stmt_init(flow7_t), out of memory");
        return;
    }

    /* prepare statements */
    if (mysql_stmt_prepare(ctx->flow1hdr_t_stmt, ctx->flow1hdr_t_insert,
                strlen(ctx->flow1hdr_t_insert)))
    {
        syslog(LOG_ERR, "mysql_stmt_prepare(), flow1hdr_t_insert failed\n");
        syslog(LOG_ERR, mysql_stmt_error(ctx->flow1hdr_t_stmt));
        return;
    }

    if (mysql_stmt_prepare(ctx->flow5hdr_t_stmt, ctx->flow5hdr_t_insert,
                strlen(ctx->flow5hdr_t_insert)))
    {
        syslog(LOG_ERR, "mysql_stmt_prepare(), flow5hdr_t_insert failed\n");
        syslog(LOG_ERR, mysql_stmt_error(ctx->flow5hdr_t_stmt));
        return;
    }

    if (mysql_stmt_prepare(ctx->flow7hdr_t_stmt, ctx->flow7hdr_t_insert,
                strlen(ctx->flow7hdr_t_insert)))
    {
        syslog(LOG_ERR, "mysql_stmt_prepare(), flow7hdr_t_insert failed\n");
        syslog(LOG_ERR, mysql_stmt_error(ctx->flow7hdr_t_stmt));
        return;
    }

    if (mysql_stmt_prepare(ctx->flow1_t_stmt, ctx->flow1_t_insert,
                strlen(ctx->flow1_t_insert)))
    {
        syslog(LOG_ERR, "mysql_stmt_prepare(), flow1_t_insert failed\n");
        syslog(LOG_ERR, mysql_stmt_error(ctx->flow1_t_stmt));
        return;
    }

    if (mysql_stmt_prepare(ctx->flow5_t_stmt, ctx->flow5_t_insert,
                strlen(ctx->flow5_t_insert)))
    {
        syslog(LOG_ERR, "mysql_stmt_prepare(), flow5_t_insert failed\n");
        syslog(LOG_ERR, mysql_stmt_error(ctx->flow5_t_stmt));
        return;
    }

    if (mysql_stmt_prepare(ctx->flow7_t_stmt, ctx->flow7_t_insert,
                strlen(ctx->flow7_t_insert)))
    {
        syslog(LOG_ERR, "mysql_stmt_prepare(), flow7_t_insert failed\n");
        syslog(LOG_ERR, mysql_stmt_error(ctx->flow7_t_stmt));
        return;
    }

    /* bind the data for parameters */
    memset(ctx->flow1hdr_t_bind, 0, sizeof(ctx->flow1hdr_t_bind));
    memset(ctx->flow5hdr_t_bind, 0, sizeof(ctx->flow5hdr_t_bind));
    memset(ctx->flow7hdr_t_bind, 0, sizeof(ctx->flow7hdr_t_bind));
    memset(ctx->flow1_t_bind, 0, sizeof(ctx->flow1_t_bind));
    memset(ctx->flow5_t_bind, 0, sizeof(ctx->flow5_t_bind));
    memset(ctx->flow7_t_bind, 0, sizeof(ctx->flow7_t_bind));

    /* We neet to create bind array in order that 
     * fields appear in SQL-insert statement */

    /* netflow v1 header */
    /* address */
    ctx->flow1hdr_t_bind[0].buffer_type = MYSQL_TYPE_LONG;
    ctx->flow1hdr_t_bind[0].buffer = &ctx->flowhdr_t_packet.address;
    ctx->flow1hdr_t_bind[0].is_null = 0;
    ctx->flow1hdr_t_bind[0].is_unsigned = 1;
    ctx->flow1hdr_t_bind[0].length = 0;

    /* version_id */
    ctx->flow1hdr_t_bind[1].buffer_type = MYSQL_TYPE_SHORT;
    ctx->flow1hdr_t_bind[1].buffer = &ctx->flowhdr_t_packet.version;
    ctx->flow1hdr_t_bind[1].is_null = 0;
    ctx->flow1hdr_t_bind[1].is_unsigned = 1;
    ctx->flow1hdr_t_bind[1].length = 0;

    /* count_id */
    ctx->flow1hdr_t_bind[2].buffer_type = MYSQL_TYPE_SHORT;
    ctx->flow1hdr_t_bind[2].buffer = &ctx->flowhdr_t_packet.count;
    ctx->flow1hdr_t_bind[2].is_null = 0;
    ctx->flow1hdr_t_bind[2].is_unsigned = 1;
    ctx->flow1hdr_t_bind[2].length = 0;

    /* uptime */
    ctx->flow1hdr_t_bind[3].buffer_type = MYSQL_TYPE_LONG;
    ctx->flow1hdr_t_bind[3].buffer = &ctx->flowhdr_t_packet.uptime;
    ctx->flow1hdr_t_bind[3].is_null = 0;
    ctx->flow1hdr_t_bind[3].is_unsigned = 1;
    ctx->flow1hdr_t_bind[3].length = 0;

    /* secs */
    ctx->flow1hdr_t_bind[4].buffer_type = MYSQL_TYPE_LONG;
    ctx->flow1hdr_t_bind[4].buffer = &ctx->flowhdr_t_packet.secs;
    ctx->flow1hdr_t_bind[4].is_null = 0;
    ctx->flow1hdr_t_bind[4].is_unsigned = 1;
    ctx->flow1hdr_t_bind[4].length = 0;

    /* nsecs */
    ctx->flow1hdr_t_bind[5].buffer_type = MYSQL_TYPE_LONG;
    ctx->flow1hdr_t_bind[5].buffer = &ctx->flowhdr_t_packet.nsecs;
    ctx->flow1hdr_t_bind[5].is_null = 0;
    ctx->flow1hdr_t_bind[5].is_unsigned = 1;
    ctx->flow1hdr_t_bind[5].length = 0;

    /* netflow v5 header */
    /* address */
    ctx->flow5hdr_t_bind[0].buffer_type = MYSQL_TYPE_LONG;
    ctx->flow5hdr_t_bind[0].buffer = &ctx->flowhdr_t_packet.address;
    ctx->flow5hdr_t_bind[0].is_null = 0;
    ctx->flow5hdr_t_bind[0].is_unsigned = 1;
    ctx->flow5hdr_t_bind[0].length = 0;

    /* version_id */
    ctx->flow5hdr_t_bind[1].buffer_type = MYSQL_TYPE_SHORT;
    ctx->flow5hdr_t_bind[1].buffer = &ctx->flowhdr_t_packet.version;
    ctx->flow5hdr_t_bind[1].is_null = 0;
    ctx->flow5hdr_t_bind[1].is_unsigned = 1;
    ctx->flow5hdr_t_bind[1].length = 0;

    /* count_id */
    ctx->flow5hdr_t_bind[2].buffer_type = MYSQL_TYPE_SHORT;
    ctx->flow5hdr_t_bind[2].buffer = &ctx->flowhdr_t_packet.count;
    ctx->flow5hdr_t_bind[2].is_null = 0;
    ctx->flow5hdr_t_bind[2].is_unsigned = 1;
    ctx->flow5hdr_t_bind[2].length = 0;

    /* uptime */
    ctx->flow5hdr_t_bind[3].buffer_type = MYSQL_TYPE_LONG;
    ctx->flow5hdr_t_bind[3].buffer = &ctx->flowhdr_t_packet.uptime;
    ctx->flow5hdr_t_bind[3].is_null = 0;
    ctx->flow5hdr_t_bind[3].is_unsigned = 1;
    ctx->flow5hdr_t_bind[3].length = 0;

    /* secs */
    ctx->flow5hdr_t_bind[4].buffer_type = MYSQL_TYPE_LONG;
    ctx->flow5hdr_t_bind[4].buffer = &ctx->flowhdr_t_packet.secs;
    ctx->flow5hdr_t_bind[4].is_null = 0;
    ctx->flow5hdr_t_bind[4].is_unsigned = 1;
    ctx->flow5hdr_t_bind[4].length = 0;

    /* nsecs */
    ctx->flow5hdr_t_bind[5].buffer_type = MYSQL_TYPE_LONG;
    ctx->flow5hdr_t_bind[5].buffer = &ctx->flowhdr_t_packet.nsecs;
    ctx->flow5hdr_t_bind[5].is_null = 0;
    ctx->flow5hdr_t_bind[5].is_unsigned = 1;
    ctx->flow5hdr_t_bind[5].length = 0;

    /* flowseq */
    ctx->flow5hdr_t_bind[6].buffer_type = MYSQL_TYPE_LONG;
    ctx->flow5hdr_t_bind[6].buffer = &ctx->flowhdr_t_packet.flowseq;
    ctx->flow5hdr_t_bind[6].is_null = 0;
    ctx->flow5hdr_t_bind[6].is_unsigned = 1;
    ctx->flow5hdr_t_bind[6].length = 0;

    /* engine */
    ctx->flow5hdr_t_bind[7].buffer_type = MYSQL_TYPE_TINY;
    ctx->flow5hdr_t_bind[7].buffer = &ctx->flowhdr_t_packet.engine;
    ctx->flow5hdr_t_bind[7].is_null = 0;
    ctx->flow5hdr_t_bind[7].is_unsigned = 1;
    ctx->flow5hdr_t_bind[7].length = 0;

    /* eng_id */
    ctx->flow5hdr_t_bind[8].buffer_type = MYSQL_TYPE_TINY;
    ctx->flow5hdr_t_bind[8].buffer = &ctx->flowhdr_t_packet.eng_id;
    ctx->flow5hdr_t_bind[8].is_null = 0;
    ctx->flow5hdr_t_bind[8].is_unsigned = 1;
    ctx->flow5hdr_t_bind[8].length = 0;

    /* sampling */
    ctx->flow5hdr_t_bind[9].buffer_type = MYSQL_TYPE_SHORT;
    ctx->flow5hdr_t_bind[9].buffer = &ctx->flowhdr_t_packet.sampling;
    ctx->flow5hdr_t_bind[9].is_null = 0;
    ctx->flow5hdr_t_bind[9].is_unsigned = 1;
    ctx->flow5hdr_t_bind[9].length = 0;

    /* netflow v7 header */
    /* address */
    ctx->flow7hdr_t_bind[0].buffer_type = MYSQL_TYPE_LONG;
    ctx->flow7hdr_t_bind[0].buffer = &ctx->flowhdr_t_packet.address;
    ctx->flow7hdr_t_bind[0].is_null = 0;
    ctx->flow7hdr_t_bind[0].is_unsigned = 1;
    ctx->flow7hdr_t_bind[0].length = 0;

    /* version_id */
    ctx->flow7hdr_t_bind[1].buffer_type = MYSQL_TYPE_SHORT;
    ctx->flow7hdr_t_bind[1].buffer = &ctx->flowhdr_t_packet.version;
    ctx->flow7hdr_t_bind[1].is_null = 0;
    ctx->flow7hdr_t_bind[1].is_unsigned = 1;
    ctx->flow7hdr_t_bind[1].length = 0;

    /* count_id */
    ctx->flow7hdr_t_bind[2].buffer_type = MYSQL_TYPE_SHORT;
    ctx->flow7hdr_t_bind[2].buffer = &ctx->flowhdr_t_packet.count;
    ctx->flow7hdr_t_bind[2].is_null = 0;
    ctx->flow7hdr_t_bind[2].is_unsigned = 1;
    ctx->flow7hdr_t_bind[2].length = 0;

    /* uptime */
    ctx->flow7hdr_t_bind[3].buffer_type = MYSQL_TYPE_LONG;
    ctx->flow7hdr_t_bind[3].buffer = &ctx->flowhdr_t_packet.uptime;
    ctx->flow7hdr_t_bind[3].is_null = 0;
    ctx->flow7hdr_t_bind[3].is_unsigned = 1;
    ctx->flow7hdr_t_bind[3].length = 0;

    /* secs */
    ctx->flow7hdr_t_bind[4].buffer_type = MYSQL_TYPE_LONG;
    ctx->flow7hdr_t_bind[4].buffer = &ctx->flowhdr_t_packet.secs;
    ctx->flow7hdr_t_bind[4].is_null = 0;
    ctx->flow7hdr_t_bind[4].is_unsigned = 1;
    ctx->flow7hdr_t_bind[4].length = 0;

    /* nsecs */
    ctx->flow7hdr_t_bind[5].buffer_type = MYSQL_TYPE_LONG;
    ctx->flow7hdr_t_bind[5].buffer = &ctx->flowhdr_t_packet.nsecs;
    ctx->flow7hdr_t_bind[5].is_null = 0;
    ctx->flow7hdr_t_bind[5].is_unsigned = 1;
    ctx->flow7hdr_t_bind[5].length = 0;

    /* flowseq */
    ctx->flow7hdr_t_bind[6].buffer_type = MYSQL_TYPE_LONG;
    ctx->flow7hdr_t_bind[6].buffer = &ctx->flowhdr_t_packet.flowseq;
    ctx->flow7hdr_t_bind[6].is_null = 0;
    ctx->flow7hdr_t_bind[6].is_unsigned = 1;
    ctx->flow7hdr_t_bind[6].length = 0;

    /* resv */
    ctx->flow7hdr_t_bind[7].buffer_type = MYSQL_TYPE_LONG;
    ctx->flow7hdr_t_bind[7].buffer = &ctx->flowhdr_t_packet.resv;
    ctx->flow7hdr_t_bind[7].is_null = 0;
    ctx->flow7hdr_t_bind[7].is_unsigned = 1;
    ctx->flow7hdr_t_bind[7].length = 0;

    /* netflow v1 */
    /* address */
    ctx->flow1_t_bind[0].buffer_type = MYSQL_TYPE_LONG;
    ctx->flow1_t_bind[0].buffer = &ctx->flow_t_packet.address;
    ctx->flow1_t_bind[0].is_null = 0;
    ctx->flow1_t_bind[0].is_unsigned = 1;
    ctx->flow1_t_bind[0].length = 0;

    /* version */
    ctx->flow1_t_bind[1].buffer_type = MYSQL_TYPE_SHORT;
    ctx->flow1_t_bind[1].buffer = &ctx->flow_t_packet.version;
    ctx->flow1_t_bind[1].is_null = 0;
    ctx->flow1_t_bind[1].is_unsigned = 1;
    ctx->flow1_t_bind[1].length = 0;

    /* number */
    ctx->flow1_t_bind[2].buffer_type = MYSQL_TYPE_LONG;
    ctx->flow1_t_bind[2].buffer = &ctx->flow_t_packet.number;
    ctx->flow1_t_bind[2].is_null = 0;
    ctx->flow1_t_bind[2].is_unsigned = 1;
    ctx->flow1_t_bind[2].length = 0;

    /* srcaddr */
    ctx->flow1_t_bind[3].buffer_type = MYSQL_TYPE_LONG;
    ctx->flow1_t_bind[3].buffer = &ctx->flow_t_packet.srcaddr;
    ctx->flow1_t_bind[3].is_null = 0;
    ctx->flow1_t_bind[3].is_unsigned = 1;
    ctx->flow1_t_bind[3].length = 0;

    /* dstaddr */
    ctx->flow1_t_bind[4].buffer_type = MYSQL_TYPE_LONG;
    ctx->flow1_t_bind[4].buffer = &ctx->flow_t_packet.dstaddr;
    ctx->flow1_t_bind[4].is_null = 0;
    ctx->flow1_t_bind[4].is_unsigned = 1;
    ctx->flow1_t_bind[4].length = 0;

    /* nexthop */
    ctx->flow1_t_bind[5].buffer_type = MYSQL_TYPE_LONG;
    ctx->flow1_t_bind[5].buffer = &ctx->flow_t_packet.nexthop;
    ctx->flow1_t_bind[5].is_null = 0;
    ctx->flow1_t_bind[5].is_unsigned = 1;
    ctx->flow1_t_bind[5].length = 0;

    /* input_if */
    ctx->flow1_t_bind[6].buffer_type = MYSQL_TYPE_SHORT;
    ctx->flow1_t_bind[6].buffer = &ctx->flow_t_packet.input_if;
    ctx->flow1_t_bind[6].is_null = 0;
    ctx->flow1_t_bind[6].is_unsigned = 1;
    ctx->flow1_t_bind[6].length = 0;

    /* output_if */
    ctx->flow1_t_bind[7].buffer_type = MYSQL_TYPE_SHORT;
    ctx->flow1_t_bind[7].buffer = &ctx->flow_t_packet.output_if;
    ctx->flow1_t_bind[7].is_null = 0;
    ctx->flow1_t_bind[7].is_unsigned = 1;
    ctx->flow1_t_bind[7].length = 0;

    /* dpkts */
    ctx->flow1_t_bind[8].buffer_type = MYSQL_TYPE_LONG;
    ctx->flow1_t_bind[8].buffer = &ctx->flow_t_packet.dpkts;
    ctx->flow1_t_bind[8].is_null = 0;
    ctx->flow1_t_bind[8].is_unsigned = 1;
    ctx->flow1_t_bind[8].length = 0;

    /* doctets */
    ctx->flow1_t_bind[9].buffer_type = MYSQL_TYPE_LONG;
    ctx->flow1_t_bind[9].buffer = &ctx->flow_t_packet.doctets;
    ctx->flow1_t_bind[9].is_null = 0;
    ctx->flow1_t_bind[9].is_unsigned = 1;
    ctx->flow1_t_bind[9].length = 0;

    /* fuptime */
    ctx->flow1_t_bind[10].buffer_type = MYSQL_TYPE_LONG;
    ctx->flow1_t_bind[10].buffer = &ctx->flow_t_packet.fuptime;
    ctx->flow1_t_bind[10].is_null = 0;
    ctx->flow1_t_bind[10].is_unsigned = 1;
    ctx->flow1_t_bind[10].length = 0;

    /* luptime */
    ctx->flow1_t_bind[11].buffer_type = MYSQL_TYPE_LONG;
    ctx->flow1_t_bind[11].buffer = &ctx->flow_t_packet.luptime;
    ctx->flow1_t_bind[11].is_null = 0;
    ctx->flow1_t_bind[11].is_unsigned = 1;
    ctx->flow1_t_bind[11].length = 0;

    /* srcport */
    ctx->flow1_t_bind[12].buffer_type = MYSQL_TYPE_SHORT;
    ctx->flow1_t_bind[12].buffer = &ctx->flow_t_packet.srcport;
    ctx->flow1_t_bind[12].is_null = 0;
    ctx->flow1_t_bind[12].is_unsigned = 1;
    ctx->flow1_t_bind[12].length = 0;

    /* dstport */
    ctx->flow1_t_bind[13].buffer_type = MYSQL_TYPE_SHORT;
    ctx->flow1_t_bind[13].buffer = &ctx->flow_t_packet.dstport;
    ctx->flow1_t_bind[13].is_null = 0;
    ctx->flow1_t_bind[13].is_unsigned = 1;
    ctx->flow1_t_bind[13].length = 0;

    /* resv1 */
    ctx->flow1_t_bind[14].buffer_type = MYSQL_TYPE_SHORT;
    ctx->flow1_t_bind[14].buffer = &ctx->flow_t_packet.resv1;
    ctx->flow1_t_bind[14].is_null = 0;
    ctx->flow1_t_bind[14].is_unsigned = 1;
    ctx->flow1_t_bind[14].length = 0;

    /* flags */
    ctx->flow1_t_bind[15].buffer_type = MYSQL_TYPE_TINY;
    ctx->flow1_t_bind[15].buffer = &ctx->flow_t_packet.flags;
    ctx->flow1_t_bind[15].is_null = 0;
    ctx->flow1_t_bind[15].is_unsigned = 1;
    ctx->flow1_t_bind[15].length = 0;

    /* prot */
    ctx->flow1_t_bind[16].buffer_type = MYSQL_TYPE_TINY;
    ctx->flow1_t_bind[16].buffer = &ctx->flow_t_packet.prot;
    ctx->flow1_t_bind[16].is_null = 0;
    ctx->flow1_t_bind[16].is_unsigned = 1;
    ctx->flow1_t_bind[16].length = 0;

    /* tos */
    ctx->flow1_t_bind[17].buffer_type = MYSQL_TYPE_TINY;
    ctx->flow1_t_bind[17].buffer = &ctx->flow_t_packet.tos;
    ctx->flow1_t_bind[17].is_null = 0;
    ctx->flow1_t_bind[17].is_unsigned = 1;
    ctx->flow1_t_bind[17].length = 0;

    /* tcp_retx_cnt */
    ctx->flow1_t_bind[18].buffer_type = MYSQL_TYPE_TINY;
    ctx->flow1_t_bind[18].buffer = &ctx->flow_t_packet.tcp_retx_cnt;
    ctx->flow1_t_bind[18].is_null = 0;
    ctx->flow1_t_bind[18].is_unsigned = 1;
    ctx->flow1_t_bind[18].length = 0;

    /* tcp_retx_sec */
    ctx->flow1_t_bind[19].buffer_type = MYSQL_TYPE_TINY;
    ctx->flow1_t_bind[19].buffer = &ctx->flow_t_packet.tcp_retx_sec;
    ctx->flow1_t_bind[19].is_null = 0;
    ctx->flow1_t_bind[19].is_unsigned = 1;
    ctx->flow1_t_bind[19].length = 0;

    /* tcp_misseq_cnt */
    ctx->flow1_t_bind[20].buffer_type = MYSQL_TYPE_TINY;
    ctx->flow1_t_bind[20].buffer = &ctx->flow_t_packet.tcp_misseq_cnt;
    ctx->flow1_t_bind[20].is_null = 0;
    ctx->flow1_t_bind[20].is_unsigned = 1;
    ctx->flow1_t_bind[20].length = 0;

    /* resv2 */
    ctx->flow1_t_bind[21].buffer_type = MYSQL_TYPE_LONG;
    ctx->flow1_t_bind[21].buffer = &ctx->flow_t_packet.resv2;
    ctx->flow1_t_bind[21].is_null = 0;
    ctx->flow1_t_bind[21].is_unsigned = 1;
    ctx->flow1_t_bind[21].length = 0;

    /* netflow v5 */
    /* address */
    ctx->flow5_t_bind[0].buffer_type = MYSQL_TYPE_LONG;
    ctx->flow5_t_bind[0].buffer = &ctx->flow_t_packet.address;
    ctx->flow5_t_bind[0].is_null = 0;
    ctx->flow5_t_bind[0].is_unsigned = 1;
    ctx->flow5_t_bind[0].length = 0;

    /* version */
    ctx->flow5_t_bind[1].buffer_type = MYSQL_TYPE_SHORT;
    ctx->flow5_t_bind[1].buffer = &ctx->flow_t_packet.version;
    ctx->flow5_t_bind[1].is_null = 0;
    ctx->flow5_t_bind[1].is_unsigned = 1;
    ctx->flow5_t_bind[1].length = 0;

    /* number */
    ctx->flow5_t_bind[2].buffer_type = MYSQL_TYPE_LONG;
    ctx->flow5_t_bind[2].buffer = &ctx->flow_t_packet.number;
    ctx->flow5_t_bind[2].is_null = 0;
    ctx->flow5_t_bind[2].is_unsigned = 1;
    ctx->flow5_t_bind[2].length = 0;

    /* srcaddr */
    ctx->flow5_t_bind[3].buffer_type = MYSQL_TYPE_LONG;
    ctx->flow5_t_bind[3].buffer = &ctx->flow_t_packet.srcaddr;
    ctx->flow5_t_bind[3].is_null = 0;
    ctx->flow5_t_bind[3].is_unsigned = 1;
    ctx->flow5_t_bind[3].length = 0;

    /* dstaddr */
    ctx->flow5_t_bind[4].buffer_type = MYSQL_TYPE_LONG;
    ctx->flow5_t_bind[4].buffer = &ctx->flow_t_packet.dstaddr;
    ctx->flow5_t_bind[4].is_null = 0;
    ctx->flow5_t_bind[4].is_unsigned = 1;
    ctx->flow5_t_bind[4].length = 0;

    /* nexthop */
    ctx->flow5_t_bind[5].buffer_type = MYSQL_TYPE_LONG;
    ctx->flow5_t_bind[5].buffer = &ctx->flow_t_packet.nexthop;
    ctx->flow5_t_bind[5].is_null = 0;
    ctx->flow5_t_bind[5].is_unsigned = 1;
    ctx->flow5_t_bind[5].length = 0;

    /* input_if */
    ctx->flow5_t_bind[6].buffer_type = MYSQL_TYPE_SHORT;
    ctx->flow5_t_bind[6].buffer = &ctx->flow_t_packet.input_if;
    ctx->flow5_t_bind[6].is_null = 0;
    ctx->flow5_t_bind[6].is_unsigned = 1;
    ctx->flow5_t_bind[6].length = 0;

    /* output_if */
    ctx->flow5_t_bind[7].buffer_type = MYSQL_TYPE_SHORT;
    ctx->flow5_t_bind[7].buffer = &ctx->flow_t_packet.output_if;
    ctx->flow5_t_bind[7].is_null = 0;
    ctx->flow5_t_bind[7].is_unsigned = 1;
    ctx->flow5_t_bind[7].length = 0;

    /* dpkts */
    ctx->flow5_t_bind[8].buffer_type = MYSQL_TYPE_LONG;
    ctx->flow5_t_bind[8].buffer = &ctx->flow_t_packet.dpkts;
    ctx->flow5_t_bind[8].is_null = 0;
    ctx->flow5_t_bind[8].is_unsigned = 1;
    ctx->flow5_t_bind[8].length = 0;

    /* doctets */
    ctx->flow5_t_bind[9].buffer_type = MYSQL_TYPE_LONG;
    ctx->flow5_t_bind[9].buffer = &ctx->flow_t_packet.doctets;
    ctx->flow5_t_bind[9].is_null = 0;
    ctx->flow5_t_bind[9].is_unsigned = 1;
    ctx->flow5_t_bind[9].length = 0;

    /* fuptime */
    ctx->flow5_t_bind[10].buffer_type = MYSQL_TYPE_LONG;
    ctx->flow5_t_bind[10].buffer = &ctx->flow_t_packet.fuptime;
    ctx->flow5_t_bind[10].is_null = 0;
    ctx->flow5_t_bind[10].is_unsigned = 1;
    ctx->flow5_t_bind[10].length = 0;

    /* luptime */
    ctx->flow5_t_bind[11].buffer_type = MYSQL_TYPE_LONG;
    ctx->flow5_t_bind[11].buffer = &ctx->flow_t_packet.luptime;
    ctx->flow5_t_bind[11].is_null = 0;
    ctx->flow5_t_bind[11].is_unsigned = 1;
    ctx->flow5_t_bind[11].length = 0;

    /* srcport */
    ctx->flow5_t_bind[12].buffer_type = MYSQL_TYPE_SHORT;
    ctx->flow5_t_bind[12].buffer = &ctx->flow_t_packet.srcport;
    ctx->flow5_t_bind[12].is_null = 0;
    ctx->flow5_t_bind[12].is_unsigned = 1;
    ctx->flow5_t_bind[12].length = 0;

    /* dstport */
    ctx->flow5_t_bind[13].buffer_type = MYSQL_TYPE_SHORT;
    ctx->flow5_t_bind[13].buffer = &ctx->flow_t_packet.dstport;
    ctx->flow5_t_bind[13].is_null = 0;
    ctx->flow5_t_bind[13].is_unsigned = 1;
    ctx->flow5_t_bind[13].length = 0;

    /* resv1 */
    ctx->flow5_t_bind[14].buffer_type = MYSQL_TYPE_SHORT;
    ctx->flow5_t_bind[14].buffer = &ctx->flow_t_packet.resv1;
    ctx->flow5_t_bind[14].is_null = 0;
    ctx->flow5_t_bind[14].is_unsigned = 1;
    ctx->flow5_t_bind[14].length = 0;

    /* flags */
    ctx->flow5_t_bind[15].buffer_type = MYSQL_TYPE_TINY;
    ctx->flow5_t_bind[15].buffer = &ctx->flow_t_packet.flags;
    ctx->flow5_t_bind[15].is_null = 0;
    ctx->flow5_t_bind[15].is_unsigned = 1;
    ctx->flow5_t_bind[15].length = 0;

    /* prot */
    ctx->flow5_t_bind[16].buffer_type = MYSQL_TYPE_TINY;
    ctx->flow5_t_bind[16].buffer = &ctx->flow_t_packet.prot;
    ctx->flow5_t_bind[16].is_null = 0;
    ctx->flow5_t_bind[16].is_unsigned = 1;
    ctx->flow5_t_bind[16].length = 0;

    /* tos */
    ctx->flow5_t_bind[17].buffer_type = MYSQL_TYPE_TINY;
    ctx->flow5_t_bind[17].buffer = &ctx->flow_t_packet.tos;
    ctx->flow5_t_bind[17].is_null = 0;
    ctx->flow5_t_bind[17].is_unsigned = 1;
    ctx->flow5_t_bind[17].length = 0;

    /* srcmask */
    ctx->flow5_t_bind[18].buffer_type = MYSQL_TYPE_TINY;
    ctx->flow5_t_bind[18].buffer = &ctx->flow_t_packet.srcmask;
    ctx->flow5_t_bind[18].is_null = 0;
    ctx->flow5_t_bind[18].is_unsigned = 1;
    ctx->flow5_t_bind[18].length = 0;

    /* dstmask */
    ctx->flow5_t_bind[19].buffer_type = MYSQL_TYPE_TINY;
    ctx->flow5_t_bind[19].buffer = &ctx->flow_t_packet.dstmask;
    ctx->flow5_t_bind[19].is_null = 0;
    ctx->flow5_t_bind[19].is_unsigned = 1;
    ctx->flow5_t_bind[19].length = 0;

    /* netflow v7 */
    /* address */
    ctx->flow7_t_bind[0].buffer_type = MYSQL_TYPE_LONG;
    ctx->flow7_t_bind[0].buffer = &ctx->flow_t_packet.address;
    ctx->flow7_t_bind[0].is_null = 0;
    ctx->flow7_t_bind[0].is_unsigned = 1;
    ctx->flow7_t_bind[0].length = 0;

    /* version */
    ctx->flow7_t_bind[1].buffer_type = MYSQL_TYPE_SHORT;
    ctx->flow7_t_bind[1].buffer = &ctx->flow_t_packet.version;
    ctx->flow7_t_bind[1].is_null = 0;
    ctx->flow7_t_bind[1].is_unsigned = 1;
    ctx->flow7_t_bind[1].length = 0;

    /* number */
    ctx->flow7_t_bind[2].buffer_type = MYSQL_TYPE_LONG;
    ctx->flow7_t_bind[2].buffer = &ctx->flow_t_packet.number;
    ctx->flow7_t_bind[2].is_null = 0;
    ctx->flow7_t_bind[2].is_unsigned = 1;
    ctx->flow7_t_bind[2].length = 0;

    /* srcaddr */
    ctx->flow7_t_bind[3].buffer_type = MYSQL_TYPE_LONG;
    ctx->flow7_t_bind[3].buffer = &ctx->flow_t_packet.srcaddr;
    ctx->flow7_t_bind[3].is_null = 0;
    ctx->flow7_t_bind[3].is_unsigned = 1;
    ctx->flow7_t_bind[3].length = 0;

    /* dstaddr */
    ctx->flow7_t_bind[4].buffer_type = MYSQL_TYPE_LONG;
    ctx->flow7_t_bind[4].buffer = &ctx->flow_t_packet.dstaddr;
    ctx->flow7_t_bind[4].is_null = 0;
    ctx->flow7_t_bind[4].is_unsigned = 1;
    ctx->flow7_t_bind[4].length = 0;

    /* nexthop */
    ctx->flow7_t_bind[5].buffer_type = MYSQL_TYPE_LONG;
    ctx->flow7_t_bind[5].buffer = &ctx->flow_t_packet.nexthop;
    ctx->flow7_t_bind[5].is_null = 0;
    ctx->flow7_t_bind[5].is_unsigned = 1;
    ctx->flow7_t_bind[5].length = 0;

    /* input_if */
    ctx->flow7_t_bind[6].buffer_type = MYSQL_TYPE_SHORT;
    ctx->flow7_t_bind[6].buffer = &ctx->flow_t_packet.input_if;
    ctx->flow7_t_bind[6].is_null = 0;
    ctx->flow7_t_bind[6].is_unsigned = 1;
    ctx->flow7_t_bind[6].length = 0;

    /* output_if */
    ctx->flow7_t_bind[7].buffer_type = MYSQL_TYPE_SHORT;
    ctx->flow7_t_bind[7].buffer = &ctx->flow_t_packet.output_if;
    ctx->flow7_t_bind[7].is_null = 0;
    ctx->flow7_t_bind[7].is_unsigned = 1;
    ctx->flow7_t_bind[7].length = 0;

    /* dpkts */
    ctx->flow7_t_bind[8].buffer_type = MYSQL_TYPE_LONG;
    ctx->flow7_t_bind[8].buffer = &ctx->flow_t_packet.dpkts;
    ctx->flow7_t_bind[8].is_null = 0;
    ctx->flow7_t_bind[8].is_unsigned = 1;
    ctx->flow7_t_bind[8].length = 0;

    /* doctets */
    ctx->flow7_t_bind[9].buffer_type = MYSQL_TYPE_LONG;
    ctx->flow7_t_bind[9].buffer = &ctx->flow_t_packet.doctets;
    ctx->flow7_t_bind[9].is_null = 0;
    ctx->flow7_t_bind[9].is_unsigned = 1;
    ctx->flow7_t_bind[9].length = 0;

    /* fuptime */
    ctx->flow7_t_bind[10].buffer_type = MYSQL_TYPE_LONG;
    ctx->flow7_t_bind[10].buffer = &ctx->flow_t_packet.fuptime;
    ctx->flow7_t_bind[10].is_null = 0;
    ctx->flow7_t_bind[10].is_unsigned = 1;
    ctx->flow7_t_bind[10].length = 0;

    /* luptime */
    ctx->flow7_t_bind[11].buffer_type = MYSQL_TYPE_LONG;
    ctx->flow7_t_bind[11].buffer = &ctx->flow_t_packet.luptime;
    ctx->flow7_t_bind[11].is_null = 0;
    ctx->flow7_t_bind[11].is_unsigned = 1;
    ctx->flow7_t_bind[11].length = 0;

    /* srcport */
    ctx->flow7_t_bind[12].buffer_type = MYSQL_TYPE_SHORT;
    ctx->flow7_t_bind[12].buffer = &ctx->flow_t_packet.srcport;
    ctx->flow7_t_bind[12].is_null = 0;
    ctx->flow7_t_bind[12].is_unsigned = 1;
    ctx->flow7_t_bind[12].length = 0;

    /* dstport */
    ctx->flow7_t_bind[13].buffer_type = MYSQL_TYPE_SHORT;
    ctx->flow7_t_bind[13].buffer = &ctx->flow_t_packet.dstport;
    ctx->flow7_t_bind[13].is_null = 0;
    ctx->flow7_t_bind[13].is_unsigned = 1;
    ctx->flow7_t_bind[13].length = 0;

    /* mode */
    ctx->flow7_t_bind[14].buffer_type = MYSQL_TYPE_TINY;
    ctx->flow7_t_bind[14].buffer = &ctx->flow_t_packet.mode;
    ctx->flow7_t_bind[14].is_null = 0;
    ctx->flow7_t_bind[14].is_unsigned = 1;
    ctx->flow7_t_bind[14].length = 0;

    /* flags */
    ctx->flow7_t_bind[15].buffer_type = MYSQL_TYPE_TINY;
    ctx->flow7_t_bind[15].buffer = &ctx->flow_t_packet.flags;
    ctx->flow7_t_bind[15].is_null = 0;
    ctx->flow7_t_bind[15].is_unsigned = 1;
    ctx->flow7_t_bind[15].length = 0;

    /* prot */
    ctx->flow7_t_bind[16].buffer_type = MYSQL_TYPE_TINY;
    ctx->flow7_t_bind[16].buffer = &ctx->flow_t_packet.prot;
    ctx->flow7_t_bind[16].is_null = 0;
    ctx->flow7_t_bind[16].is_unsigned = 1;
    ctx->flow7_t_bind[16].length = 0;

    /* tos */
    ctx->flow7_t_bind[17].buffer_type = MYSQL_TYPE_TINY;
    ctx->flow7_t_bind[17].buffer = &ctx->flow_t_packet.tos;
    ctx->flow7_t_bind[17].is_null = 0;
    ctx->flow7_t_bind[17].is_unsigned = 1;
    ctx->flow7_t_bind[17].length = 0;

    /* src_as */
    ctx->flow7_t_bind[18].buffer_type = MYSQL_TYPE_SHORT;
    ctx->flow7_t_bind[18].buffer = &ctx->flow_t_packet.src_as;
    ctx->flow7_t_bind[18].is_null = 0;
    ctx->flow7_t_bind[18].is_unsigned = 1;
    ctx->flow7_t_bind[18].length = 0;

    /* dst_as */
    ctx->flow7_t_bind[19].buffer_type = MYSQL_TYPE_SHORT;
    ctx->flow7_t_bind[19].buffer = &ctx->flow_t_packet.dst_as;
    ctx->flow7_t_bind[19].is_null = 0;
    ctx->flow7_t_bind[19].is_unsigned = 1;
    ctx->flow7_t_bind[19].length = 0;

    /* srcmask */
    ctx->flow7_t_bind[20].buffer_type = MYSQL_TYPE_TINY;
    ctx->flow7_t_bind[20].buffer = &ctx->flow_t_packet.srcmask;
    ctx->flow7_t_bind[20].is_null = 0;
    ctx->flow7_t_bind[20].is_unsigned = 1;
    ctx->flow7_t_bind[20].length = 0;

    /* dstmask */
    ctx->flow7_t_bind[21].buffer_type = MYSQL_TYPE_TINY;
    ctx->flow7_t_bind[21].buffer = &ctx->flow_t_packet.dstmask;
    ctx->flow7_t_bind[21].is_null = 0;
    ctx->flow7_t_bind[21].is_unsigned = 1;
    ctx->flow7_t_bind[21].length = 0;

    /* resv2 */
    ctx->flow7_t_bind[22].buffer_type = MYSQL_TYPE_LONG;
    ctx->flow7_t_bind[22].buffer = &ctx->flow_t_packet.resv2;
    ctx->flow7_t_bind[22].is_null = 0;
    ctx->flow7_t_bind[22].is_unsigned = 1;
    ctx->flow7_t_bind[22].length = 0;

    /* router */
    ctx->flow7_t_bind[23].buffer_type = MYSQL_TYPE_LONG;
    ctx->flow7_t_bind[23].buffer = &ctx->flow_t_packet.router;
    ctx->flow7_t_bind[23].is_null = 0;
    ctx->flow7_t_bind[23].is_unsigned = 1;
    ctx->flow7_t_bind[23].length = 0;

    /* Bind the buffers */
    if (mysql_stmt_bind_param(ctx->flow1hdr_t_stmt, ctx->flow1hdr_t_bind))
    {
        syslog(LOG_ERR, "mysql_stmt_bind_param(flow1hdr_t) failed");
        syslog(LOG_ERR, mysql_stmt_error(ctx->flow1hdr_t_stmt));
        return;
    }

    if (mysql_stmt_bind_param(ctx->flow5hdr_t_stmt, ctx->flow5hdr_t_bind))
    {
        syslog(LOG_ERR, "mysql_stmt_bind_param(flow5hdr_t) failed");
        syslog(LOG_ERR, mysql_stmt_error(ctx->flow5hdr_t_stmt));
        return;
    }

    if (mysql_stmt_bind_param(ctx->flow7hdr_t_stmt, ctx->flow7hdr_t_bind))
    {
        syslog(LOG_ERR, "mysql_stmt_bind_param(flow7hdr_t) failed");
        syslog(LOG_ERR, mysql_stmt_error(ctx->flow7hdr_t_stmt));
        return;
    }

    if (mysql_stmt_bind_param(ctx->flow1_t_stmt, ctx->flow1_t_bind))
    {
        syslog(LOG_ERR, "mysql_stmt_bind_param(flow1_t) failed");
        syslog(LOG_ERR, mysql_stmt_error(ctx->flow1_t_stmt));
        return;
    }

    if (mysql_stmt_bind_param(ctx->flow5_t_stmt, ctx->flow5_t_bind))
    {
        syslog(LOG_ERR, "mysql_stmt_bind_param(flow5_t) failed");
        syslog(LOG_ERR, mysql_stmt_error(ctx->flow5_t_stmt));
        return;
    }

    if (mysql_stmt_bind_param(ctx->flow7_t_stmt, ctx->flow7_t_bind))
    {
        syslog(LOG_ERR, "mysql_stmt_bind_param(flow7_t) failed");
        syslog(LOG_ERR, mysql_stmt_error(ctx->flow7_t_stmt));
        return;
    }

    ctx->connected = 1;
}

/* begins transaction */
void begin(void *dbparam)
{
    thread_context *ctx = (thread_context *) dbparam;
    if (!ctx->trans_enabled) return;

    mysql_query(&ctx->conn, "START TRANSACTION");
}

/* ends transaction */
void commit(void *dbparam)
{
    thread_context *ctx = (thread_context *) dbparam;
    if (!ctx->trans_enabled) return;

    mysql_commit(&ctx->conn);
}

/* closes connection to database */
void _close(void *dbparam)
{
    thread_context *ctx = (thread_context *) dbparam;
    if (!ctx->connected) return;

    commit(dbparam);
    mysql_stmt_close(ctx->flow1hdr_t_stmt);
    mysql_stmt_close(ctx->flow5hdr_t_stmt);
    mysql_stmt_close(ctx->flow7hdr_t_stmt);
    mysql_stmt_close(ctx->flow1_t_stmt);
    mysql_stmt_close(ctx->flow5_t_stmt);
    mysql_stmt_close(ctx->flow7_t_stmt);
    mysql_close(&ctx->conn);
    mysql_thread_end();

    syslog(LOG_DEBUG, "Connection to database closed");
    free(ctx);
}

/* inserts netflow packet's header into database */
void inserthdr(uint32_t address, void *flowhdr, void *dbparam)
{
    struct in_addr  in;
    char            str[256];
    char            addr[128];
    short           version = ntohs(*((short *)flowhdr));
    struct tm       *stm;
    time_t          ttm;
    thread_context *ctx = (thread_context *) dbparam;

    if (!ctx->connected) {
        _connect(ctx->config, dbparam);
        if (!ctx->connected) 
            return;
    }

    if (!flowhdr) {
        syslog(LOG_ERR, "NULL packet received");
        return;
    }

    in.s_addr = address;
    inet_ntop(AF_INET, &in, addr, sizeof(addr));

    ttm = time(NULL);
    stm = localtime(&ttm);

    switch (version) {
    case 1: {           /* netflow v1 */
        struct flowhdr1_t   *hdr = (struct flowhdr1_t *)flowhdr;

        ctx->flowhdr_t_packet.address = address;
        ctx->flowhdr_t_packet.version = version;
        ctx->flowhdr_t_packet.count = ntohs(hdr->count);
        ctx->flowhdr_t_packet.uptime = (unsigned long)ntohl(hdr->uptime);
        ctx->flowhdr_t_packet.secs = (unsigned long)ntohl(hdr->secs);
        ctx->flowhdr_t_packet.nsecs = (unsigned long)ntohl(hdr->nsecs);

        /* Execute the flow1hdr_t_stmt statement */
        if (mysql_stmt_execute(ctx->flow1hdr_t_stmt))
        {
            syslog(LOG_ERR, "mysql_stmt_execute(flow1hdr_t) failed");
            syslog(LOG_ERR, mysql_stmt_error(ctx->flow1hdr_t_stmt));
            return;
        }
        break;
    }
    case 5: {           /* netflow v5 */
        struct flowhdr5_t   *hdr = (struct flowhdr5_t *)flowhdr;

        ctx->flowhdr_t_packet.address = address;
        ctx->flowhdr_t_packet.version = version;
        ctx->flowhdr_t_packet.count = ntohs(hdr->count);
        ctx->flowhdr_t_packet.uptime = (unsigned long)ntohl(hdr->uptime);
        ctx->flowhdr_t_packet.secs = (unsigned long)ntohl(hdr->secs);
        ctx->flowhdr_t_packet.nsecs = (unsigned long)ntohl(hdr->nsecs);
        ctx->flowhdr_t_packet.flowseq = (unsigned long)ntohl(hdr->flowseq);
        ctx->flowhdr_t_packet.engine = hdr->engine;
        ctx->flowhdr_t_packet.eng_id = hdr->eng_id;
        ctx->flowhdr_t_packet.sampling = hdr->sampling;

        /* Execute the flow5hdr_t_stmt statement */
        if (mysql_stmt_execute(ctx->flow5hdr_t_stmt))
        {
            syslog(LOG_ERR, "mysql_stmt_execute(flow5hdr_t) failed");
            syslog(LOG_ERR, mysql_stmt_error(ctx->flow5hdr_t_stmt));
            return;
        }
        break;
    }
	case 7: {           /* netflow v7 */
        struct flowhdr7_t   *hdr = (struct flowhdr7_t *)flowhdr;

        ctx->flowhdr_t_packet.address = address;
        ctx->flowhdr_t_packet.version = version;
        ctx->flowhdr_t_packet.count = ntohs(hdr->count);
        ctx->flowhdr_t_packet.uptime = (unsigned long)ntohl(hdr->uptime);
        ctx->flowhdr_t_packet.secs = (unsigned long)ntohl(hdr->secs);
        ctx->flowhdr_t_packet.nsecs = (unsigned long)ntohl(hdr->nsecs);
        ctx->flowhdr_t_packet.flowseq = (unsigned long)ntohl(hdr->flowseq);
        ctx->flowhdr_t_packet.resv = (unsigned long)ntohl(hdr->resv);

        /* Execute the flow7hdr_t_stmt statement */
        if (mysql_stmt_execute(ctx->flow7hdr_t_stmt))
        {
            syslog(LOG_ERR, "mysql_stmt_execute(flow7hdr_t) failed");
            syslog(LOG_ERR, mysql_stmt_error(ctx->flow7hdr_t_stmt));
            return;
        }
    }
    default:
        syslog(LOG_ERR, "Unknown version of netflow packet");
        return;
    }
}

/* inserts flow into database */
void insertflow(uint32_t address, void *flowhdr, void *flow, int num, void *dbparam)
{
    struct in_addr  in;
    char            addr[128];
    char            str[512];
    short           version = ntohs(*((short *)flowhdr));
    struct tm       *stm;
    time_t          ttm;
    unsigned int    uptime;
    unsigned int    luptime;
    struct flowhdr5_t   *hdr = (struct flowhdr5_t *)flowhdr;
    struct flow1_t      *fl1 = (struct flow1_t *)flow;
    struct flow5_t      *fl5 = (struct flow5_t *)flow;
    struct flow7_t      *fl7 = (struct flow7_t *)flow;

    thread_context *ctx = (thread_context *) dbparam;

    if (!ctx->connected) {
        _connect(ctx->config, dbparam);
        if (!ctx->connected) 
            return;
    }

    if (!flowhdr || !flow) {
        syslog(LOG_ERR, "NULL packet received");
        return;
    }

    in.s_addr = address;
    inet_ntop(AF_INET, &in, addr, sizeof(addr));

    if (ctx->ltime) 
        ttm = time(NULL);
    else {
        uptime = htonl((unsigned int)hdr->uptime);
        luptime = htonl((unsigned int)fl5->luptime);
        if (uptime > luptime) {
            ttm = htonl(hdr->secs) - (uptime - luptime)/1000;
        } else {
            ttm = htonl(hdr->secs) + (luptime - uptime)/1000;
        }
    }
    stm = localtime(&ttm);

    switch (version) {
    case 1:             /* netflow v1 */
        ctx->flow_t_packet.address = address;
        ctx->flow_t_packet.version = version;
        ctx->flow_t_packet.number = num;
        ctx->flow_t_packet.srcaddr = htonl((fl1->srcaddr[0] << 24) + 
                                            (fl1->srcaddr[1] << 16) +
                                            (fl1->srcaddr[2] << 8) + 
                                            fl1->srcaddr[3]);
        ctx->flow_t_packet.dstaddr = htonl((fl1->dstaddr[0] << 24) + 
                                            (fl1->dstaddr[1] << 16) +
                                            (fl1->dstaddr[2] << 8) + 
                                            fl1->dstaddr[3]);
        ctx->flow_t_packet.nexthop = htonl((fl1->nexthop[0] << 24) + 
                                            (fl1->nexthop[1] << 16) +
                                            (fl1->nexthop[2] << 8) + 
                                            fl1->nexthop[3]);
        ctx->flow_t_packet.input_if = htons(fl1->input_if);
        ctx->flow_t_packet.output_if = htons(fl1->output_if),
        ctx->flow_t_packet.doctets = (unsigned long) htonl(fl1->doctets);
        ctx->flow_t_packet.dpkts = (unsigned long) htonl(fl1->dpkts);
        ctx->flow_t_packet.fuptime = (unsigned long) htonl(fl1->fuptime);
        ctx->flow_t_packet.luptime = (unsigned long) htonl(fl1->luptime);
        ctx->flow_t_packet.srcport = htons(fl1->srcport);
        ctx->flow_t_packet.dstport = htons(fl1->dstport);
        ctx->flow_t_packet.resv1 = htons(fl1->resv1);
        ctx->flow_t_packet.flags = fl1->flags;
        ctx->flow_t_packet.prot = fl1->prot;
        ctx->flow_t_packet.tos = fl1->tos;
        ctx->flow_t_packet.tcp_retx_cnt = fl1->tcp_retx_cnt;
        ctx->flow_t_packet.tcp_retx_sec = fl1->tcp_retx_sec;
        ctx->flow_t_packet.tcp_misseq_cnt = fl1->tcp_misseq_cnt;
        ctx->flow_t_packet.resv2 = (unsigned long)htonl(fl1->resv2);

        /* Execute the flow1_t_stmt statement */
        if (mysql_stmt_execute(ctx->flow1_t_stmt))
        {
            syslog(LOG_ERR, "mysql_stmt_execute(flow1_t) failed");
            syslog(LOG_ERR, mysql_stmt_error(ctx->flow1_t_stmt));
            return;
        }
        break;
    case 5:             /* netflow v5 */
        ctx->flow_t_packet.address = address ;
        ctx->flow_t_packet.version = version;
        ctx->flow_t_packet.number = (unsigned long)(num+htonl(hdr->flowseq)),
        ctx->flow_t_packet.srcaddr = htonl((fl5->srcaddr[0] << 24) + 
                                            (fl5->srcaddr[1] << 16) +
                                            (fl5->srcaddr[2] << 8) + 
                                            fl5->srcaddr[3]);
        ctx->flow_t_packet.dstaddr = htonl((fl5->dstaddr[0] << 24) + 
                                            (fl5->dstaddr[1] << 16) +
                                            (fl5->dstaddr[2] << 8) + 
                                            fl5->dstaddr[3]);
        ctx->flow_t_packet.nexthop = htonl((fl5->nexthop[0] << 24) + 
                                            (fl5->nexthop[1] << 16) +
                                            (fl5->nexthop[2] << 8) + 
                                            fl5->nexthop[3]);
        ctx->flow_t_packet.input_if = htons(fl5->input_if);
        ctx->flow_t_packet.output_if = htons(fl5->output_if);
        ctx->flow_t_packet.dpkts = (unsigned long) htonl(fl5->dpkts);
        ctx->flow_t_packet.doctets = (unsigned long) htonl(fl5->doctets);
        ctx->flow_t_packet.fuptime = (unsigned long) htonl(fl5->fuptime);
        ctx->flow_t_packet.luptime = (unsigned long) htonl(fl5->luptime);
        ctx->flow_t_packet.srcport = htons(fl5->srcport);
        ctx->flow_t_packet.dstport = htons(fl5->dstport);
        ctx->flow_t_packet.resv1 = (unsigned char)fl5->resv1;
        ctx->flow_t_packet.flags = fl5->flags;
        ctx->flow_t_packet.prot = fl5->prot;
        ctx->flow_t_packet.tos = fl5->tos;
        ctx->flow_t_packet.src_as = htons(fl5->src_as);
        ctx->flow_t_packet.dst_as = htons(fl5->dst_as);
        ctx->flow_t_packet.resv2 = htons(fl5->resv2);
        ctx->flow_t_packet.srcmask = (unsigned char) fl5->srcmask;
        ctx->flow_t_packet.dstmask = (unsigned char) fl5->dstmask;

        /* Execute the flow5_t_stmt statement */
        if (mysql_stmt_execute(ctx->flow5_t_stmt))
        {
            syslog(LOG_ERR, "mysql_stmt_execute(flow5_t) failed");
            syslog(LOG_ERR, mysql_stmt_error(ctx->flow5_t_stmt));
            return;
        }
        break;
    case 7:                 /* netflow v7 */
        ctx->flow_t_packet.address = address;
        ctx->flow_t_packet.version = version;
        ctx->flow_t_packet.number = (unsigned long)(num+htonl(hdr->flowseq));
        ctx->flow_t_packet.srcaddr = htonl((fl7->srcaddr[0] << 24) + 
                                            (fl7->srcaddr[1] << 16) +
                                            (fl7->srcaddr[2] << 8) + 
                                            fl7->srcaddr[3]);
        ctx->flow_t_packet.dstaddr = htonl((fl7->dstaddr[0] << 24) + 
                                            (fl7->dstaddr[1] << 16) +
                                            (fl7->dstaddr[2] << 8) + 
                                            fl7->dstaddr[3]);
        ctx->flow_t_packet.nexthop = htonl((fl7->nexthop[0] << 24) + 
                                            (fl7->nexthop[1] << 16) +
                                            (fl7->nexthop[2] << 8) + 
                                            fl7->nexthop[3]);
        ctx->flow_t_packet.input_if = htons(fl7->input_if);
        ctx->flow_t_packet.output_if = htons(fl7->output_if),
        ctx->flow_t_packet.doctets = (unsigned long) htonl(fl7->doctets);
        ctx->flow_t_packet.dpkts = (unsigned long) htonl(fl7->dpkts);
        ctx->flow_t_packet.fuptime = (unsigned long) htonl(fl7->fuptime);
        ctx->flow_t_packet.luptime = (unsigned long) htonl(fl7->luptime);
        ctx->flow_t_packet.srcport = htons(fl7->srcport);
        ctx->flow_t_packet.dstport = htons(fl7->dstport);
        ctx->flow_t_packet.mode = (unsigned char) fl7->mode;
        ctx->flow_t_packet.flags = (unsigned char) fl7->flags;
        ctx->flow_t_packet.prot = (unsigned char) fl7->prot;
        ctx->flow_t_packet.tos = (unsigned char) fl7->tos;
        ctx->flow_t_packet.src_as = htons(fl7->src_as);
        ctx->flow_t_packet.dst_as = htons(fl7->dst_as);
        ctx->flow_t_packet.srcmask = (unsigned char) fl7->srcmask;
        ctx->flow_t_packet.dstmask = (unsigned char) fl7->dstmask;
        ctx->flow_t_packet.resv2 = htons(fl7->resv2);
        ctx->flow_t_packet.router = htonl((fl7->router[0] << 24) + 
                                            (fl7->router[1] << 16) +
                                            (fl7->router[2] << 8) + 
                                            fl7->router[3]);

        /* Execute the flow7_t_stmt statement */
        if (mysql_stmt_execute(ctx->flow7_t_stmt))
        {
            syslog(LOG_ERR, "mysql_stmt_execute(flow7_t) failed");
            syslog(LOG_ERR, mysql_stmt_error(ctx->flow7_t_stmt));
            return;
        }
        break;
    default:
        syslog(LOG_ERR, "Unknown version of netflow packet");
        return;
    }
}

void insertpkt(uint32_t address, void *packet, void *dbparam)
{
    struct flowhdr1_t   *hdr1 = (struct flowhdr1_t *)packet;
    short               version = ntohs(*((short *)packet));
    int                 i;
    void                *flow;

    inserthdr(address, packet, dbparam);

    for (i = 0; i < htons(hdr1->count); i++) {
        switch (version) {
        case 1:
            flow = packet + sizeof(struct flowhdr1_t) +
                sizeof(struct flow1_t)*i;
            break;
        case 5:
            flow = packet + sizeof(struct flowhdr5_t) +
                sizeof(struct flow5_t)*i;
            break;
        case 7:
            flow = packet + sizeof(struct flowhdr7_t) +
                sizeof(struct flow7_t)*i;
            break;
        }
        insertflow(address, packet, flow, i, dbparam);
    }
}
