/**************************************************************************
 * Copyright (c) 2008-2009, Vyacheslav Yurkov <uvv.mail@gmail.com>        *
 * All rights reserved.                                                   *
 *                                                                        *
 * Redistribution and use in source and binary forms, with or without     *
 * modification, are permitted provided that the following conditions     *
 * are met:                                                               *
 *                                                                        *
 * 1. Redistributions of source code must retain the above copyright      *
 *    notice, this list of conditions and the following disclaimer.       *
 * 2. Redistributions in binary form must reproduce the above copyright   *
 *    notice, this list of conditions and the following disclaimer in the *
 *    documentation and/or other materials provided with the distribution.*
 * 3. Neither the name of the Andrew Klyachkin nor the names of his       *
 *    contributors may be used to endorse or promote products derived     *
 *    from this software without specific prior written permission.       *
 *                                                                        *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS    *
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT      *
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR  *
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT   *
 * AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, *
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       *
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  *
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  *
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    *
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  *
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   *
 **************************************************************************/

#include <sys/types.h>
#include <sys/socket.h>
#include <syslog.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <string.h>
#include <stdlib.h>
#include <mysql.h>

#include <libnnfc.h>

typedef struct
{
    uint32_t address;
    unsigned short version;
    unsigned long number;
    uint32_t srcaddr;
    uint32_t dstaddr;
    uint32_t nexthop;
    unsigned short input_if;
    unsigned short output_if;
    unsigned long doctets;
    unsigned long dpkts;
    unsigned long fuptime;
    unsigned long luptime;
    unsigned short srcport;
    unsigned short dstport;
    unsigned char resv1;
    char flags;
    char prot;
    char tos;
    char mode;
    unsigned short src_as;
    unsigned short dst_as;
    unsigned char srcmask;
    unsigned char dstmask;
    char tcp_retx_cnt;
    char tcp_retx_sec;
    char tcp_misseq_cnt;
    uint32_t router;
    unsigned short resv2;
    char time_id[20];
} flow_for_mysql;

typedef struct
{
    uint32_t address;
    unsigned short version;
    unsigned short count;
    uint32_t uptime;
    uint32_t secs;
    uint32_t nsecs;
    uint32_t flowseq;
    char engine;
    char eng_id;
    short sampling;
    uint32_t resv;
    char time_id[20];
} header_for_mysql;

typedef struct
{
    /*
     *  conn = mysql_init();
     *  mysql_real_connect(conn);
     *  mysql_real_query();
     *  mysql_close(conn);
     *  mysql_thread_end();
     */

    MYSQL    conn;
    int             connected;
    char            config[MAXPATHLEN];

    /*
     *  Parameters for this module:
     *  localtime - store local or remote time?
     *  dbhost - mysql server
     *  dbport - mysql server port
     *  dbuser - mysql user's name
     *  dbpass - mysql user's password
     *  dbname - mysql database
     *  dbheader - mysql table for netflow headers
     *  dbflows - mysql table for netflow packets
     */
    char    *dbhost;
    int     dbport;
    char    *dbuser;
    char    *dbpass;
    char    *dbname;
    char    *dbheader;
    char    *dbflows;
    int     ltime;
    /* insert queries */
    char flow1hdr_t_insert[1024];
    char flow5hdr_t_insert[1024];
    char flow7hdr_t_insert[1024];
    char flow1_t_insert[1024];
    char flow5_t_insert[1024];
    char flow7_t_insert[1024];
    /* statement handlers */
    MYSQL_STMT *flow1hdr_t_stmt;
    MYSQL_STMT *flow5hdr_t_stmt;
    MYSQL_STMT *flow7hdr_t_stmt;
    MYSQL_STMT *flow1_t_stmt;
    MYSQL_STMT *flow5_t_stmt;
    MYSQL_STMT *flow7_t_stmt;
    header_for_mysql flowhdr_t_packet;
    flow_for_mysql flow_t_packet;
    /* bind handlers */
    MYSQL_BIND flow1hdr_t_bind[6];
    MYSQL_BIND flow5hdr_t_bind[10];
    MYSQL_BIND flow7hdr_t_bind[8];
    MYSQL_BIND flow1_t_bind[22];
    MYSQL_BIND flow5_t_bind[20];
    MYSQL_BIND flow7_t_bind[24];
    int     trans_enabled;
} thread_context;

