/**************************************************************************
 * Copyright (c) 2004-2006, Andrew Klyachkin <andrewk@region73.ru>        *
 * All rights reserved.                                                   *
 *                                                                        *
 * Redistribution and use in source and binary forms, with or without     *
 * modification, are permitted provided that the following conditions     *
 * are met:                                                               *
 *                                                                        *
 * 1. Redistributions of source code must retain the above copyright      *
 *    notice, this list of conditions and the following disclaimer.       *
 * 2. Redistributions in binary form must reproduce the above copyright   *
 *    notice, this list of conditions and the following disclaimer in the *
 *    documentation and/or other materials provided with the distribution.*
 * 3. Neither the name of the Andrew Klyachkin nor the names of his       *
 *    contributors may be used to endorse or promote products derived     *
 *    from this software without specific prior written permission.       *
 *                                                                        *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS    *
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT      *
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR  *
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT   *
 * AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, *
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       *
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  *
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  *
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    *
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  *
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   *
 **************************************************************************/

#include <sys/types.h>
#include <sys/socket.h>
#include <syslog.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <libnnfc.h>

void * _create_thread_context()
{
}

/* establishes connection to database */
void _connect(char *fname, void *param)
{
    syslog(LOG_DEBUG, "Connection to null://null established");
}

/* closes connection to database */
void _close(void *param)
{
    syslog(LOG_DEBUG, "Connection to null://null closed");
}

/* inserts netflow packet's header into database */
void inserthdr(uint32_t address, void *flowhdr, void *param)
{
    struct in_addr  in;
    char            str[128];
    int             version = ntohs(*((short *)flowhdr));

    if (flowhdr == NULL) {
        syslog(LOG_ERR, "NULL packet received");
        return;
    }

    in.s_addr = address;
    inet_ntop(AF_INET, &in, str, sizeof(str));
    syslog(LOG_DEBUG, "Packet v%d from %s", version, str);

    switch (version) {
    case 1: {           /* netflow v1 */
        struct flowhdr1_t   *hdr = (struct flowhdr1_t *)flowhdr;
        syslog(LOG_DEBUG, "count = %d", ntohs(hdr->count));
        syslog(LOG_DEBUG, "uptime = %lu", (unsigned long)ntohl(hdr->uptime));
        syslog(LOG_DEBUG, "secs = %lu", (unsigned long)ntohl(hdr->secs));
        syslog(LOG_DEBUG, "nsecs = %lu", (unsigned long)ntohl(hdr->nsecs));
        break;
    }
	case 5: {           /* netflow v5 */
        struct flowhdr5_t   *hdr = (struct flowhdr5_t *)flowhdr;
        syslog(LOG_DEBUG, "count = %d", ntohs(hdr->count));
        syslog(LOG_DEBUG, "uptime = %lu", (unsigned long)ntohl(hdr->uptime));
        syslog(LOG_DEBUG, "secs = %lu", (unsigned long)ntohl(hdr->secs));
        syslog(LOG_DEBUG, "nsecs = %lu", (unsigned long)ntohl(hdr->nsecs));
        syslog(LOG_DEBUG, "flowseq = %lu", (unsigned long)ntohl(hdr->flowseq));
        syslog(LOG_DEBUG, "engine = %d", hdr->engine);
        syslog(LOG_DEBUG, "eng_id = %d", hdr->eng_id);
        syslog(LOG_DEBUG, "sampling = %d", ntohs(hdr->sampling));
        break;
    }
	case 7: {           /* netflow v7 */
        struct flowhdr7_t   *hdr = (struct flowhdr7_t *)flowhdr;
        syslog(LOG_DEBUG, "count = %d", ntohs(hdr->count));
        syslog(LOG_DEBUG, "uptime = %lu", (unsigned long)ntohl(hdr->uptime));
        syslog(LOG_DEBUG, "secs = %lu", (unsigned long)ntohl(hdr->secs));
        syslog(LOG_DEBUG, "nsecs = %lu", (unsigned long)ntohl(hdr->nsecs));
        syslog(LOG_DEBUG, "flowseq = %lu", (unsigned long)ntohl(hdr->flowseq));
        syslog(LOG_DEBUG, "resv = %lu", (unsigned long)ntohl(hdr->resv));
        break;
    }
    default:
        syslog(LOG_ERR, "Unknown version of netflow packet");
    }
}

/* inserts flow into database */
void insertflow(uint32_t address, void *flowhdr, void *flow, int num, void *param)
{
    struct in_addr  in;
    char            str[128];
    int             version = ntohs(*((short *)flowhdr));

    if (flowhdr == NULL || flow == NULL) {
        syslog(LOG_ERR, "NULL packet received");
        return;
    }

    in.s_addr = address;
    inet_ntop(AF_INET, &in, str, sizeof(str));

    switch (version) {
    case 1: {       /* netflow v1 */
        struct flow1_t  *fl = (struct flow1_t *)flow;
        syslog(LOG_DEBUG, "Flow v%d N%u from %s", version, num, str);
        syslog(LOG_DEBUG, "srcaddr = '%u.%u.%u.%u'",
            fl->srcaddr[0], fl->srcaddr[1], fl->srcaddr[2], fl->srcaddr[3]);
        syslog(LOG_DEBUG, "dstaddr = '%u.%u.%u.%u'",
            fl->dstaddr[0], fl->dstaddr[1], fl->dstaddr[2], fl->dstaddr[3]);
        syslog(LOG_DEBUG, "nexthop = '%u.%u.%u.%u'",
            fl->nexthop[0], fl->nexthop[1], fl->nexthop[2], fl->nexthop[3]);
        syslog(LOG_DEBUG, "input_if = %u", htons(fl->input_if));
        syslog(LOG_DEBUG, "output_if = %u", htons(fl->output_if));
        syslog(LOG_DEBUG, "dpkts = %lu", (unsigned long)htonl(fl->dpkts));
        syslog(LOG_DEBUG, "doctets = %lu", (unsigned long)htonl(fl->doctets));
        syslog(LOG_DEBUG, "fuptime = %lu", (unsigned long)htonl(fl->fuptime));
        syslog(LOG_DEBUG, "luptime = %lu", (unsigned long)htonl(fl->luptime));
        syslog(LOG_DEBUG, "srcport = %u", htons(fl->srcport));
        syslog(LOG_DEBUG, "dstport = %u", htons(fl->dstport));
        syslog(LOG_DEBUG, "resv1 = %u", htons(fl->resv1));
        syslog(LOG_DEBUG, "prot = %u", fl->prot);
        syslog(LOG_DEBUG, "tos = %u", fl->tos);
        syslog(LOG_DEBUG, "flags = %u", fl->flags);
        syslog(LOG_DEBUG, "tcp_retx_cnt = %u", fl->tcp_retx_cnt);
        syslog(LOG_DEBUG, "tcp_retx_sec = %u", fl->tcp_retx_sec);
        syslog(LOG_DEBUG, "tcp_misseq_cnt = %u", fl->tcp_misseq_cnt);
        syslog(LOG_DEBUG, "resv2 = %lu", (unsigned long)htonl(fl->resv2));
        break;
    }
	case 5: {           /* netflow v5 */
        struct flowhdr5_t   *hdr = (struct flowhdr5_t *)flowhdr;
        struct flow5_t      *fl = (struct flow5_t *)flow;
        syslog(LOG_DEBUG, "Flow v%d N%lu from %s", version, 
            (unsigned long)(num+htonl(hdr->flowseq)), str);
        syslog(LOG_DEBUG, "srcaddr = '%u.%u.%u.%u'",
            fl->srcaddr[0], fl->srcaddr[1], fl->srcaddr[2], fl->srcaddr[3]);
        syslog(LOG_DEBUG, "dstaddr = '%u.%u.%u.%u'",
            fl->dstaddr[0], fl->dstaddr[1], fl->dstaddr[2], fl->dstaddr[3]);
        syslog(LOG_DEBUG, "nexthop = '%u.%u.%u.%u'",
            fl->nexthop[0], fl->nexthop[1], fl->nexthop[2], fl->nexthop[3]);
        syslog(LOG_DEBUG, "input_if = %u", htons(fl->input_if));
        syslog(LOG_DEBUG, "output_if = %u", htons(fl->output_if));
        syslog(LOG_DEBUG, "dpkts = %lu", (unsigned long)htonl(fl->dpkts));
        syslog(LOG_DEBUG, "doctets = %lu", (unsigned long)htonl(fl->doctets));
        syslog(LOG_DEBUG, "fuptime = %lu", (unsigned long)htonl(fl->fuptime));
        syslog(LOG_DEBUG, "luptime = %lu", (unsigned long)htonl(fl->luptime));
        syslog(LOG_DEBUG, "srcport = %u", htons(fl->srcport));
        syslog(LOG_DEBUG, "dstport = %u", htons(fl->dstport));
        syslog(LOG_DEBUG, "resv1 = %u", fl->resv1);
        syslog(LOG_DEBUG, "flags = %u", fl->flags);
        syslog(LOG_DEBUG, "prot = %u", fl->prot);
        syslog(LOG_DEBUG, "tos = %u", fl->tos);
        syslog(LOG_DEBUG, "src_as = %u", htons(fl->src_as));
        syslog(LOG_DEBUG, "dst_as = %u", htons(fl->dst_as));
        syslog(LOG_DEBUG, "srcmask = %u", fl->srcmask);
        syslog(LOG_DEBUG, "dstmask = %u", fl->dstmask);
        syslog(LOG_DEBUG, "resv2 = %u", htons(fl->resv2));
        break;
    }
	case 7: {           /* netflow v7 */
        struct flowhdr7_t   *hdr = (struct flowhdr7_t *)flowhdr;
        struct flow7_t      *fl = (struct flow7_t *)flow;
        syslog(LOG_DEBUG, "Flow v%d N%lu from %s", version, 
            (unsigned long)(num+htonl(hdr->flowseq)), str);
        syslog(LOG_DEBUG, "srcaddr = %u.%u.%u.%u",
            fl->srcaddr[0], fl->srcaddr[1], fl->srcaddr[2], fl->srcaddr[3]);
        syslog(LOG_DEBUG, "dstaddr = %u.%u.%u.%u",
            fl->dstaddr[0], fl->dstaddr[1], fl->dstaddr[2], fl->dstaddr[3]);
        syslog(LOG_DEBUG, "nexthop = %u.%u.%u.%u",
            fl->nexthop[0], fl->nexthop[1], fl->nexthop[2], fl->nexthop[3]);
        syslog(LOG_DEBUG, "input_if = %u", ntohs(fl->input_if));
        syslog(LOG_DEBUG, "output_if = %u", ntohs(fl->output_if));
        syslog(LOG_DEBUG, "dpkts = %lu", (unsigned long)ntohl(fl->dpkts));
        syslog(LOG_DEBUG, "doctets = %lu", (unsigned long)ntohl(fl->doctets));
        syslog(LOG_DEBUG, "fuptime = %lu", (unsigned long)ntohl(fl->fuptime));
        syslog(LOG_DEBUG, "luptime = %lu", (unsigned long)ntohl(fl->luptime));
        syslog(LOG_DEBUG, "srcport = %u", htons(fl->srcport));
        syslog(LOG_DEBUG, "dstport = %u", htons(fl->dstport));
        syslog(LOG_DEBUG, "mode = %u", (unsigned char)fl->mode);
        syslog(LOG_DEBUG, "flags = %u", (unsigned char)fl->flags);
        syslog(LOG_DEBUG, "prot = %u", (unsigned char)fl->prot);
        syslog(LOG_DEBUG, "tos = %u", (unsigned char)fl->tos);
        syslog(LOG_DEBUG, "src_as = %u", htons(fl->src_as));
        syslog(LOG_DEBUG, "dst_as = %u", htons(fl->dst_as));
        syslog(LOG_DEBUG, "srcmask = %u", (unsigned char)fl->srcmask);
        syslog(LOG_DEBUG, "dstmask = %u", (unsigned char)fl->dstmask);
        syslog(LOG_DEBUG, "resv2 = %lu", (unsigned long)ntohl(fl->resv2));
        syslog(LOG_DEBUG, "router = %u.%u.%u.%u",
            fl->router[0], fl->router[1], fl->router[2], fl->router[3]);
        break;
    }
	default:
        syslog(LOG_ERR, "Unknown version of netflow packet");
    }
}

void insertpkt(uint32_t address, void *packet, void *param)
{
    struct flowhdr1_t   *hdr1 = (struct flowhdr1_t *)packet;
    short               version = ntohs(*((short *)packet));
    int                 i;
    void                *flow;

    inserthdr(address, packet, param);

    for (i = 0; i < htons(hdr1->count); i++) {
        switch (version) {
        case 1:
            flow = packet + sizeof(struct flowhdr1_t) +
                sizeof(struct flow1_t)*i;
            break;
        case 5:
            flow = packet + sizeof(struct flowhdr5_t) +
                sizeof(struct flow5_t)*i;
            break;
        case 7:
            flow = packet + sizeof(struct flowhdr7_t) +
                sizeof(struct flow7_t)*i;
            break;
        }
        insertflow(address, packet, flow, i, param);
    }
}

void begin(void *param)
{
    return;
}

void commit(void *param)
{
    return;
}
