/**************************************************************************
 * Copyright (c) 2008-2009, Vyacheslav Yurkov <uvv.mail@gmail.com>        *
 * All rights reserved.                                                   *
 *                                                                        *
 * Redistribution and use in source and binary forms, with or without     *
 * modification, are permitted provided that the following conditions     *
 * are met:                                                               *
 *                                                                        *
 * 1. Redistributions of source code must retain the above copyright      *
 *    notice, this list of conditions and the following disclaimer.       *
 * 2. Redistributions in binary form must reproduce the above copyright   *
 *    notice, this list of conditions and the following disclaimer in the *
 *    documentation and/or other materials provided with the distribution.*
 * 3. Neither the name of the Vyacheslav Yurkov nor the names of his      *
 *    contributors may be used to endorse or promote products derived     *
 *    from this software without specific prior written permission.       *
 *                                                                        *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS    *
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT      *
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR  *
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT   *
 * AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, *
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       *
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  *
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  *
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    *
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  *
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   *
 **************************************************************************/

#include "oracle.h"

static OCIEnv *envhp = NULL;
static pthread_mutex_t      mutex_init = PTHREAD_MUTEX_INITIALIZER;      /* init mutex */

void * _create_thread_context()
{
    thread_context *ctx = (thread_context*) malloc(sizeof(thread_context));
//    ctx->envhp = NULL;
    ctx->srvhp = NULL;
    ctx->svchp = NULL;
    ctx->authp = NULL;
    ctx->stmt_flow1_t = NULL;
    ctx->stmt_flow5_t = NULL;
    ctx->stmt_flow7_t = NULL;
    ctx->ltime = 1;
    ctx->full = 0;
    ctx->agg = 0;
    ctx->accounting = 0;
    ctx->dbuser = 0;
    ctx->dbpass = 0;
    ctx->dbname = 0;
    ctx->dbheader = 0;
    ctx->dbflows = 0;
 //   ctx->config[MAXPATHLEN];
    ctx->connected = 0;
    ctx->counting = 0;
    ctx->inc_net = 0;
    ctx->inc_addr = 0;
    ctx->inc_net_count = 0;
    ctx->inc_addr_count = 0;
    ctx->exc_net = 0;
    ctx->exc_addr = 0;
    ctx->exc_net_count = 0;
    ctx->exc_addr_count = 0;
    /*ctx->insert = (text *)"insert into :table_name (srcaddr, dstaddr, "
                          "doctets)"
                          "values (:srcaddr, :dstaddr, :doctets)";
    */
    return ctx;
}

void __attribute__ ((constructor)) my_init(void)
{
    static int     env_initialized = 0;
    sword errcode;

    /* Initialize the mode to be the threaded and object environment */
    pthread_mutex_lock(&mutex_init);
    if (!env_initialized) {
        if ((errcode = OCIEnvNlsCreate((OCIEnv **) &envhp, OCI_THREADED, (dvoid *)0,
                            NULL, NULL, NULL, 0, (dvoid *) 0, (ub2) 0, (ub2) 0))) {
            printf("%s\n", "FAILED: OCIEnvCreate()");
            if (errcode != 0) {
                (void) printf("OCIEnvCreate failed with errcode = %d.\n", errcode);
                exit(1);
            }
            /*    return OCI_ERROR;*/
        }
        else {
            printf("Environment created\n");
            env_initialized = 1;
        }
    }
    pthread_mutex_unlock(&mutex_init);
}

void __attribute__ ((destructor)) my_fini(void)
{
    pthread_mutex_lock(&mutex_init);
    if (envhp)
        (void) OCIHandleFree((dvoid *) envhp, (ub4) OCI_HTYPE_ENV);
    pthread_mutex_unlock(&mutex_init);
}

/* Print the error message */
void report_error(errhp)
OCIError *errhp;
{
    text  msgbuf[1024];
    sb4   errcode = 0;

    (void) OCIErrorGet((dvoid *) errhp, (ub4) 1, (text *) NULL, &errcode,
                       msgbuf, (ub4) sizeof(msgbuf), (ub4) OCI_HTYPE_ERROR);
    syslog(LOG_ERR, "ERROR CODE = %d\n", errcode);
    syslog(LOG_ERR, "%.*s\n", 1024, msgbuf);
    return;
}

void _close(void *dbparam)
{
    thread_context *ctx = (thread_context *) dbparam;
    printf("Pointer close dbparam = %p\n", dbparam);
    printf("try to close module\n");
    printf("Before close: ctx->dbuser = %s\n", ctx->dbuser);
    if (ctx->connected) {
    printf("try to close session\n");
        /* Logout and detach from the server */
        if (OCISessionEnd(ctx->svchp, ctx->errhp, ctx->authp, (ub4) 0)) {
            syslog(LOG_ERR, "FAILED: OCISessionEnd()\n");
            report_error(ctx->errhp);
        }

        syslog(LOG_ERR, "%s Logged off.\n", ctx->dbuser);

        if (OCIServerDetach(ctx->srvhp, ctx->errhp, (ub4) OCI_DEFAULT)) {
            syslog(LOG_ERR, "FAILED: OCISessionEnd()\n");
            report_error(ctx->errhp);
        }

        syslog(LOG_ERR, "Detached from server.\n");
    }
    
    /* Free the specified handles */
    syslog(LOG_ERR, "Freeing handles ...\n");

    if (ctx->srvhp)
        (void) OCIHandleFree((dvoid *) ctx->srvhp, (ub4) OCI_HTYPE_SERVER);
    if (ctx->svchp)
        (void) OCIHandleFree((dvoid *) ctx->svchp, (ub4) OCI_HTYPE_SVCCTX);
    if (ctx->errhp)
        (void) OCIHandleFree((dvoid *) ctx->errhp, (ub4) OCI_HTYPE_ERROR);
    if (ctx->authp)
        (void) OCIHandleFree((dvoid *) ctx->authp, (ub4) OCI_HTYPE_SESSION);
    if (ctx->stmt_flow1_t)
        (void) OCIHandleFree((dvoid *) ctx->stmt_flow1_t, (ub4) OCI_HTYPE_STMT);
    if (ctx->stmt_flow5_t)
        (void) OCIHandleFree((dvoid *) ctx->stmt_flow5_t, (ub4) OCI_HTYPE_STMT);
    if (ctx->stmt_flow7_t)
        (void) OCIHandleFree((dvoid *) ctx->stmt_flow7_t, (ub4) OCI_HTYPE_STMT);
    /*if (ctx->envhp)
        (void) OCIHandleFree((dvoid *) ctx->envhp, (ub4) OCI_HTYPE_ENV);
    */
    syslog(LOG_DEBUG, "Connection to database closed");

    /* Free the memory allocated for pointers */
    if (ctx->exc_addr) free(ctx->exc_addr);
    if (ctx->exc_net) free(ctx->exc_net);
    if (ctx->inc_addr) free(ctx->inc_addr);
    if (ctx->inc_net) free(ctx->inc_net);
    
    /* free thread_context structure */
    free(ctx);
}

/**
 * Parse a network specification. The argument specifies
 * a list of networks. The format is
 * <tt>[network/netmask;]*</tt> (no whitespace, must be terminated
 * with a semicolon). The network must be given in dotted-decimal
 * notation. The netmask can be given in CIDR notation (/16) or
 * in dotted-decimal (/255.255.0.0).
 * This function was taken from GNUnet project.
 * <p>
 * @param routeList a string specifying the forbidden networks
 * @return the converted list, NULL if the synatx is flawed
 * @author Christian Grothoff
 */
CIDRNetwork *
GNUNET_parse_ipv4_network_specification(const char *routeList)
{
    unsigned int count;
    unsigned int i;
    unsigned int j;
    unsigned int len;
    int cnt;
    unsigned int pos;
    unsigned int temps[8];
    int slash;
    CIDRNetwork *result;

    if (routeList == NULL)
        return NULL;
    len = strlen (routeList);
    if (len == 0)
        return NULL;
    count = 0;
    for (i = 0; i < len; i++)
        if (routeList[i] == ';')
            count++;
    result = (CIDRNetwork *)calloc(count + 1, sizeof (CIDRNetwork) * (count + 1));
    /* add termination */
    memset (result, 0, sizeof (CIDRNetwork) * (count + 1));
    i = 0; 
    pos = 0;
    while (i < count)
    {
        cnt = sscanf (&routeList[pos],
                      "%u.%u.%u.%u/%u.%u.%u.%u;",
                      &temps[0],
                      &temps[1],
                      &temps[2],
                      &temps[3], &temps[4], &temps[5], &temps[6], &temps[7]);
        if (cnt == 8)
        {
            for (j = 0; j < 8; j++)
                if (temps[j] > 0xFF)
                {
                    printf("Invalid format for IP: `%s'\n", &routeList[pos]);
                    free (result);
                    return NULL;
                }
            result[i].network.s_addr =
                htonl ((temps[0] << 24) + (temps[1] << 16) + (temps[2] << 8) +
                        temps[3]);
            result[i].netmask.s_addr =
                htonl ((temps[4] << 24) + (temps[5] << 16) + (temps[6] << 8) +
                        temps[7]);
            while (routeList[pos] != ';')
                pos++;
            pos++;
            i++;
            continue;
        }
        /* try second notation */
        cnt = sscanf (&routeList[pos],
                      "%u.%u.%u.%u/%u;",
                      &temps[0], &temps[1], &temps[2], &temps[3], &slash);
        if (cnt == 5)
        {
            for (j = 0; j < 4; j++)
                if (temps[j] > 0xFF)
                {
                    printf("Invalid format for IP: `%s'\n", &routeList[pos]);
                    free (result);
                    return NULL;
                }
            result[i].network.s_addr =
                htonl ((temps[0] << 24) + (temps[1] << 16) + (temps[2] << 8) +
                        temps[3]);
            if ((slash <= 32) && (slash >= 0))
            {
                result[i].netmask.s_addr = 0;
                while (slash > 0)
                {
                    result[i].netmask.s_addr = (result[i].netmask.s_addr >> 1) + 0x80000000;
                    slash--;
                }
                result[i].netmask.s_addr = htonl (result[i].netmask.s_addr);
                while (routeList[pos] != ';')
                    pos++;
                pos++;
                i++;
                continue;
            }
            else
            {
                printf("Invalid network notation ('/%d' is not legal in IPv4 CIDR).", slash);
                free (result);
                return NULL;      /* error */
            }
        }
        /* try third notation */
        slash = 32;
        cnt = sscanf (&routeList[pos],
                      "%u.%u.%u.%u;",
                      &temps[0], &temps[1], &temps[2], &temps[3]);
        if (cnt == 4)
        {
            for (j = 0; j < 4; j++)
                if (temps[j] > 0xFF)
                {
                    printf("Invalid format for IP: `%s'\n", &routeList[pos]);
                    free (result);
                    return NULL;
                }
            result[i].network.s_addr =
                htonl ((temps[0] << 24) + (temps[1] << 16) + (temps[2] << 8) +
                        temps[3]);
            result[i].netmask.s_addr = 0;
            while (slash > 0)
            {
                result[i].netmask.s_addr = (result[i].netmask.s_addr >> 1) + 0x80000000;
                slash--;
            }
            result[i].netmask.s_addr = htonl (result[i].netmask.s_addr);
            while (routeList[pos] != ';')
                pos++;
            pos++;
            i++;
            continue;
        }
        printf("Invalid format for IP: `%s'\n", &routeList[pos]);
        free (result);
        return NULL;              /* error */
    }
    if (pos < strlen (routeList))
    {
        printf("Invalid format for IP: `%s'\n", &routeList[pos]);
        free (result);
        return NULL;              /* oops */
    }
    return result;                /* ok */
}

void _connect(char *fname, void *dbparam)
{
    struct command  *commands = 0;
    int             n = 0, i;
    struct in_addr  addr;
    struct in_addr  sub;
    struct hostent  *he = 0;
    struct value    *v = 0;
	char            *route_list = 0;
    sword       status;

    thread_context *ctx = (thread_context *) dbparam;
    
    if (!fname) {
        printf("return\n");
        return;
    }

    if (!ctx->config[0]) {
        memset(ctx->config, 0, MAXPATHLEN);
        strncpy(ctx->config, fname, MAXPATHLEN);
    }

    if (readcfg(ctx->config, &commands)) {
        syslog(LOG_ERR, "oracle: parse error!");
        return;
    }

    CFG_GETI(ctx->full, "full", commands);
    CFG_GETI(ctx->agg, "agg", commands);
    CFG_GETI(ctx->accounting, "accounting", commands);
 
    CFG_GETS(ctx->dbuser, "dbuser", commands);
    CFG_GETS(ctx->dbpass, "dbpass", commands);
    CFG_GETS(ctx->dbname, "dbname", commands);
    CFG_GETS(ctx->dbheader, "dbheader", commands);
    CFG_GETS(ctx->dbflows, "dbflows", commands);

    v = getallvalues("included_addresses", commands);
    ctx->inc_addr_count = 0;
    while (v) {
        if (inet_pton(AF_INET, v->value, &addr) <= 0) {
            if ((he = gethostbyname(v->value)) == 0) {
                /* TODO: The function is obsolete. 
                 * Applications should use getaddrinfo(3)
                 * and getnameinfo(3) instead */
                syslog(LOG_ERR, "oracle: invalid address "
                    "in parameter included_addresses");
                return;
            }
            addr = *(struct in_addr *)(he->h_addr);
        }
        /* Allocate memory for the list of excluded addresses */
        if (!ctx->inc_addr) {
            ctx->inc_addr = (uint32_t *)calloc(1, sizeof(uint32_t));
        } else {
            ctx->inc_addr = (uint32_t *)realloc(ctx->inc_addr, sizeof(uint32_t)*(ctx->inc_addr_count+1));
        }
        ctx->inc_addr[ctx->inc_addr_count] = addr.s_addr;
        ctx->inc_addr_count++;
        v = v->next;
    }
    
    v = getallvalues("excluded_addresses", commands);
    ctx->exc_addr_count = 0;
    while (v) {
        if (inet_pton(AF_INET, v->value, &addr) <= 0) {
            if ((he = gethostbyname(v->value)) == 0) {
                /* TODO: The function is obsolete. 
                 * Applications should use getaddrinfo(3)
                 * and getnameinfo(3) instead */
                syslog(LOG_ERR, "oracle: invalid address "
                    "in parameter excluded_addresses");
                return;
            }
            addr = *(struct in_addr *)(he->h_addr);
        }
        /* Allocate memory for the list of excluded addresses */
        if (!ctx->exc_addr) {
            ctx->exc_addr = (uint32_t *)calloc(1, sizeof(uint32_t));
        } else {
            ctx->exc_addr = (uint32_t *)realloc(ctx->exc_addr, sizeof(uint32_t)*(ctx->exc_addr_count+1));
        }
        ctx->exc_addr[ctx->exc_addr_count] = addr.s_addr;
        ctx->exc_addr_count++;
        v = v->next;
    }

    /* Get included networks */
    v = getallvalues("included_networks", commands);
    ctx->inc_net_count = 0;
    while (v) {
        /* Allocate memory for the list of excluded networks */
        if (!route_list) {
            route_list = (char *)calloc(1, 2*INET_ADDRSTRLEN);
        } else {
            route_list = (char *)realloc(route_list, 2*INET_ADDRSTRLEN*(ctx->inc_net_count + 1));
        }
        strcat(route_list, v->value);
        strcat(route_list, ";");
        /*printf("DEBUG: subnet value = %s\n", route_list); */
        v = v->next;
        ctx->inc_net_count++;
    }
    ctx->inc_net = GNUNET_parse_ipv4_network_specification(route_list);
    if (!route_list) free (route_list);
    route_list = NULL;

    /* Get excluded networks */
    v = getallvalues("excluded_networks", commands);
    ctx->exc_net_count = 0;
    while (v) {
        /* Allocate memory for the list of excluded networks */
        if (!route_list) {
            route_list = (char *)calloc(1, 2*INET_ADDRSTRLEN);
        } else {
            route_list = (char *)realloc(route_list, 2*INET_ADDRSTRLEN*(ctx->exc_net_count + 1));
        }
        strcat(route_list, v->value);
        strcat(route_list, ";");
        /*printf("DEBUG: subnet value = %s\n", route_list); */
        v = v->next;
        ctx->exc_net_count++;
    }

    ctx->exc_net = GNUNET_parse_ipv4_network_specification(route_list);
    if (!route_list) free (route_list);

    freecmds(commands);

    if (!ctx->dbuser || !ctx->dbpass || !ctx->dbname || !ctx->dbheader || !ctx->dbflows) {
        syslog(LOG_ERR, "oracle: required parameter missing ;)");
        return;
    }

    /* Initialize the Environment and allocate handles */
    syslog(LOG_ERR, "Environment setup ....\n");

    /* Initialize the mode to be the threaded and object environment */
    /*if (OCIEnvNlsCreate((OCIEnv **) &ctx->envhp, OCI_THREADED | OCI_OBJECT | OCI_NO_MUTEX, (dvoid *)0,
                        NULL, NULL, NULL, 0, (dvoid *) 0, (ub2) 0, (ub2) 0)) {
        printf("%s\n", "FAILED: OCIEnvCreate()");
        /*    return OCI_ERROR;*
    }
    printf("Environment created\n");

    /* Allocate a service handle */
    if (OCIHandleAlloc((dvoid *) envhp, (dvoid **) &ctx->svchp,
                       OCI_HTYPE_SVCCTX, (size_t) 0, (dvoid **) 0)) {
        syslog(LOG_ERR, "FAILED: OCIHandleAlloc() on svchp\n");
        /*    return OCI_ERROR;*/
    }

    /* Allocate an error handle */
    if (OCIHandleAlloc((dvoid *) envhp, (dvoid **) &ctx->errhp,
                       OCI_HTYPE_ERROR, (size_t) 0, (dvoid **) 0)) {
        syslog(LOG_ERR, "FAILED: OCIHandleAlloc() on errhp\n");
        /*    return OCI_ERROR;*/
    }

    /* Allocate a server handle */
    if (OCIHandleAlloc((dvoid *) envhp, (dvoid **) &ctx->srvhp,
                       OCI_HTYPE_SERVER, (size_t) 0, (dvoid **) 0)) {
        syslog(LOG_ERR, "FAILED: OCIHandleAlloc() on srvhp\n");
        /*    return OCI_ERROR;*/
    }

    /* Allocate a authentication handle */
    if (OCIHandleAlloc((dvoid *) envhp, (dvoid **) &ctx->authp,
                       OCI_HTYPE_SESSION, (size_t) 0, (dvoid **) 0)) {
        syslog(LOG_ERR, "FAILED: OCIHandleAlloc() on authp\n");
        /*    return OCI_ERROR;*/
    }

    /* Attach to the database server */
    if (OCIServerAttach(ctx->srvhp, ctx->errhp, (text *)ctx->dbname,
                        strlen(ctx->dbname), 0)) {
        syslog(LOG_ERR, "FAILED: OCIServerAttach()\n");
        /*    return OCI_ERROR;*/
    }

    /* Set the server handle in the service handle */
    if (OCIAttrSet((dvoid *) ctx->svchp, (ub4) OCI_HTYPE_SVCCTX,
                   (dvoid *) ctx->srvhp, (ub4) 0, (ub4) OCI_ATTR_SERVER, ctx->errhp)) {
        syslog(LOG_ERR, "FAILED: OCIAttrSet() server attribute\n");
        /*    return OCI_ERROR;*/
    }

    /* Logon to the server and begin a session */
    /* Set attributes in the authentication handle */
    if (OCIAttrSet((dvoid *) ctx->authp, (ub4) OCI_HTYPE_SESSION,
                   (dvoid *) ctx->dbuser, (ub4) strlen(ctx->dbuser),
                   (ub4) OCI_ATTR_USERNAME, ctx->errhp)) {
        syslog(LOG_ERR, "FAILED: OCIAttrSet() userid\n");
        /*    return OCI_ERROR;*/
    }
    if (OCIAttrSet((dvoid *) ctx->authp, (ub4) OCI_HTYPE_SESSION,
                   (dvoid *) ctx->dbpass, (ub4) strlen(ctx->dbpass),
                   (ub4) OCI_ATTR_PASSWORD, ctx->errhp)) {
        syslog(LOG_ERR, "FAILED: OCIAttrSet() passwd\n");
        /*    return OCI_ERROR;*/
    }

    syslog(LOG_ERR, "Logging on as %s  ....\n", ctx->dbuser);

    if (OCISessionBegin(ctx->svchp, ctx->errhp, ctx->authp, OCI_CRED_RDBMS, OCI_DEFAULT)) {
        syslog(LOG_ERR, "FAILED: OCIAttrSet() passwd\n");
        /*    return OCI_ERROR;*/
    }

    syslog(LOG_ERR, "%s logged on.\n", ctx->dbuser);

    /* Set the authentication handle in the Service handle */
    if (OCIAttrSet((dvoid *) ctx->svchp, (ub4) OCI_HTYPE_SVCCTX,
                   (dvoid *) ctx->authp, (ub4) 0, (ub4) OCI_ATTR_SESSION, ctx->errhp)) {
        syslog(LOG_ERR, "FAILED: OCIAttrSet() session\n");
        sb4 errcode = 0;
        text errbuf[1024];
        (void) OCIErrorGet((dvoid *)ctx->errhp, (ub4) 1, (text *) NULL, &errcode,
                           errbuf, (ub4) sizeof(errbuf), OCI_HTYPE_ERROR);
        syslog(LOG_ERR, "Error - %.*s\n", 1024, errbuf);
        /*    return OCI_ERROR;*/
    }

    /* Allocate statement handles */
    if (OCIHandleAlloc((dvoid *) envhp, (dvoid **) &ctx->stmt_flow1hdr_t,
                       (ub4)OCI_HTYPE_STMT, (CONST size_t) 0, (dvoid **) 0)) {
        syslog(LOG_ERR, "FAILED: alloc stmt_flow1hdr_t statement handle\n");
        return;
    }

    if (OCIHandleAlloc((dvoid *) envhp, (dvoid **) &ctx->stmt_flow5hdr_t,
                       (ub4)OCI_HTYPE_STMT, (CONST size_t) 0, (dvoid **) 0)) {
        syslog(LOG_ERR, "FAILED: alloc stmt_flow5hdr_t statement handle\n");
        return;
    }

    if (OCIHandleAlloc((dvoid *) envhp, (dvoid **) &ctx->stmt_flow7hdr_t,
                       (ub4)OCI_HTYPE_STMT, (CONST size_t) 0, (dvoid **) 0)) {
        syslog(LOG_ERR, "FAILED: alloc stmt_flow7hdr_t statement handle\n");
        return;
    }

    if (OCIHandleAlloc((dvoid *) envhp, (dvoid **) &ctx->stmt_flow1_t,
                       (ub4)OCI_HTYPE_STMT, (CONST size_t) 0, (dvoid **) 0)) {
        syslog(LOG_ERR, "FAILED: alloc stmt_flow1_t statement handle\n");
        return;
    }

    if (OCIHandleAlloc((dvoid *) envhp, (dvoid **) &ctx->stmt_flow5_t,
                       (ub4)OCI_HTYPE_STMT, (CONST size_t) 0, (dvoid **) 0)) {
        syslog(LOG_ERR, "FAILED: alloc stmt_flow5_t statement handle\n");
        return;
    }

    if (OCIHandleAlloc((dvoid *) envhp, (dvoid **) &ctx->stmt_flow7_t,
                       (ub4)OCI_HTYPE_STMT, (CONST size_t) 0, (dvoid **) 0)) {
        syslog(LOG_ERR, "FAILED: alloc stmt_flow7_t statement handle\n");
        return;
    }

    /* Connedtion is established */
    ctx->connected = 1;

    /* Create SQL-insert statements */
    sprintf(ctx->flow1hdr_t_insert, "insert into %s (address, version_id, "
            "count_id, uptime, secs, nsecs) values (:address, :version_id, "
            ":count_id, :uptime, :secs, :nsecs)", ctx->dbheader);

    sprintf(ctx->flow5hdr_t_insert, "insert into %s (address, version_id, "
            "count_id, uptime, secs, nsecs, flowseq, engine, eng_id, sampling) "
            "values (:address, :version_id, :count_id, :uptime, :secs, :nsecs, "
            ":flowseq, :engine, :eng_id, :sampling)", ctx->dbheader);

    sprintf(ctx->flow7hdr_t_insert, "insert into %s (address, version_id, "
            "count_id, uptime, secs, nsecs, flowseq, resv) values (:address, "
            ":version_id, :count_id, :uptime, :secs, :nsecs, :flowseq, :resv)",
            ctx->dbheader);

    sprintf(ctx->flow1_t_insert, "insert into %s (address, version_id, number_id, "
                "srcaddr, dstaddr, nexthop, input_if, output_if, dpkts, doctets, "
                "fuptime, luptime, srcport, dstport, resv1, flags, prot, tos, "
                "tcp_retx_cnt, tcp_retx_sec, tcp_misseq_cnt, resv2) values "
                "(:address, :version_id, :number_id, :srcaddr, :dstaddr, :nexthop, "
                ":input_if, :output_if, :dpkts, :doctets, :fuptime, :luptime, "
                ":srcport, :dstport, :resv1, :flags, :prot, :tos, :tcp_retx_cnt, "
                ":tcp_retx_sec, :tcp_misseq_cnt, :resv2)", ctx->dbflows);

    sprintf(ctx->flow5_t_insert, "insert into %s (address, version_id, number_id, "
                "srcaddr, dstaddr, nexthop, input_if, output_if, doctets, dpkts, "
                "fuptime, luptime, srcport, dstport, resv1, flags, prot, tos, "
                "srcmask, dstmask) values "
                "(:address, :version_id, :number_id, :srcaddr, :dstaddr, :nexthop, "
                ":input_if, :output_if, :doctets, :dpkts, :fuptime, :luptime, "
                ":srcport, :dstport, :resv1, :flags, :prot, :tos, :srcmask, "
                ":dstmask)", ctx->dbflows);

    sprintf(ctx->flow7_t_insert, "insert into %s (address, version_id, number_id, "
                "srcaddr, dstaddr, nexthop, input_if, output_if, doctets, dpkts, "
                "fuptime, luptime, srcport, dstport, mode_id, flags, prot, tos, "
                "src_as, dst_as, srcmask, dstmask, resv2, router) values "
                "(:address, :version_id, :number_id, :srcaddr, :dstaddr, :nexthop, "
                ":input_if, :output_if, :doctets, :dpkts, :fuptime, :luptime, "
                ":srcport, :dstport, :mode_id, :flags, :prot, :tos, :src_as, :dst_as, "
                ":srcmask, :dstmask, :resv2, :router)", ctx->dbflows);

    /* Prepare queries */
    if (OCIStmtPrepare(ctx->stmt_flow1hdr_t, ctx->errhp, ctx->flow1hdr_t_insert,
                      (ub4) strlen((char *)ctx->flow1hdr_t_insert),
                      (ub4) OCI_NTV_SYNTAX, (ub4) OCI_DEFAULT))
    {
        syslog(LOG_ERR, "FAILED: OCIStmtPrepare() stmt_flow1hdr_t");
        report_error(ctx->errhp);
    }

    if (OCIStmtPrepare(ctx->stmt_flow5hdr_t, ctx->errhp, ctx->flow5hdr_t_insert,
                      (ub4) strlen((char *)ctx->flow5hdr_t_insert),
                      (ub4) OCI_NTV_SYNTAX, (ub4) OCI_DEFAULT))
    {
        syslog(LOG_ERR, "FAILED: OCIStmtPrepare() stmt_flow5hdr_t");
        report_error(ctx->errhp);
    }

    if (OCIStmtPrepare(ctx->stmt_flow7hdr_t, ctx->errhp, ctx->flow7hdr_t_insert,
                      (ub4) strlen((char *)ctx->flow7hdr_t_insert),
                      (ub4) OCI_NTV_SYNTAX, (ub4) OCI_DEFAULT))
    {
        syslog(LOG_ERR, "FAILED: OCIStmtPrepare() stmt_flow7hdr_t");
        report_error(ctx->errhp);
    }

    if (OCIStmtPrepare(ctx->stmt_flow1_t, ctx->errhp, ctx->flow1_t_insert,
                      (ub4) strlen((char *)ctx->flow1_t_insert),
                      (ub4) OCI_NTV_SYNTAX, (ub4) OCI_DEFAULT))
    {
        syslog(LOG_ERR, "FAILED: OCIStmtPrepare() stmt_flow1_t");
        report_error(ctx->errhp);
    }

    if (OCIStmtPrepare(ctx->stmt_flow5_t, ctx->errhp, ctx->flow5_t_insert,
                      (ub4) strlen((char *)ctx->flow5_t_insert),
                      (ub4) OCI_NTV_SYNTAX, (ub4) OCI_DEFAULT))
    {
        syslog(LOG_ERR, "FAILED: OCIStmtPrepare() stmt_flow5_t");
        report_error(ctx->errhp);
    }

    if (OCIStmtPrepare(ctx->stmt_flow7_t, ctx->errhp, ctx->flow7_t_insert,
                      (ub4) strlen((char *)ctx->flow7_t_insert),
                      (ub4) OCI_NTV_SYNTAX, (ub4) OCI_DEFAULT))
    {
        syslog(LOG_ERR, "FAILED: OCIStmtPrepare() stmt_flow7_t");
        report_error(ctx->errhp);
    }

    /* Bind flow1header data to query parameters */
    if ((status = OCIBindByName(ctx->stmt_flow1hdr_t, &ctx->flow1hdr_t_bind.address,
                    ctx->errhp, (text *) ":address", -1, (dvoid *) &ctx->flowhdr_t_packet.address,
                    (sword) sizeof(ctx->flowhdr_t_packet.address), SQLT_UIN, (dvoid *) 0, (ub2 *) 0,
                    (ub2 *) 0, (ub4) 0, (ub4 *) 0, OCI_DEFAULT)) ||
        (status = OCIBindByName(ctx->stmt_flow1hdr_t, &ctx->flow1hdr_t_bind.version,
                    ctx->errhp, (text *) ":version_id", -1, (dvoid *) &ctx->flowhdr_t_packet.version,
                    (sword) sizeof(ctx->flowhdr_t_packet.version), SQLT_UIN, (dvoid *) 0, (ub2 *) 0,
                    (ub2 *) 0, (ub4) 0, (ub4 *) 0, OCI_DEFAULT)) ||
        (status = OCIBindByName(ctx->stmt_flow1hdr_t, &ctx->flow1hdr_t_bind.count,
                    ctx->errhp, (text *) ":count_id", -1, (dvoid *) &ctx->flowhdr_t_packet.count,
                    (sword) sizeof(ctx->flowhdr_t_packet.count), SQLT_UIN, (dvoid *) 0, (ub2 *) 0,
                    (ub2 *) 0, (ub4) 0, (ub4 *) 0, OCI_DEFAULT)) ||
        (status = OCIBindByName(ctx->stmt_flow1hdr_t, &ctx->flow1hdr_t_bind.uptime,
                    ctx->errhp, (text *) ":uptime", -1, (dvoid *) &ctx->flowhdr_t_packet.uptime,
                    (sword) sizeof(ctx->flowhdr_t_packet.uptime), SQLT_UIN, (dvoid *) 0, (ub2 *) 0,
                    (ub2 *) 0, (ub4) 0, (ub4 *) 0, OCI_DEFAULT)) ||
        (status = OCIBindByName(ctx->stmt_flow1hdr_t, &ctx->flow1hdr_t_bind.secs,
                    ctx->errhp, (text *) ":secs", -1, (dvoid *) &ctx->flowhdr_t_packet.secs,
                    (sword) sizeof(ctx->flowhdr_t_packet.secs), SQLT_UIN, (dvoid *) 0, (ub2 *) 0,
                    (ub2 *) 0, (ub4) 0, (ub4 *) 0, OCI_DEFAULT)) ||
        (status = OCIBindByName(ctx->stmt_flow1hdr_t, &ctx->flow1hdr_t_bind.nsecs,
                    ctx->errhp, (text *) ":nsecs", -1, (dvoid *) &ctx->flowhdr_t_packet.nsecs,
                    (sword) sizeof(ctx->flowhdr_t_packet.nsecs), SQLT_UIN, (dvoid *) 0, (ub2 *) 0,
                    (ub2 *) 0, (ub4) 0, (ub4 *) 0, OCI_DEFAULT)))
                    {
        syslog(LOG_ERR, "FAILED: BINDByName() for flow1header data");
        report_error(ctx->errhp);
                    }

    /* Bind flow5header data to query parameters */
    if ((status = OCIBindByName(ctx->stmt_flow5hdr_t, &ctx->flow5hdr_t_bind.address,
                    ctx->errhp, (text *) ":address", -1, (dvoid *) &ctx->flowhdr_t_packet.address,
                    (sword) sizeof(ctx->flowhdr_t_packet.address), SQLT_UIN, (dvoid *) 0, (ub2 *) 0,
                    (ub2 *) 0, (ub4) 0, (ub4 *) 0, OCI_DEFAULT)) ||
        (status = OCIBindByName(ctx->stmt_flow5hdr_t, &ctx->flow5hdr_t_bind.version,
                    ctx->errhp, (text *) ":version_id", -1, (dvoid *) &ctx->flowhdr_t_packet.version,
                    (sword) sizeof(ctx->flowhdr_t_packet.version), SQLT_UIN, (dvoid *) 0, (ub2 *) 0,
                    (ub2 *) 0, (ub4) 0, (ub4 *) 0, OCI_DEFAULT)) ||
        (status = OCIBindByName(ctx->stmt_flow5hdr_t, &ctx->flow5hdr_t_bind.count,
                    ctx->errhp, (text *) ":count_id", -1, (dvoid *) &ctx->flowhdr_t_packet.count,
                    (sword) sizeof(ctx->flowhdr_t_packet.count), SQLT_UIN, (dvoid *) 0, (ub2 *) 0,
                    (ub2 *) 0, (ub4) 0, (ub4 *) 0, OCI_DEFAULT)) ||
        (status = OCIBindByName(ctx->stmt_flow5hdr_t, &ctx->flow5hdr_t_bind.uptime,
                    ctx->errhp, (text *) ":uptime", -1, (dvoid *) &ctx->flowhdr_t_packet.uptime,
                    (sword) sizeof(ctx->flowhdr_t_packet.uptime), SQLT_UIN, (dvoid *) 0, (ub2 *) 0,
                    (ub2 *) 0, (ub4) 0, (ub4 *) 0, OCI_DEFAULT)) ||
        (status = OCIBindByName(ctx->stmt_flow5hdr_t, &ctx->flow5hdr_t_bind.secs,
                    ctx->errhp, (text *) ":secs", -1, (dvoid *) &ctx->flowhdr_t_packet.secs,
                    (sword) sizeof(ctx->flowhdr_t_packet.secs), SQLT_UIN, (dvoid *) 0, (ub2 *) 0,
                    (ub2 *) 0, (ub4) 0, (ub4 *) 0, OCI_DEFAULT)) ||
        (status = OCIBindByName(ctx->stmt_flow5hdr_t, &ctx->flow5hdr_t_bind.nsecs,
                    ctx->errhp, (text *) ":nsecs", -1, (dvoid *) &ctx->flowhdr_t_packet.nsecs,
                    (sword) sizeof(ctx->flowhdr_t_packet.nsecs), SQLT_UIN, (dvoid *) 0, (ub2 *) 0,
                    (ub2 *) 0, (ub4) 0, (ub4 *) 0, OCI_DEFAULT)) ||
        (status = OCIBindByName(ctx->stmt_flow5hdr_t, &ctx->flow5hdr_t_bind.flowseq,
                    ctx->errhp, (text *) ":flowseq", -1, (dvoid *) &ctx->flowhdr_t_packet.flowseq,
                    (sword) sizeof(ctx->flowhdr_t_packet.flowseq), SQLT_UIN, (dvoid *) 0, (ub2 *) 0,
                    (ub2 *) 0, (ub4) 0, (ub4 *) 0, OCI_DEFAULT)) ||
        (status = OCIBindByName(ctx->stmt_flow5hdr_t, &ctx->flow5hdr_t_bind.engine,
                    ctx->errhp, (text *) ":engine", -1, (dvoid *) &ctx->flowhdr_t_packet.engine,
                    (sword) sizeof(ctx->flowhdr_t_packet.engine), SQLT_UIN, (dvoid *) 0, (ub2 *) 0,
                    (ub2 *) 0, (ub4) 0, (ub4 *) 0, OCI_DEFAULT)) ||
        (status = OCIBindByName(ctx->stmt_flow5hdr_t, &ctx->flow5hdr_t_bind.eng_id,
                    ctx->errhp, (text *) ":eng_id", -1, (dvoid *) &ctx->flowhdr_t_packet.eng_id,
                    (sword) sizeof(ctx->flowhdr_t_packet.eng_id), SQLT_UIN, (dvoid *) 0, (ub2 *) 0,
                    (ub2 *) 0, (ub4) 0, (ub4 *) 0, OCI_DEFAULT)) ||
        (status = OCIBindByName(ctx->stmt_flow5hdr_t, &ctx->flow5hdr_t_bind.sampling,
                    ctx->errhp, (text *) ":sampling", -1, (dvoid *) &ctx->flowhdr_t_packet.sampling,
                    (sword) sizeof(ctx->flowhdr_t_packet.sampling), SQLT_UIN, (dvoid *) 0, (ub2 *) 0,
                    (ub2 *) 0, (ub4) 0, (ub4 *) 0, OCI_DEFAULT)))
                    {
        syslog(LOG_ERR, "FAILED: BINDByName() for flow5header data");
        report_error(ctx->errhp);
                    }

    /* Bind flow7header data to query parameters */
    if ((status = OCIBindByName(ctx->stmt_flow7hdr_t, &ctx->flow7hdr_t_bind.address,
                    ctx->errhp, (text *) ":address", -1, (dvoid *) &ctx->flowhdr_t_packet.address,
                    (sword) sizeof(ctx->flowhdr_t_packet.address), SQLT_UIN, (dvoid *) 0, (ub2 *) 0,
                    (ub2 *) 0, (ub4) 0, (ub4 *) 0, OCI_DEFAULT)) ||
        (status = OCIBindByName(ctx->stmt_flow7hdr_t, &ctx->flow7hdr_t_bind.version,
                    ctx->errhp, (text *) ":version_id", -1, (dvoid *) &ctx->flowhdr_t_packet.version,
                    (sword) sizeof(ctx->flowhdr_t_packet.version), SQLT_UIN, (dvoid *) 0, (ub2 *) 0,
                    (ub2 *) 0, (ub4) 0, (ub4 *) 0, OCI_DEFAULT)) ||
        (status = OCIBindByName(ctx->stmt_flow7hdr_t, &ctx->flow7hdr_t_bind.count,
                    ctx->errhp, (text *) ":count_id", -1, (dvoid *) &ctx->flowhdr_t_packet.count,
                    (sword) sizeof(ctx->flowhdr_t_packet.count), SQLT_UIN, (dvoid *) 0, (ub2 *) 0,
                    (ub2 *) 0, (ub4) 0, (ub4 *) 0, OCI_DEFAULT)) ||
        (status = OCIBindByName(ctx->stmt_flow7hdr_t, &ctx->flow7hdr_t_bind.uptime,
                    ctx->errhp, (text *) ":uptime", -1, (dvoid *) &ctx->flowhdr_t_packet.uptime,
                    (sword) sizeof(ctx->flowhdr_t_packet.uptime), SQLT_UIN, (dvoid *) 0, (ub2 *) 0,
                    (ub2 *) 0, (ub4) 0, (ub4 *) 0, OCI_DEFAULT)) ||
        (status = OCIBindByName(ctx->stmt_flow7hdr_t, &ctx->flow7hdr_t_bind.secs,
                    ctx->errhp, (text *) ":secs", -1, (dvoid *) &ctx->flowhdr_t_packet.secs,
                    (sword) sizeof(ctx->flowhdr_t_packet.secs), SQLT_UIN, (dvoid *) 0, (ub2 *) 0,
                    (ub2 *) 0, (ub4) 0, (ub4 *) 0, OCI_DEFAULT)) ||
        (status = OCIBindByName(ctx->stmt_flow7hdr_t, &ctx->flow7hdr_t_bind.flowseq,
                    ctx->errhp, (text *) ":flowseq", -1, (dvoid *) &ctx->flowhdr_t_packet.flowseq,
                    (sword) sizeof(ctx->flowhdr_t_packet.flowseq), SQLT_UIN, (dvoid *) 0, (ub2 *) 0,
                    (ub2 *) 0, (ub4) 0, (ub4 *) 0, OCI_DEFAULT)) ||
        (status = OCIBindByName(ctx->stmt_flow7hdr_t, &ctx->flow7hdr_t_bind.resv,
                    ctx->errhp, (text *) ":resv", -1, (dvoid *) &ctx->flowhdr_t_packet.resv,
                    (sword) sizeof(ctx->flowhdr_t_packet.resv), SQLT_UIN, (dvoid *) 0, (ub2 *) 0,
                    (ub2 *) 0, (ub4) 0, (ub4 *) 0, OCI_DEFAULT)))
                    {
        syslog(LOG_ERR, "FAILED: BINDByName() for flow7header data");
        report_error(ctx->errhp);
                    }

    /* Bind flow1 data to query parameters */
    if ((status = OCIBindByName(ctx->stmt_flow1_t, &ctx->flow1_t_bind.address,
                    ctx->errhp, (text *) ":address", -1, (dvoid *) &ctx->flow_t_packet.address,
                    (sword) sizeof(ctx->flow_t_packet.address), SQLT_UIN, (dvoid *) 0, (ub2 *) 0,
                    (ub2 *) 0, (ub4) 0, (ub4 *) 0, OCI_DEFAULT)) ||
        (status = OCIBindByName(ctx->stmt_flow1_t, &ctx->flow1_t_bind.version,
                    ctx->errhp, (text *) ":version_id", -1, (dvoid *) &ctx->flow_t_packet.version,
                    (sword) sizeof(ctx->flow_t_packet.version), SQLT_UIN, (dvoid *) 0, (ub2 *) 0,
                    (ub2 *) 0, (ub4) 0, (ub4 *) 0, OCI_DEFAULT)) ||
        (status = OCIBindByName(ctx->stmt_flow1_t, &ctx->flow1_t_bind.number,
                    ctx->errhp, (text *) ":number_id", -1, (dvoid *) &ctx->flow_t_packet.number,
                    (sword) sizeof(ctx->flow_t_packet.number), SQLT_UIN, (dvoid *) 0, (ub2 *) 0,
                    (ub2 *) 0, (ub4) 0, (ub4 *) 0, OCI_DEFAULT)) ||
        (status = OCIBindByName(ctx->stmt_flow1_t, &ctx->flow1_t_bind.srcaddr,
                    ctx->errhp, (text *) ":srcaddr", -1, (dvoid *) &ctx->flow_t_packet.srcaddr,
                    (sword) sizeof(ctx->flow_t_packet.srcaddr), SQLT_UIN, (dvoid *) 0, (ub2 *) 0,
                    (ub2 *) 0, (ub4) 0, (ub4 *) 0, OCI_DEFAULT)) ||
        (status = OCIBindByName(ctx->stmt_flow1_t, &ctx->flow1_t_bind.dstaddr,
                    ctx->errhp, (text *) ":dstaddr", -1, (dvoid *) &ctx->flow_t_packet.dstaddr,
                    (sword) sizeof(ctx->flow_t_packet.dstaddr), SQLT_UIN, (dvoid *) 0, (ub2 *) 0,
                    (ub2 *) 0, (ub4) 0, (ub4 *) 0, OCI_DEFAULT)) ||
        (status = OCIBindByName(ctx->stmt_flow1_t, &ctx->flow1_t_bind.nexthop,
                    ctx->errhp, (text *) ":nexthop", -1, (dvoid *) &ctx->flow_t_packet.nexthop,
                    (sword) sizeof(ctx->flow_t_packet.nexthop), SQLT_UIN, (dvoid *) 0, (ub2 *) 0,
                    (ub2 *) 0, (ub4) 0, (ub4 *) 0, OCI_DEFAULT)) ||
        (status = OCIBindByName(ctx->stmt_flow1_t, &ctx->flow1_t_bind.input_if,
                    ctx->errhp, (text *) ":input_if", -1, (dvoid *) &ctx->flow_t_packet.input_if,
                    (sword) sizeof(ctx->flow_t_packet.input_if), SQLT_UIN, (dvoid *) 0, (ub2 *) 0,
                    (ub2 *) 0, (ub4) 0, (ub4 *) 0, OCI_DEFAULT)) ||
        (status = OCIBindByName(ctx->stmt_flow1_t, &ctx->flow1_t_bind.output_if,
                    ctx->errhp, (text *) ":output_if", -1, (dvoid *) &ctx->flow_t_packet.output_if,
                    (sword) sizeof(ctx->flow_t_packet.output_if), SQLT_UIN, (dvoid *) 0, (ub2 *) 0,
                    (ub2 *) 0, (ub4) 0, (ub4 *) 0, OCI_DEFAULT)) ||
        (status = OCIBindByName(ctx->stmt_flow1_t, &ctx->flow1_t_bind.doctets,
                    ctx->errhp, (text *) ":doctets", -1, (dvoid *) &ctx->flow_t_packet.doctets,
                    (sword) sizeof(ctx->flow_t_packet.doctets), SQLT_UIN, (dvoid *) 0, (ub2 *) 0,
                    (ub2 *) 0, (ub4) 0, (ub4 *) 0, OCI_DEFAULT)) ||
        (status = OCIBindByName(ctx->stmt_flow1_t, &ctx->flow1_t_bind.dpkts,
                    ctx->errhp, (text *) ":dpkts", -1, (dvoid *) &ctx->flow_t_packet.dpkts,
                    (sword) sizeof(ctx->flow_t_packet.dpkts), SQLT_UIN, (dvoid *) 0, (ub2 *) 0,
                    (ub2 *) 0, (ub4) 0, (ub4 *) 0, OCI_DEFAULT)) ||
        (status = OCIBindByName(ctx->stmt_flow1_t, &ctx->flow1_t_bind.fuptime,
                    ctx->errhp, (text *) ":fuptime", -1, (dvoid *) &ctx->flow_t_packet.fuptime,
                    (sword) sizeof(ctx->flow_t_packet.fuptime), SQLT_UIN, (dvoid *) 0, (ub2 *) 0,
                    (ub2 *) 0, (ub4) 0, (ub4 *) 0, OCI_DEFAULT)) ||
        (status = OCIBindByName(ctx->stmt_flow1_t, &ctx->flow1_t_bind.luptime,
                    ctx->errhp, (text *) ":luptime", -1, (dvoid *) &ctx->flow_t_packet.luptime,
                    (sword) sizeof(ctx->flow_t_packet.luptime), SQLT_UIN, (dvoid *) 0, (ub2 *) 0,
                    (ub2 *) 0, (ub4) 0, (ub4 *) 0, OCI_DEFAULT)) ||
        (status = OCIBindByName(ctx->stmt_flow1_t, &ctx->flow1_t_bind.srcport,
                    ctx->errhp, (text *) ":srcport", -1, (dvoid *) &ctx->flow_t_packet.srcport,
                    (sword) sizeof(ctx->flow_t_packet.srcport), SQLT_UIN, (dvoid *) 0, (ub2 *) 0,
                    (ub2 *) 0, (ub4) 0, (ub4 *) 0, OCI_DEFAULT)) ||
        (status = OCIBindByName(ctx->stmt_flow1_t, &ctx->flow1_t_bind.dstport,
                    ctx->errhp, (text *) ":dstport", -1, (dvoid *) &ctx->flow_t_packet.dstport,
                    (sword) sizeof(ctx->flow_t_packet.dstport), SQLT_UIN, (dvoid *) 0, (ub2 *) 0,
                    (ub2 *) 0, (ub4) 0, (ub4 *) 0, OCI_DEFAULT)) ||
        (status = OCIBindByName(ctx->stmt_flow1_t, &ctx->flow1_t_bind.resv1,
                    ctx->errhp, (text *) ":resv1", -1, (dvoid *) &ctx->flow_t_packet.resv1,
                    (sword) sizeof(ctx->flow_t_packet.resv1), SQLT_UIN, (dvoid *) 0, (ub2 *) 0,
                    (ub2 *) 0, (ub4) 0, (ub4 *) 0, OCI_DEFAULT)) ||
        (status = OCIBindByName(ctx->stmt_flow1_t, &ctx->flow1_t_bind.flags,
                    ctx->errhp, (text *) ":flags", -1, (dvoid *) &ctx->flow_t_packet.flags,
                    (sword) sizeof(ctx->flow_t_packet.flags), SQLT_INT, (dvoid *) 0, (ub2 *) 0,
                    (ub2 *) 0, (ub4) 0, (ub4 *) 0, OCI_DEFAULT)) ||
        (status = OCIBindByName(ctx->stmt_flow1_t, &ctx->flow1_t_bind.prot,
                    ctx->errhp, (text *) ":prot", -1, (dvoid *) &ctx->flow_t_packet.prot,
                    (sword) sizeof(ctx->flow_t_packet.prot), SQLT_INT, (dvoid *) 0, (ub2 *) 0,
                    (ub2 *) 0, (ub4) 0, (ub4 *) 0, OCI_DEFAULT)) ||
        (status = OCIBindByName(ctx->stmt_flow1_t, &ctx->flow1_t_bind.tos,
                    ctx->errhp, (text *) ":tos", -1, (dvoid *) &ctx->flow_t_packet.tos,
                    (sword) sizeof(ctx->flow_t_packet.tos), SQLT_INT, (dvoid *) 0, (ub2 *) 0,
                    (ub2 *) 0, (ub4) 0, (ub4 *) 0, OCI_DEFAULT)) ||
        (status = OCIBindByName(ctx->stmt_flow1_t, &ctx->flow1_t_bind.tcp_retx_cnt,
                    ctx->errhp, (text *) ":tcp_retx_cnt", -1, (dvoid *) &ctx->flow_t_packet.tcp_retx_cnt,
                    (sword) sizeof(ctx->flow_t_packet.tcp_retx_cnt), SQLT_INT, (dvoid *) 0, (ub2 *) 0,
                    (ub2 *) 0, (ub4) 0, (ub4 *) 0, OCI_DEFAULT)) ||
        (status = OCIBindByName(ctx->stmt_flow1_t, &ctx->flow1_t_bind.tcp_retx_sec,
                    ctx->errhp, (text *) ":tcp_retx_sec", -1, (dvoid *) &ctx->flow_t_packet.tcp_retx_sec,
                    (sword) sizeof(ctx->flow_t_packet.tcp_retx_sec), SQLT_INT, (dvoid *) 0, (ub2 *) 0,
                    (ub2 *) 0, (ub4) 0, (ub4 *) 0, OCI_DEFAULT)) ||
        (status = OCIBindByName(ctx->stmt_flow1_t, &ctx->flow1_t_bind.tcp_misseq_cnt,
                    ctx->errhp, (text *) ":tcp_misseq_cnt", -1, (dvoid *) &ctx->flow_t_packet.tcp_misseq_cnt,
                    (sword) sizeof(ctx->flow_t_packet.tcp_misseq_cnt), SQLT_INT, (dvoid *) 0, (ub2 *) 0,
                    (ub2 *) 0, (ub4) 0, (ub4 *) 0, OCI_DEFAULT)) ||
        (status = OCIBindByName(ctx->stmt_flow1_t, &ctx->flow1_t_bind.resv2,
                    ctx->errhp, (text *) ":resv2", -1, (dvoid *) &ctx->flow_t_packet.resv2,
                    (sword) sizeof(ctx->flow_t_packet.resv2), SQLT_UIN, (dvoid *) 0, (ub2 *) 0,
                    (ub2 *) 0, (ub4) 0, (ub4 *) 0, OCI_DEFAULT)))
                    {
        syslog(LOG_ERR, "FAILED: BINDByName() for flow1 data");
        report_error(ctx->errhp);
                    }

    /* Bind flow5 data to query parameters */
    if ((status = OCIBindByName(ctx->stmt_flow5_t, &ctx->flow5_t_bind.address,
                    ctx->errhp, (text *) ":address", -1, (dvoid *) &ctx->flow_t_packet.address,
                    (sword) sizeof(ctx->flow_t_packet.address), SQLT_UIN, (dvoid *) 0, (ub2 *) 0,
                    (ub2 *) 0, (ub4) 0, (ub4 *) 0, OCI_DEFAULT)) ||
        (status = OCIBindByName(ctx->stmt_flow5_t, &ctx->flow5_t_bind.version,
                    ctx->errhp, (text *) ":version_id", -1, (dvoid *) &ctx->flow_t_packet.version,
                    (sword) sizeof(ctx->flow_t_packet.version), SQLT_UIN, (dvoid *) 0, (ub2 *) 0,
                    (ub2 *) 0, (ub4) 0, (ub4 *) 0, OCI_DEFAULT)) ||
        (status = OCIBindByName(ctx->stmt_flow5_t, &ctx->flow5_t_bind.number,
                    ctx->errhp, (text *) ":number_id", -1, (dvoid *) &ctx->flow_t_packet.number,
                    (sword) sizeof(ctx->flow_t_packet.number), SQLT_UIN, (dvoid *) 0, (ub2 *) 0,
                    (ub2 *) 0, (ub4) 0, (ub4 *) 0, OCI_DEFAULT)) ||
        (status = OCIBindByName(ctx->stmt_flow5_t, &ctx->flow5_t_bind.srcaddr,
                    ctx->errhp, (text *) ":srcaddr", -1, (dvoid *) &ctx->flow_t_packet.srcaddr,
                    (sword) sizeof(ctx->flow_t_packet.srcaddr), SQLT_UIN, (dvoid *) 0, (ub2 *) 0,
                    (ub2 *) 0, (ub4) 0, (ub4 *) 0, OCI_DEFAULT)) ||
        (status = OCIBindByName(ctx->stmt_flow5_t, &ctx->flow5_t_bind.dstaddr,
                    ctx->errhp, (text *) ":dstaddr", -1, (dvoid *) &ctx->flow_t_packet.dstaddr,
                    (sword) sizeof(ctx->flow_t_packet.dstaddr), SQLT_UIN, (dvoid *) 0, (ub2 *) 0,
                    (ub2 *) 0, (ub4) 0, (ub4 *) 0, OCI_DEFAULT)) ||
        (status = OCIBindByName(ctx->stmt_flow5_t, &ctx->flow5_t_bind.nexthop,
                    ctx->errhp, (text *) ":nexthop", -1, (dvoid *) &ctx->flow_t_packet.nexthop,
                    (sword) sizeof(ctx->flow_t_packet.nexthop), SQLT_UIN, (dvoid *) 0, (ub2 *) 0,
                    (ub2 *) 0, (ub4) 0, (ub4 *) 0, OCI_DEFAULT)) ||
        (status = OCIBindByName(ctx->stmt_flow5_t, &ctx->flow5_t_bind.input_if,
                    ctx->errhp, (text *) ":input_if", -1, (dvoid *) &ctx->flow_t_packet.input_if,
                    (sword) sizeof(ctx->flow_t_packet.input_if), SQLT_UIN, (dvoid *) 0, (ub2 *) 0,
                    (ub2 *) 0, (ub4) 0, (ub4 *) 0, OCI_DEFAULT)) ||
        (status = OCIBindByName(ctx->stmt_flow5_t, &ctx->flow5_t_bind.output_if,
                    ctx->errhp, (text *) ":output_if", -1, (dvoid *) &ctx->flow_t_packet.output_if,
                    (sword) sizeof(ctx->flow_t_packet.output_if), SQLT_UIN, (dvoid *) 0, (ub2 *) 0,
                    (ub2 *) 0, (ub4) 0, (ub4 *) 0, OCI_DEFAULT)) ||
        (status = OCIBindByName(ctx->stmt_flow5_t, &ctx->flow5_t_bind.doctets,
                    ctx->errhp, (text *) ":doctets", -1, (dvoid *) &ctx->flow_t_packet.doctets,
                    (sword) sizeof(ctx->flow_t_packet.doctets), SQLT_UIN, (dvoid *) 0, (ub2 *) 0,
                    (ub2 *) 0, (ub4) 0, (ub4 *) 0, OCI_DEFAULT)) ||
        (status = OCIBindByName(ctx->stmt_flow5_t, &ctx->flow5_t_bind.dpkts,
                    ctx->errhp, (text *) ":dpkts", -1, (dvoid *) &ctx->flow_t_packet.dpkts,
                    (sword) sizeof(ctx->flow_t_packet.dpkts), SQLT_UIN, (dvoid *) 0, (ub2 *) 0,
                    (ub2 *) 0, (ub4) 0, (ub4 *) 0, OCI_DEFAULT)) ||
        (status = OCIBindByName(ctx->stmt_flow5_t, &ctx->flow5_t_bind.fuptime,
                    ctx->errhp, (text *) ":fuptime", -1, (dvoid *) &ctx->flow_t_packet.fuptime,
                    (sword) sizeof(ctx->flow_t_packet.fuptime), SQLT_UIN, (dvoid *) 0, (ub2 *) 0,
                    (ub2 *) 0, (ub4) 0, (ub4 *) 0, OCI_DEFAULT)) ||
        (status = OCIBindByName(ctx->stmt_flow5_t, &ctx->flow5_t_bind.luptime,
                    ctx->errhp, (text *) ":luptime", -1, (dvoid *) &ctx->flow_t_packet.luptime,
                    (sword) sizeof(ctx->flow_t_packet.luptime), SQLT_UIN, (dvoid *) 0, (ub2 *) 0,
                    (ub2 *) 0, (ub4) 0, (ub4 *) 0, OCI_DEFAULT)) ||
        (status = OCIBindByName(ctx->stmt_flow5_t, &ctx->flow5_t_bind.srcport,
                    ctx->errhp, (text *) ":srcport", -1, (dvoid *) &ctx->flow_t_packet.srcport,
                    (sword) sizeof(ctx->flow_t_packet.srcport), SQLT_UIN, (dvoid *) 0, (ub2 *) 0,
                    (ub2 *) 0, (ub4) 0, (ub4 *) 0, OCI_DEFAULT)) ||
        (status = OCIBindByName(ctx->stmt_flow5_t, &ctx->flow5_t_bind.dstport,
                    ctx->errhp, (text *) ":dstport", -1, (dvoid *) &ctx->flow_t_packet.dstport,
                    (sword) sizeof(ctx->flow_t_packet.dstport), SQLT_UIN, (dvoid *) 0, (ub2 *) 0,
                    (ub2 *) 0, (ub4) 0, (ub4 *) 0, OCI_DEFAULT)) ||
        (status = OCIBindByName(ctx->stmt_flow5_t, &ctx->flow5_t_bind.resv1,
                    ctx->errhp, (text *) ":resv1", -1, (dvoid *) &ctx->flow_t_packet.resv1,
                    (sword) sizeof(ctx->flow_t_packet.resv1), SQLT_UIN, (dvoid *) 0, (ub2 *) 0,
                    (ub2 *) 0, (ub4) 0, (ub4 *) 0, OCI_DEFAULT)) ||
        (status = OCIBindByName(ctx->stmt_flow5_t, &ctx->flow5_t_bind.flags,
                    ctx->errhp, (text *) ":flags", -1, (dvoid *) &ctx->flow_t_packet.flags,
                    (sword) sizeof(ctx->flow_t_packet.flags), SQLT_INT, (dvoid *) 0, (ub2 *) 0,
                    (ub2 *) 0, (ub4) 0, (ub4 *) 0, OCI_DEFAULT)) ||
        (status = OCIBindByName(ctx->stmt_flow5_t, &ctx->flow5_t_bind.prot,
                    ctx->errhp, (text *) ":prot", -1, (dvoid *) &ctx->flow_t_packet.prot,
                    (sword) sizeof(ctx->flow_t_packet.prot), SQLT_INT, (dvoid *) 0, (ub2 *) 0,
                    (ub2 *) 0, (ub4) 0, (ub4 *) 0, OCI_DEFAULT)) ||
        (status = OCIBindByName(ctx->stmt_flow5_t, &ctx->flow5_t_bind.tos,
                    ctx->errhp, (text *) ":tos", -1, (dvoid *) &ctx->flow_t_packet.tos,
                    (sword) sizeof(ctx->flow_t_packet.tos), SQLT_INT, (dvoid *) 0, (ub2 *) 0,
                    (ub2 *) 0, (ub4) 0, (ub4 *) 0, OCI_DEFAULT)) ||
        (status = OCIBindByName(ctx->stmt_flow5_t, &ctx->flow5_t_bind.srcmask,
                    ctx->errhp, (text *) ":srcmask", -1, (dvoid *) &ctx->flow_t_packet.srcmask,
                    (sword) sizeof(ctx->flow_t_packet.srcmask), SQLT_UIN, (dvoid *) 0, (ub2 *) 0,
                    (ub2 *) 0, (ub4) 0, (ub4 *) 0, OCI_DEFAULT)) ||
        (status = OCIBindByName(ctx->stmt_flow5_t, &ctx->flow5_t_bind.dstmask,
                    ctx->errhp, (text *) ":dstmask", -1, (dvoid *) &ctx->flow_t_packet.dstmask,
                    (sword) sizeof(ctx->flow_t_packet.dstmask), SQLT_UIN, (dvoid *) 0, (ub2 *) 0,
                    (ub2 *) 0, (ub4) 0, (ub4 *) 0, OCI_DEFAULT)))
                    {
        syslog(LOG_ERR, "FAILED: BINDByName() for flow5 data");
        report_error(ctx->errhp);
                    }

    /* Bind flow7 data to query parameters */
    if ((status = OCIBindByName(ctx->stmt_flow7_t, &ctx->flow7_t_bind.address,
                    ctx->errhp, (text *) ":address", -1, (dvoid *) &ctx->flow_t_packet.address,
                    (sword) sizeof(ctx->flow_t_packet.address), SQLT_UIN, (dvoid *) 0, (ub2 *) 0,
                    (ub2 *) 0, (ub4) 0, (ub4 *) 0, OCI_DEFAULT)) ||
        (status = OCIBindByName(ctx->stmt_flow7_t, &ctx->flow7_t_bind.version,
                    ctx->errhp, (text *) ":version_id", -1, (dvoid *) &ctx->flow_t_packet.version,
                    (sword) sizeof(ctx->flow_t_packet.version), SQLT_UIN, (dvoid *) 0, (ub2 *) 0,
                    (ub2 *) 0, (ub4) 0, (ub4 *) 0, OCI_DEFAULT)) ||
        (status = OCIBindByName(ctx->stmt_flow7_t, &ctx->flow7_t_bind.number,
                    ctx->errhp, (text *) ":number_id", -1, (dvoid *) &ctx->flow_t_packet.number,
                    (sword) sizeof(ctx->flow_t_packet.number), SQLT_UIN, (dvoid *) 0, (ub2 *) 0,
                    (ub2 *) 0, (ub4) 0, (ub4 *) 0, OCI_DEFAULT)) ||
        (status = OCIBindByName(ctx->stmt_flow7_t, &ctx->flow7_t_bind.srcaddr,
                    ctx->errhp, (text *) ":srcaddr", -1, (dvoid *) &ctx->flow_t_packet.srcaddr,
                    (sword) sizeof(ctx->flow_t_packet.srcaddr), SQLT_UIN, (dvoid *) 0, (ub2 *) 0,
                    (ub2 *) 0, (ub4) 0, (ub4 *) 0, OCI_DEFAULT)) ||
        (status = OCIBindByName(ctx->stmt_flow7_t, &ctx->flow7_t_bind.dstaddr,
                    ctx->errhp, (text *) ":dstaddr", -1, (dvoid *) &ctx->flow_t_packet.dstaddr,
                    (sword) sizeof(ctx->flow_t_packet.dstaddr), SQLT_UIN, (dvoid *) 0, (ub2 *) 0,
                    (ub2 *) 0, (ub4) 0, (ub4 *) 0, OCI_DEFAULT)) ||
        (status = OCIBindByName(ctx->stmt_flow7_t, &ctx->flow7_t_bind.nexthop,
                    ctx->errhp, (text *) ":nexthop", -1, (dvoid *) &ctx->flow_t_packet.nexthop,
                    (sword) sizeof(ctx->flow_t_packet.nexthop), SQLT_UIN, (dvoid *) 0, (ub2 *) 0,
                    (ub2 *) 0, (ub4) 0, (ub4 *) 0, OCI_DEFAULT)) ||
        (status = OCIBindByName(ctx->stmt_flow7_t, &ctx->flow7_t_bind.input_if,
                    ctx->errhp, (text *) ":input_if", -1, (dvoid *) &ctx->flow_t_packet.input_if,
                    (sword) sizeof(ctx->flow_t_packet.input_if), SQLT_UIN, (dvoid *) 0, (ub2 *) 0,
                    (ub2 *) 0, (ub4) 0, (ub4 *) 0, OCI_DEFAULT)) ||
        (status = OCIBindByName(ctx->stmt_flow7_t, &ctx->flow7_t_bind.output_if,
                    ctx->errhp, (text *) ":output_if", -1, (dvoid *) &ctx->flow_t_packet.output_if,
                    (sword) sizeof(ctx->flow_t_packet.output_if), SQLT_UIN, (dvoid *) 0, (ub2 *) 0,
                    (ub2 *) 0, (ub4) 0, (ub4 *) 0, OCI_DEFAULT)) ||
        (status = OCIBindByName(ctx->stmt_flow7_t, &ctx->flow7_t_bind.doctets,
                    ctx->errhp, (text *) ":doctets", -1, (dvoid *) &ctx->flow_t_packet.doctets,
                    (sword) sizeof(ctx->flow_t_packet.doctets), SQLT_UIN, (dvoid *) 0, (ub2 *) 0,
                    (ub2 *) 0, (ub4) 0, (ub4 *) 0, OCI_DEFAULT)) ||
        (status = OCIBindByName(ctx->stmt_flow7_t, &ctx->flow7_t_bind.dpkts,
                    ctx->errhp, (text *) ":dpkts", -1, (dvoid *) &ctx->flow_t_packet.dpkts,
                    (sword) sizeof(ctx->flow_t_packet.dpkts), SQLT_UIN, (dvoid *) 0, (ub2 *) 0,
                    (ub2 *) 0, (ub4) 0, (ub4 *) 0, OCI_DEFAULT)) ||
        (status = OCIBindByName(ctx->stmt_flow7_t, &ctx->flow7_t_bind.fuptime,
                    ctx->errhp, (text *) ":fuptime", -1, (dvoid *) &ctx->flow_t_packet.fuptime,
                    (sword) sizeof(ctx->flow_t_packet.fuptime), SQLT_UIN, (dvoid *) 0, (ub2 *) 0,
                    (ub2 *) 0, (ub4) 0, (ub4 *) 0, OCI_DEFAULT)) ||
        (status = OCIBindByName(ctx->stmt_flow7_t, &ctx->flow7_t_bind.luptime,
                    ctx->errhp, (text *) ":luptime", -1, (dvoid *) &ctx->flow_t_packet.luptime,
                    (sword) sizeof(ctx->flow_t_packet.luptime), SQLT_UIN, (dvoid *) 0, (ub2 *) 0,
                    (ub2 *) 0, (ub4) 0, (ub4 *) 0, OCI_DEFAULT)) ||
        (status = OCIBindByName(ctx->stmt_flow7_t, &ctx->flow7_t_bind.srcport,
                    ctx->errhp, (text *) ":srcport", -1, (dvoid *) &ctx->flow_t_packet.srcport,
                    (sword) sizeof(ctx->flow_t_packet.srcport), SQLT_UIN, (dvoid *) 0, (ub2 *) 0,
                    (ub2 *) 0, (ub4) 0, (ub4 *) 0, OCI_DEFAULT)) ||
        (status = OCIBindByName(ctx->stmt_flow7_t, &ctx->flow7_t_bind.dstport,
                    ctx->errhp, (text *) ":dstport", -1, (dvoid *) &ctx->flow_t_packet.dstport,
                    (sword) sizeof(ctx->flow_t_packet.dstport), SQLT_UIN, (dvoid *) 0, (ub2 *) 0,
                    (ub2 *) 0, (ub4) 0, (ub4 *) 0, OCI_DEFAULT)) ||
        (status = OCIBindByName(ctx->stmt_flow7_t, &ctx->flow7_t_bind.mode,
                    ctx->errhp, (text *) ":mode_id", -1, (dvoid *) &ctx->flow_t_packet.mode,
                    (sword) sizeof(ctx->flow_t_packet.mode), SQLT_INT, (dvoid *) 0, (ub2 *) 0,
                    (ub2 *) 0, (ub4) 0, (ub4 *) 0, OCI_DEFAULT)) ||
        (status = OCIBindByName(ctx->stmt_flow7_t, &ctx->flow7_t_bind.flags,
                    ctx->errhp, (text *) ":flags", -1, (dvoid *) &ctx->flow_t_packet.flags,
                    (sword) sizeof(ctx->flow_t_packet.flags), SQLT_INT, (dvoid *) 0, (ub2 *) 0,
                    (ub2 *) 0, (ub4) 0, (ub4 *) 0, OCI_DEFAULT)) ||
        (status = OCIBindByName(ctx->stmt_flow7_t, &ctx->flow7_t_bind.prot,
                    ctx->errhp, (text *) ":prot", -1, (dvoid *) &ctx->flow_t_packet.prot,
                    (sword) sizeof(ctx->flow_t_packet.prot), SQLT_INT, (dvoid *) 0, (ub2 *) 0,
                    (ub2 *) 0, (ub4) 0, (ub4 *) 0, OCI_DEFAULT)) ||
        (status = OCIBindByName(ctx->stmt_flow7_t, &ctx->flow7_t_bind.tos,
                    ctx->errhp, (text *) ":tos", -1, (dvoid *) &ctx->flow_t_packet.tos,
                    (sword) sizeof(ctx->flow_t_packet.tos), SQLT_INT, (dvoid *) 0, (ub2 *) 0,
                    (ub2 *) 0, (ub4) 0, (ub4 *) 0, OCI_DEFAULT)) ||
        (status = OCIBindByName(ctx->stmt_flow7_t, &ctx->flow7_t_bind.src_as,
                    ctx->errhp, (text *) ":src_as", -1, (dvoid *) &ctx->flow_t_packet.src_as,
                    (sword) sizeof(ctx->flow_t_packet.src_as), SQLT_UIN, (dvoid *) 0, (ub2 *) 0,
                    (ub2 *) 0, (ub4) 0, (ub4 *) 0, OCI_DEFAULT)) ||
        (status = OCIBindByName(ctx->stmt_flow7_t, &ctx->flow7_t_bind.dst_as,
                    ctx->errhp, (text *) ":dst_as", -1, (dvoid *) &ctx->flow_t_packet.dst_as,
                    (sword) sizeof(ctx->flow_t_packet.dst_as), SQLT_UIN, (dvoid *) 0, (ub2 *) 0,
                    (ub2 *) 0, (ub4) 0, (ub4 *) 0, OCI_DEFAULT)) ||
        (status = OCIBindByName(ctx->stmt_flow7_t, &ctx->flow7_t_bind.srcmask,
                    ctx->errhp, (text *) ":srcmask", -1, (dvoid *) &ctx->flow_t_packet.srcmask,
                    (sword) sizeof(ctx->flow_t_packet.srcmask), SQLT_UIN, (dvoid *) 0, (ub2 *) 0,
                    (ub2 *) 0, (ub4) 0, (ub4 *) 0, OCI_DEFAULT)) ||
        (status = OCIBindByName(ctx->stmt_flow7_t, &ctx->flow7_t_bind.dstmask,
                    ctx->errhp, (text *) ":dstmask", -1, (dvoid *) &ctx->flow_t_packet.dstmask,
                    (sword) sizeof(ctx->flow_t_packet.dstmask), SQLT_UIN, (dvoid *) 0, (ub2 *) 0,
                    (ub2 *) 0, (ub4) 0, (ub4 *) 0, OCI_DEFAULT)) ||
        (status = OCIBindByName(ctx->stmt_flow7_t, &ctx->flow7_t_bind.resv2,
                    ctx->errhp, (text *) ":resv2", -1, (dvoid *) &ctx->flow_t_packet.resv2,
                    (sword) sizeof(ctx->flow_t_packet.resv2), SQLT_UIN, (dvoid *) 0, (ub2 *) 0,
                    (ub2 *) 0, (ub4) 0, (ub4 *) 0, OCI_DEFAULT)) ||
        (status = OCIBindByName(ctx->stmt_flow7_t, &ctx->flow7_t_bind.router,
                    ctx->errhp, (text *) ":router", -1, (dvoid *) &ctx->flow_t_packet.router,
                    (sword) sizeof(ctx->flow_t_packet.router), SQLT_UIN, (dvoid *) 0, (ub2 *) 0,
                    (ub2 *) 0, (ub4) 0, (ub4 *) 0, OCI_DEFAULT)))
                    {
        syslog(LOG_ERR, "FAILED: BINDByName() for flow7 data");
        report_error(ctx->errhp);
                    }
}

/* begin transaction */
void begin(void *dbparam)
{
    thread_context *ctx = (thread_context *) dbparam;

    (void) OCITransStart(ctx->svchp, ctx->errhp, 60, (ub4)OCI_TRANS_NEW);
}

/* end transaction */
void commit(void *dbparam)
{
    thread_context *ctx = (thread_context *) dbparam;
    /* Commit the changes */
    (void) OCITransCommit(ctx->svchp, ctx->errhp, (ub4) 0);
}

/* inserts netflow packet's header into database */
void inserthdr(uint32_t address, void *flowhdr, void *dbparam)
{
    struct in_addr  in;
    char            str[256];
    char            addr[128];
    short           version = ntohs(*((short *)flowhdr));
    struct tm       *stm;
    time_t          ttm;

    thread_context *ctx = (thread_context *) dbparam;

    if (!ctx->connected) {
        _connect(ctx->config, dbparam);
        if (!ctx->connected) 
            return;
    }

    if (!flowhdr) {
        syslog(LOG_ERR, "NULL packet received");
        return;
    }

    in.s_addr = address;
    inet_ntop(AF_INET, &in, addr, sizeof(addr));
    /* TODO */
    /* change sizeof to constant INET_ADDRSTRLEN */

    ttm = time(NULL);
    stm = localtime(&ttm);

    switch (version) {
    case 1: {           /* netflow v1 */
        struct flowhdr1_t   *hdr = (struct flowhdr1_t *)flowhdr;
        snprintf(str, sizeof(str), "INSERT INTO %s "
            "(address, version_, count_, uptime, secs, nsecs, time_) "
            "VALUES(\'%s\', %d, %d, %lu, %lu, %lu, "
            "to_date(\'%04d-%02d-%02d %02d:%02d:%02d\', "
            "\'YYYY-MM-DD HH24:MI:SS\'))",
            ctx->dbheader, addr, version, ntohs(hdr->count), 
            (unsigned long)ntohl(hdr->uptime),
            (unsigned long)ntohl(hdr->secs), (unsigned long)ntohl(hdr->nsecs),
            stm->tm_year+1900, stm->tm_mon+1, stm->tm_mday,
            stm->tm_hour, stm->tm_min, stm->tm_sec
            );

        ctx->flowhdr_t_packet.address = address;
        ctx->flowhdr_t_packet.version = version;
        ctx->flowhdr_t_packet.count = ntohs(hdr->count);
        ctx->flowhdr_t_packet.uptime = (unsigned long)ntohl(hdr->uptime);
        ctx->flowhdr_t_packet.secs = (unsigned long)ntohl(hdr->secs);
        ctx->flowhdr_t_packet.nsecs = (unsigned long)ntohl(hdr->nsecs);

        if (OCIStmtExecute(ctx->svchp, ctx->stmt_flow1hdr_t, ctx->errhp,
                          (ub4) 1, (ub4) 0, (CONST OCISnapshot*) 0,
                          (OCISnapshot*) 0, (ub4) OCI_DEFAULT))
        {
            syslog(LOG_ERR, "FAILED: OCIStmtExecute() stmt_flow1hdr_t");
            report_error(ctx->errhp);
        }
        break;
    }
    case 5: {           /* netflow v5 */
        struct flowhdr5_t   *hdr = (struct flowhdr5_t *)flowhdr;
        snprintf(str, sizeof(str), "INSERT INTO %s "
            "(address, version_, count_, uptime, secs, nsecs, "
            "flowseq, engine, eng_id, sampling, time_) "
            "VALUES(\'%s\', %d, %d, %lu, %lu, %lu, %lu, %d, %d, "
            "%d, to_date(\'%04d-%02d-%02d %02d:%02d:%02d\', "
            "\'YYYY-MM-DD HH24:MI:SS\'))",
            ctx->dbheader, addr, version, ntohs(hdr->count),
            (unsigned long)ntohl(hdr->uptime),
            (unsigned long)ntohl(hdr->secs), (unsigned long)ntohl(hdr->nsecs),
            (unsigned long)ntohl(hdr->flowseq),
            hdr->engine, hdr->eng_id, hdr->sampling,
            stm->tm_year+1900, stm->tm_mon+1, stm->tm_mday,
            stm->tm_hour, stm->tm_min, stm->tm_sec
            );

        ctx->flowhdr_t_packet.address = address;
        ctx->flowhdr_t_packet.version = version;
        ctx->flowhdr_t_packet.count = ntohs(hdr->count);
        ctx->flowhdr_t_packet.uptime = (unsigned long)ntohl(hdr->uptime);
        ctx->flowhdr_t_packet.secs = (unsigned long)ntohl(hdr->secs);
        ctx->flowhdr_t_packet.nsecs = (unsigned long)ntohl(hdr->nsecs);
        ctx->flowhdr_t_packet.flowseq = (unsigned long)ntohl(hdr->flowseq);
        ctx->flowhdr_t_packet.engine = hdr->engine;
        ctx->flowhdr_t_packet.eng_id = hdr->eng_id;
        ctx->flowhdr_t_packet.sampling = hdr->sampling;

        if (OCIStmtExecute(ctx->svchp, ctx->stmt_flow5hdr_t, ctx->errhp,
                          (ub4) 1, (ub4) 0, (CONST OCISnapshot*) 0,
                          (OCISnapshot*) 0, (ub4) OCI_DEFAULT))
        {
            syslog(LOG_ERR, "FAILED: OCIStmtExecute() stmt_flow5hdr_t");
            report_error(ctx->errhp);
        }
        break;
    }
	case 7: {           /* netflow v7 */
        struct flowhdr7_t   *hdr = (struct flowhdr7_t *)flowhdr;
        snprintf(str, sizeof(str), "INSERT INTO %s "
            "(address, version_, count_, uptime, secs, nsecs, "
            "flowseq, resv, time_) "
            "VALUES(\'%s\', %u, %u, %lu, %lu, %lu, %lu, %lu,"
            "to_date(\'%04d-%02d-%02d %02d:%02d:%02d\', "
            "\'YYYY-MM-DD HH24:MI:SS\'))",
            ctx->dbheader, addr, version, ntohs(hdr->count),
            (unsigned long)ntohl(hdr->uptime), (unsigned long)ntohl(hdr->secs),
            (unsigned long)ntohl(hdr->nsecs),
            (unsigned long)ntohl(hdr->flowseq), (unsigned long)ntohl(hdr->resv),
            stm->tm_year+1900, stm->tm_mon+1, stm->tm_mday,
            stm->tm_hour, stm->tm_min, stm->tm_sec
            );

        ctx->flowhdr_t_packet.address = address;
        ctx->flowhdr_t_packet.version = version;
        ctx->flowhdr_t_packet.count = ntohs(hdr->count);
        ctx->flowhdr_t_packet.uptime = (unsigned long)ntohl(hdr->uptime);
        ctx->flowhdr_t_packet.secs = (unsigned long)ntohl(hdr->secs);
        ctx->flowhdr_t_packet.nsecs = (unsigned long)ntohl(hdr->nsecs);
        ctx->flowhdr_t_packet.flowseq = (unsigned long)ntohl(hdr->flowseq);
        ctx->flowhdr_t_packet.resv = (unsigned long)ntohl(hdr->resv);

        if (OCIStmtExecute(ctx->svchp, ctx->stmt_flow7hdr_t, ctx->errhp,
                          (ub4) 1, (ub4) 0, (CONST OCISnapshot*) 0,
                          (OCISnapshot*) 0, (ub4) OCI_DEFAULT))
        {
            syslog(LOG_ERR, "FAILED: OCIStmtExecute() stmt_flow7hdr_t");
            report_error(ctx->errhp);
        }
    }
    default:
        syslog(LOG_ERR, "Unknown version of netflow packet");
        return;
    }
}

/* inserts flow into database */
void insertflow(uint32_t address, void *flowhdr, void *flow, int num, void* dbparam)
{
//  struct in_addr  in;
//    char            addr[128];
    char            str[512];
    short           version = ntohs(*((short *)flowhdr));
    struct tm       *stm;
    time_t          ttm;
    unsigned int    uptime;
    unsigned int    luptime;
    struct flowhdr5_t   *hdr = (struct flowhdr5_t *)flowhdr;
    struct flow1_t      *fl1 = (struct flow1_t *)flow;
    struct flow5_t      *fl5 = (struct flow5_t *)flow;
    struct flow7_t      *fl7 = (struct flow7_t *)flow;
    int    i, j;
    int     coincidence_found = 0;

    thread_context *ctx = (thread_context *) dbparam;

    if (!ctx->connected) {
        _connect(ctx->config, dbparam);
        if (!ctx->connected) 
            return;
    }

    if (!flowhdr || !flow) {
        syslog(LOG_ERR, "NULL packet received");
        return;
    }

    // Need to delete this further
    // because now ip addresses are stored in 
    // database in uint32_t format
   // in.s_addr = address;
   // inet_ntop(AF_INET, &in, addr, sizeof(addr));

    if (ctx->ltime) 
        ttm = time(NULL);
    else {
        uptime = htonl((unsigned int)hdr->uptime);
        luptime = htonl((unsigned int)fl5->luptime);
        if (uptime > luptime) {
            ttm = htonl(hdr->secs) - (uptime - luptime)/1000;
        } else {
            ttm = htonl(hdr->secs) + (luptime - uptime)/1000;
        }
    }
    stm = localtime(&ttm);
    switch (version) {
    case 1:             /* netflow v1 */
        ctx->flow_t_packet.address = address;
        ctx->flow_t_packet.version = version;
        ctx->flow_t_packet.number = num;
        ctx->flow_t_packet.srcaddr = htonl((fl1->srcaddr[0] << 24) + 
                                            (fl1->srcaddr[1] << 16) +
                                            (fl1->srcaddr[2] << 8) + 
                                            fl1->srcaddr[3]);
        ctx->flow_t_packet.dstaddr = htonl((fl1->dstaddr[0] << 24) + 
                                            (fl1->dstaddr[1] << 16) +
                                            (fl1->dstaddr[2] << 8) + 
                                            fl1->dstaddr[3]);
        ctx->flow_t_packet.nexthop = htonl((fl1->nexthop[0] << 24) + 
                                            (fl1->nexthop[1] << 16) +
                                            (fl1->nexthop[2] << 8) + 
                                            fl1->nexthop[3]);
        ctx->flow_t_packet.input_if = htons(fl1->input_if);
        ctx->flow_t_packet.output_if = htons(fl1->output_if),
        ctx->flow_t_packet.doctets = (unsigned long) htonl(fl1->doctets);
        ctx->flow_t_packet.dpkts = (unsigned long) htonl(fl1->dpkts);
        ctx->flow_t_packet.fuptime = (unsigned long) htonl(fl1->fuptime);
        ctx->flow_t_packet.luptime = (unsigned long) htonl(fl1->luptime);
        ctx->flow_t_packet.srcport = htons(fl1->srcport);
        ctx->flow_t_packet.dstport = htons(fl1->dstport);
        ctx->flow_t_packet.resv1 = htons(fl1->resv1);
        ctx->flow_t_packet.flags = fl1->flags;
        ctx->flow_t_packet.prot = fl1->prot;
        ctx->flow_t_packet.tos = fl1->tos;
        ctx->flow_t_packet.tcp_retx_cnt = fl1->tcp_retx_cnt;
        ctx->flow_t_packet.tcp_retx_sec = fl1->tcp_retx_sec;
        ctx->flow_t_packet.tcp_misseq_cnt = fl1->tcp_misseq_cnt;
        ctx->flow_t_packet.resv2 = (unsigned long)htonl(fl1->resv2);

        if (OCIStmtExecute(ctx->svchp, ctx->stmt_flow1_t, ctx->errhp, (ub4) 1, (ub4) 0,
                          (CONST OCISnapshot*) 0, (OCISnapshot*) 0,
                          (ub4) OCI_DEFAULT))
        {
            syslog(LOG_ERR, "FAILED: OCIStmtExecute() stmt_flow1_t");
            report_error(ctx->errhp);
        }
        break;
    case 5:             /* netflow v5 */

        // aggregation snipet
        /* Convert current IP addresses */
/*        ip_src = htonl((fl5->srcaddr[0] << 24) + 
                       (fl5->srcaddr[1] << 16) +
                       (fl5->srcaddr[2] << 8) + 
                       fl5->srcaddr[3]);
        ip_dst = htonl((fl5->dstaddr[0] << 24) +
                       (fl5->dstaddr[1] << 16) +
                       (fl5->dstaddr[2] << 8) + 
                       fl5->dstaddr[3]);

        /* Check current ip with the list of exceptions *
            coincidence_found = 0;
            counting = IP_NO;
            /* Excluded networks *
            for (j = 0; (j < exc_net_count) && (coincidence_found == 0); j++) {
                if ((ip_src & exc_net[j].netmask.s_addr) == exc_net[j].network.s_addr) {
                    counting = IP_SOURCE;
                    coincidence_found = 1;
                }
                if ((ip_dst & exc_net[j].netmask.s_addr) == exc_net[j].network.s_addr) {
                    counting = IP_DEST;
                    coincidence_found = 1;
                }
            }
            /* Excluded addresses *
            for (j = 0; (j < exc_addr_count) && (coincidence_found == 0); j++) {
                if (ip_src == exc_addr[j]) {
                    counting = IP_SOURCE;
                    coincidence_found = 1;
                }
                if (ip_dst == exc_addr[j]) {
                    counting = IP_DEST;
                    coincidence_found = 1;
                }
            }
            /* Included networks *
            for (j = 0; (j < inc_net_count) && (coincidence_found == 0); j++) {
                if ((ip_src & inc_net[j].netmask.s_addr) == inc_net[j].network.s_addr) {
                    counting = IP_SOURCE;
                    coincidence_found = 1;
                }
                if ((ip_dst & inc_net[j].netmask.s_addr) == inc_net[j].network.s_addr) {
                    counting = IP_DEST;
                    coincidence_found = 1;
                }
            }
            /* Included addresses *
            for (j = 0; (j < inc_addr_count) && (coincidence_found == 0); j++) {
                if (ip_src == inc_addr[j]) {
                    counting = IP_SOURCE;
                    coincidence_found = 1;
                }
                if (ip_dst == inc_addr[j]) {
                    counting = IP_DEST;
                    coincidence_found = 1;
                }
            }

                /* If address was found then we don't need to find further *
                if (!coincidence_found) break;

				snprintf(str, sizeof(str), "INSERT INTO %s "
				"(src_addr, dst_addr, doctets, src_port, dst_port, prot) "
				"VALUES(\'%u.%u.%u.%u\', \'%u.%u.%u.%u\', %d, %d, %d, %d)",
                                        dbflows,
                                        fl5->srcaddr[0], fl5->srcaddr[1], fl5->srcaddr[2], fl5->srcaddr[3],
                                        fl5->dstaddr[0], fl5->dstaddr[1], fl5->dstaddr[2], fl5->dstaddr[3],
                                        (unsigned long) htonl(fl5->doctets),
                                        htons(fl5->srcport),
                                        htons(fl5->dstport),
                                        fl5->prot
				);*/
        ctx->flow_t_packet.address = address;
        ctx->flow_t_packet.version = version;
        ctx->flow_t_packet.number = (unsigned long)(num+htonl(hdr->flowseq));
        ctx->flow_t_packet.srcaddr = htonl((fl5->srcaddr[0] << 24) + 
                                            (fl5->srcaddr[1] << 16) +
                                            (fl5->srcaddr[2] << 8) + 
                                            fl5->srcaddr[3]);
        ctx->flow_t_packet.dstaddr = htonl((fl5->dstaddr[0] << 24) + 
                                            (fl5->dstaddr[1] << 16) +
                                            (fl5->dstaddr[2] << 8) + 
                                            fl5->dstaddr[3]);
        ctx->flow_t_packet.nexthop = htonl((fl5->nexthop[0] << 24) + 
                                            (fl5->nexthop[1] << 16) +
                                            (fl5->nexthop[2] << 8) + 
                                            fl5->nexthop[3]);
        ctx->flow_t_packet.input_if = htons(fl5->input_if);
        ctx->flow_t_packet.output_if = htons(fl5->output_if),
        ctx->flow_t_packet.doctets = (unsigned long) htonl(fl5->doctets);
        ctx->flow_t_packet.dpkts = (unsigned long) htonl(fl5->dpkts);
        ctx->flow_t_packet.fuptime = (unsigned long) htonl(fl5->fuptime);
        ctx->flow_t_packet.luptime = (unsigned long) htonl(fl5->luptime);
        ctx->flow_t_packet.srcport = htons(fl5->srcport);
        ctx->flow_t_packet.dstport = htons(fl5->dstport);
        ctx->flow_t_packet.resv1 = (unsigned char)fl5->resv1;
        ctx->flow_t_packet.flags = fl5->flags;
        ctx->flow_t_packet.prot = fl5->prot;
        ctx->flow_t_packet.tos = fl5->tos;
        ctx->flow_t_packet.srcmask = (unsigned char) fl5->srcmask;
        ctx->flow_t_packet.dstmask = (unsigned char) fl5->dstmask;

        if (OCIStmtExecute(ctx->svchp, ctx->stmt_flow5_t, ctx->errhp, (ub4) 1, (ub4) 0,
                          (CONST OCISnapshot*) 0, (OCISnapshot*) 0,
                          (ub4) OCI_DEFAULT))
        {
            syslog(LOG_ERR, "FAILED: OCIStmtExecute() stmt_flow5_t");
            report_error(ctx->errhp);
        }
        break;
    case 7:                 /* netflow v7 */
        ctx->flow_t_packet.address = address;
        ctx->flow_t_packet.version = version;
        ctx->flow_t_packet.number = (unsigned long)(num+htonl(hdr->flowseq));
        ctx->flow_t_packet.srcaddr = htonl((fl7->srcaddr[0] << 24) + 
                                            (fl7->srcaddr[1] << 16) +
                                            (fl7->srcaddr[2] << 8) + 
                                            fl7->srcaddr[3]);
        ctx->flow_t_packet.dstaddr = htonl((fl7->dstaddr[0] << 24) + 
                                            (fl7->dstaddr[1] << 16) +
                                            (fl7->dstaddr[2] << 8) + 
                                            fl7->dstaddr[3]);
        ctx->flow_t_packet.nexthop = htonl((fl7->nexthop[0] << 24) + 
                                            (fl7->nexthop[1] << 16) +
                                            (fl7->nexthop[2] << 8) + 
                                            fl7->nexthop[3]);
        ctx->flow_t_packet.input_if = htons(fl7->input_if);
        ctx->flow_t_packet.output_if = htons(fl7->output_if),
        ctx->flow_t_packet.doctets = (unsigned long) htonl(fl7->doctets);
        ctx->flow_t_packet.dpkts = (unsigned long) htonl(fl7->dpkts);
        ctx->flow_t_packet.fuptime = (unsigned long) htonl(fl7->fuptime);
        ctx->flow_t_packet.luptime = (unsigned long) htonl(fl7->luptime);
        ctx->flow_t_packet.srcport = htons(fl7->srcport);
        ctx->flow_t_packet.dstport = htons(fl7->dstport);
        ctx->flow_t_packet.mode = (unsigned char) fl7->mode;
        ctx->flow_t_packet.flags = (unsigned char) fl7->flags;
        ctx->flow_t_packet.prot = (unsigned char) fl7->prot;
        ctx->flow_t_packet.tos = (unsigned char) fl7->tos;
        ctx->flow_t_packet.src_as = htons(fl7->src_as);
        ctx->flow_t_packet.dst_as = htons(fl7->dst_as);
        ctx->flow_t_packet.srcmask = (unsigned char) fl7->srcmask;
        ctx->flow_t_packet.dstmask = (unsigned char) fl7->dstmask;
        ctx->flow_t_packet.resv2 = htons(fl7->resv2);
        ctx->flow_t_packet.router = htonl((fl7->router[0] << 24) + 
                                            (fl7->router[1] << 16) +
                                            (fl7->router[2] << 8) + 
                                            fl7->router[3]);

        if (OCIStmtExecute(ctx->svchp, ctx->stmt_flow7_t, ctx->errhp, (ub4) 1, (ub4) 0,
                          (CONST OCISnapshot*) 0, (OCISnapshot*) 0,
                          (ub4) OCI_DEFAULT))
        {
            syslog(LOG_ERR, "FAILED: OCIStmtExecute() stmt_flow7_t");
            report_error(ctx->errhp);
        }
        break;
    default:
        syslog(LOG_ERR, "Unknown version of netflow packet");
        return;
    }
}

/* insert packet */
void insertpkt(uint32_t address, void *packet, void *dbparam)
{
    struct flowhdr1_t   *hdr1 = (struct flowhdr1_t *)packet;
    short               version = ntohs(*((short *)packet));
    int                 i;
    void                *flow;
 
    inserthdr(address, packet, dbparam);

    for (i = 0; i < htons(hdr1->count); i++) {
        switch (version) {
        case 1:
            flow = packet + sizeof(struct flowhdr1_t) +
                sizeof(struct flow1_t)*i;
            break;
        case 5:
            flow = packet + sizeof(struct flowhdr5_t) +
                sizeof(struct flow5_t)*i;
            break;
        case 7:
            flow = packet + sizeof(struct flowhdr7_t) +
                sizeof(struct flow7_t)*i;
            break;
        }
        insertflow(address, packet, flow, i, dbparam);
    }
}
