/**************************************************************************
 * Copyright (c) 2008-2009, Vyacheslav Yurkov <uvv.mail@gmail.com>        *
 * All rights reserved.                                                   *
 *                                                                        *
 * Redistribution and use in source and binary forms, with or without     *
 * modification, are permitted provided that the following conditions     *
 * are met:                                                               *
 *                                                                        *
 * 1. Redistributions of source code must retain the above copyright      *
 *    notice, this list of conditions and the following disclaimer.       *
 * 2. Redistributions in binary form must reproduce the above copyright   *
 *    notice, this list of conditions and the following disclaimer in the *
 *    documentation and/or other materials provided with the distribution.*
 * 3. Neither the name of the Vyacheslav Yurkov nor the names of his      *
 *    contributors may be used to endorse or promote products derived     *
 *    from this software without specific prior written permission.       *
 *                                                                        *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS    *
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT      *
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR  *
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT   *
 * AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, *
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       *
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  *
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  *
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    *
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  *
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   *
 **************************************************************************/

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <netdb.h>
#include <errno.h>
#include <oci.h>

#include <libnnfc.h>

/* Constants for counting variable */
#define IP_NO           0
#define IP_SOURCE       1
#define IP_DEST         2

/* structure for all prepared statements */
typedef struct
{
    uint32_t address;
    unsigned short version;
    unsigned long number;
    uint32_t srcaddr;
    uint32_t dstaddr;
    uint32_t nexthop;
    unsigned short input_if;
    unsigned short output_if;
    unsigned long doctets;
    unsigned long dpkts;
    unsigned long fuptime;
    unsigned long luptime;
    unsigned short srcport;
    unsigned short dstport;
    unsigned char resv1;
    char flags;
    char prot;
    char tos;
    char mode;
    unsigned short src_as;
    unsigned short dst_as;
    unsigned char srcmask;
    unsigned char dstmask;
    char tcp_retx_cnt;
    char tcp_retx_sec;
    char tcp_misseq_cnt;
    uint32_t router;
    unsigned short resv2;
    char time_id[20];
} flow_for_oracle;

typedef struct
{
    uint32_t address;
    unsigned short version;
    unsigned short count;
    uint32_t uptime;
    uint32_t secs;
    uint32_t nsecs;
    uint32_t flowseq;
    char engine;
    char eng_id;
    short sampling;
    uint32_t resv;
    char time_id[20];
} header_for_oracle;

typedef struct
{
    OCIBind *address;
    OCIBind *version;
    OCIBind *count;
    OCIBind *uptime;
    OCIBind *secs;
    OCIBind *nsecs;
    OCIBind *time_id;
} flow1hdr_t_OCIBind;

typedef struct
{
    OCIBind *address;
    OCIBind *version;
    OCIBind *count;
    OCIBind *uptime;
    OCIBind *secs;
    OCIBind *nsecs;
    OCIBind *flowseq;
    OCIBind *engine;
    OCIBind *eng_id;
    OCIBind *sampling;
    OCIBind *time_id;
} flow5hdr_t_OCIBind;

typedef struct
{
    OCIBind *address;
    OCIBind *version;
    OCIBind *count;
    OCIBind *uptime;
    OCIBind *secs;
    OCIBind *nsecs;
    OCIBind *flowseq;
    OCIBind *resv;
    OCIBind *time_id;
} flow7hdr_t_OCIBind;

typedef struct
{
    OCIBind *address;
    OCIBind *version;
    OCIBind *number;
    OCIBind *srcaddr;
    OCIBind *dstaddr;
    OCIBind *nexthop;
    OCIBind *input_if;
    OCIBind *output_if;
    OCIBind *doctets;
    OCIBind *dpkts;
    OCIBind *fuptime;
    OCIBind *luptime;
    OCIBind *srcport;
    OCIBind *dstport;
    OCIBind *resv1;
    OCIBind *flags;
    OCIBind *prot;
    OCIBind *tos;
    OCIBind *tcp_retx_cnt;
    OCIBind *tcp_retx_sec;
    OCIBind *tcp_misseq_cnt;
    OCIBind *resv2;
    OCIBind *time_id;
} flow1_t_OCIBind;

typedef struct
{
    OCIBind *address;
    OCIBind *version;
    OCIBind *number;
    OCIBind *srcaddr;
    OCIBind *dstaddr;
    OCIBind *nexthop;
    OCIBind *input_if;
    OCIBind *output_if;
    OCIBind *doctets;
    OCIBind *dpkts;
    OCIBind *fuptime;
    OCIBind *luptime;
    OCIBind *srcport;
    OCIBind *dstport;
    OCIBind *resv1;
    OCIBind *flags;
    OCIBind *prot;
    OCIBind *tos;
    OCIBind *srcmask;
    OCIBind *dstmask;
    OCIBind *time_id;
} flow5_t_OCIBind;

typedef struct
{
    OCIBind *address;
    OCIBind *version;
    OCIBind *number;
    OCIBind *srcaddr;
    OCIBind *dstaddr;
    OCIBind *nexthop;
    OCIBind *input_if;
    OCIBind *output_if;
    OCIBind *doctets;
    OCIBind *dpkts;
    OCIBind *fuptime;
    OCIBind *luptime;
    OCIBind *srcport;
    OCIBind *dstport;
    OCIBind *mode;
    OCIBind *flags;
    OCIBind *prot;
    OCIBind *tos;
    OCIBind *src_as;
    OCIBind *dst_as;
    OCIBind *srcmask;
    OCIBind *dstmask;
    OCIBind *resv2;
    OCIBind *router;
    OCIBind *time_id;
} flow7_t_OCIBind;

/**
 * @brief IPV4 network in CIDR notation.
 */
typedef struct GNUNET_IPv4NetworkSet
{
    struct in_addr network;
    struct in_addr netmask;
} CIDRNetwork;

typedef struct
{
    //OCIEnv *envhp;
    OCIServer *srvhp;
    OCISvcCtx *svchp;
    OCISession *authp;
    OCIError *errhp;
    /*
     * Parameters for this module:
     * localtime - store local or remote time?
     *  dbuser - oracle user's name
     *  dbpass - oracle user's password
     *  dbname - oracle database
     *  dbheader - mysql table for netflow headers
     *  dbflows - mysql table for netflow packets
     */

    int     ltime;
    int		full, agg, accounting;
    char    *dbuser;
    char    *dbpass;
    char    *dbname;
    char    *dbheader;
    char    *dbflows;
    char    config[MAXPATHLEN];
    int     connected;
    uint32_t buff_ip[2000][2];
    uint16_t buff_port[2000][2];
    int      buff_traf[2000][2];
    uint32_t ip_src, ip_dst;
    int		counting;
    int     k, l; /* Buffer iteration */

    /* Included networks and addresses.
     * We need to collect only this traffic */
    CIDRNetwork *inc_net;
    uint32_t    *inc_addr;
    int         inc_net_count;
    int         inc_addr_count;
    /* Excluded networks and addresses.
     * If some subnets or/and addresses overlap
     * with included networks and addresses we 
     * exclude them because we don't need to
     * collect their traffic */
    CIDRNetwork *exc_net;
    uint32_t    *exc_addr;
    int         exc_net_count;
    int         exc_addr_count;

    /* insert queries */
    text flow1hdr_t_insert[1024];
    text flow5hdr_t_insert[1024];
    text flow7hdr_t_insert[1024];
    text flow1_t_insert[1024];
    text flow5_t_insert[1024];
    text flow7_t_insert[1024];
    /* statement handlers */
    OCIStmt *stmt_flow1hdr_t;
    OCIStmt *stmt_flow5hdr_t;
    OCIStmt *stmt_flow7hdr_t;
    OCIStmt *stmt_flow1_t;
    OCIStmt *stmt_flow5_t;
    OCIStmt *stmt_flow7_t;
    header_for_oracle flowhdr_t_packet;
    flow_for_oracle flow_t_packet;
    /* bind handlers */
    flow1hdr_t_OCIBind flow1hdr_t_bind;
    flow5hdr_t_OCIBind flow5hdr_t_bind;
    flow7hdr_t_OCIBind flow7hdr_t_bind;
    flow1_t_OCIBind flow1_t_bind;
    flow5_t_OCIBind flow5_t_bind;
    flow7_t_OCIBind flow7_t_bind;

} thread_context;

