/**************************************************************************
 * Copyright (c) 2004-2006, Andrew Klyachkin <andrewk@region73.ru>        *
 * All rights reserved.                                                   *
 *                                                                        *
 * Redistribution and use in source and binary forms, with or without     *
 * modification, are permitted provided that the following conditions     *
 * are met:                                                               *
 *                                                                        *
 * 1. Redistributions of source code must retain the above copyright      *
 *    notice, this list of conditions and the following disclaimer.       *
 * 2. Redistributions in binary form must reproduce the above copyright   *
 *    notice, this list of conditions and the following disclaimer in the *
 *    documentation and/or other materials provided with the distribution.*
 * 3. Neither the name of the Andrew Klyachkin nor the names of his       *
 *    contributors may be used to endorse or promote products derived     *
 *    from this software without specific prior written permission.       *
 *                                                                        *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS    *
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT      *
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR  *
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT   *
 * AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, *
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       *
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  *
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  *
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    *
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  *
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   *
 **************************************************************************/

#include "postgres.h"

void commit(void *);

void * _create_thread_context()
{
    thread_context *ctx = (thread_context*) malloc(sizeof(thread_context));

    ctx->conn = 0;
    //config[MAXPATHLEN] = {0};
    ctx->ltime = 1;
    ctx->dbhost = 0;
    ctx->dbuser = 0;
    ctx->dbpass = 0;
    ctx->dbname = 0;
    ctx->dbheader = 0;
    ctx->dbflows = 0;
    ctx->nonblock = 0;
    return ctx;
}

/* closes connection to database */
void _close(void *dbparam)
{
    thread_context *ctx = (thread_context *) dbparam;
    int             sock, rc;
    fd_set          rset;

    if (!ctx->conn) return;

    commit(dbparam);
    if (ctx->nonblock) {
        if ((sock = PQsocket(ctx->conn)) < 0) {
            return;
        }

        FD_ZERO(&rset);
        while ((rc = PQflush(ctx->conn)) != 0) {
            for (;;) {
                FD_SET(sock, &rset);
                n_select(sock+1, NULL, &rset, NULL, NULL);
                if (FD_ISSET(sock, &rset))
                    break;
            }
        }

        for (;;) {
            FD_SET(sock, &rset);
            n_select(sock+1, &rset, NULL, NULL, NULL);
            if (FD_ISSET(sock, &rset))
                break;
        }
    }
    PQfinish(ctx->conn);

    if (ctx) free(ctx);

    syslog(LOG_DEBUG, "postgres: connection to database closed");
}

/* establishes connection to database */
void _connect(char *fname, void *dbparam)
{
    char            str[128];
    struct command  *commands = 0;
    int             n = 0, i;
    PGresult                    *res;
    int                         sock;
    fd_set                      rset;
    PostgresPollingStatusType   rc;

    thread_context *ctx = (thread_context *) dbparam;
    if (!fname) return;

    if (!ctx->config[0]) {
        memset(&ctx->config, 0, MAXPATHLEN);
        strncpy(ctx->config, fname, MAXPATHLEN);
    }

    if (readcfg(ctx->config, &commands)) {
        syslog(LOG_ERR, "postgres: parse error!");
        return;
    }

    CFG_GETI(ctx->ltime, "localtime", commands);
    CFG_GETI(ctx->dbport, "dbport", commands);
    CFG_GETI(ctx->nonblock, "nonblock", commands);

    CFG_GETS(ctx->dbhost, "dbhost", commands);
    CFG_GETS(ctx->dbuser, "dbuser", commands);
    CFG_GETS(ctx->dbpass, "dbpass", commands);
    CFG_GETS(ctx->dbname, "dbname", commands);
    CFG_GETS(ctx->dbheader, "dbheader", commands);
    CFG_GETS(ctx->dbflows, "dbflows", commands);

    freecmds(commands);

    if (!ctx->dbport || !ctx->dbhost || !ctx->dbuser || !ctx->dbpass ||
           !ctx->dbname || !ctx->dbheader || !ctx->dbflows) {

        syslog(LOG_ERR, "postgres: required parameter missing ;)");
        return;
    }

    memset((void *)&str, 0, 128);
    snprintf(str, 128, "host=%s port=%d dbname=%s user=%s password=%s",
        ctx->dbhost, ctx->dbport, ctx->dbname, ctx->dbuser, ctx->dbpass);
    if (ctx->nonblock)
        ctx->conn = PQconnectStart(str);
    else 
        ctx->conn = PQconnectdb(str);
    if (!ctx->conn || PQstatus(ctx->conn) == CONNECTION_BAD) {
        ctx->conn = 0;
        syslog(LOG_ERR, "postgres: couldn't establish connection to %s\n", 
            ctx->dbhost);
        return;
    }
    PQsetClientEncoding(ctx->conn, "UTF8");

    if (ctx->nonblock) {
        // obtain the descriptor of the socket underlying the database connection
        if ((sock = PQsocket(ctx->conn)) < 0) {
            return;
        }

        FD_ZERO(&rset);
        while ((rc = PQconnectPoll(ctx->conn)) != PGRES_POLLING_OK) {
            switch (rc) {
            case PGRES_POLLING_READING:
                FD_SET(sock, &rset);
                // wait until the socket is ready to read
                for (;;) {
                    n_select(sock+1, &rset, NULL, NULL, NULL);
                    if (FD_ISSET(sock, &rset))
                        break;
                }
                break;
            case PGRES_POLLING_WRITING:
                FD_SET(sock, &rset);
                // wait until the socket is ready to write
                for (;;) {
                    n_select(sock+1, NULL, &rset, NULL, NULL);
                    if (FD_ISSET(sock, &rset))
                        break;
                }
                break;
            case PGRES_POLLING_FAILED:
                syslog(LOG_ERR, "postgres: connection procedure has failed");
                _close(dbparam);
                ctx->conn = 0;
                return;
			default:
                /* do nothing
                 * the connection has been successfully made
                 */ ;
            }
        }
	
        if (PQstatus(ctx->conn) == CONNECTION_BAD) {
            syslog(LOG_ERR, "postgres: connection failed");
            ctx->conn = 0;
            return;
        }

        while ((res = PQgetResult(ctx->conn)) != 0);
        PQclear(res);
        if (PQsetnonblocking(ctx->conn, 1) < 0) {
            syslog(LOG_ERR, "postgres: PQsetnonblocking() failed");
            ctx->nonblock = 0;
        }
    }

    free(commands);
        if (!PQisnonblocking(ctx->conn)) {
            syslog(LOG_ERR, "postgres: isnonblocking != 1");
            ctx->nonblock = 0;
        }


    /* Create SQL-insert statements */
    sprintf(ctx->flow1hdr_t_insert, "insert into %s (address, version_id, "
            "count_id, uptime, secs, nsecs) values ($1, $2, $3, $4, $5, $6)",
            ctx->dbheader);

    sprintf(ctx->flow5hdr_t_insert, "insert into %s (address, version_id, "
            "count_id, uptime, secs, nsecs, flowseq, engine, eng_id, sampling) "
            "values ($1, $2, $3, $4, $5, $6, $7, $8, $9, $10)", ctx->dbheader);

    sprintf(ctx->flow7hdr_t_insert, "insert into %s (address, version_id, "
            "count_id, uptime, secs, nsecs, flowseq, resv) values "
            "($1, $2, $3, $4, $5, $6, $7, $8)", ctx->dbheader);

    sprintf(ctx->flow1_t_insert, "insert into %s (address, version_id, number_id, "
                "srcaddr, dstaddr, nexthop, input_if, output_if, dpkts, doctets, "
                "fuptime, luptime, srcport, dstport, resv1, flags, prot, "
                "tos, tcp_retx_cnt, tcp_retx_sec, tcp_misseq_cnt, resv2) "
                "values ($1, $2, $3, $4, $5, $6, $7, $8, $9, $10, $11, "
                        "$12, $13, $14, $15, $16, $17, $18, $19, $20, "
                        "$21, $22)", ctx->dbflows);

     sprintf(ctx->flow5_t_insert, "insert into %s (address, version_id, number_id, "
                "srcaddr, dstaddr, nexthop, input_if, output_if, dpkts, doctets, "
                "fuptime, luptime, srcport, dstport, resv1, flags, prot, tos) "
                "values ($1, $2, $3, $4, $5, $6, $7, $8, $9, $10, "
                        "$11, $12, $13, $14, $15, $16, $17, $18)", ctx->dbflows);

    sprintf(ctx->flow7_t_insert, "insert into %s (address, version_id, number_id, "
                "srcaddr, dstaddr, nexthop, input_if, output_if, dpkts, doctets, "
                "fuptime, luptime, srcport, dstport, mode_id, flags, prot, "
                "tos, src_as, dst_as, resv2, router) "
                "values ($1, $2, $3, $4, $5, $6, $7, $8, $9, $10, $11, $12, "
                        "$13, $14, $15, $16, $17, $18, $19, $20, $21, $22)",
                        ctx->dbflows);

    /* Prepare the statements */
    if (PQresultStatus(res = PQprepare(ctx->conn, STMT_FLOW1HDR_T_INSERT,
                         ctx->flow1hdr_t_insert, STMT_FLOW1HDR_T_NPARAMS,
                         NULL)) != PGRES_COMMAND_OK)
    {
        syslog(LOG_ERR, "postgres: PQprepare(flow1hdr_t_insert) failed");
        syslog(LOG_ERR, PQerrorMessage(ctx->conn));
        PQclear(res);
    }

    if (PQresultStatus(res = PQprepare(ctx->conn, STMT_FLOW5HDR_T_INSERT,
                         ctx->flow5hdr_t_insert, STMT_FLOW5HDR_T_NPARAMS,
                         NULL)) != PGRES_COMMAND_OK)
    {
        syslog(LOG_ERR, "postgres: PQprepare(flow5hdr_t_insert) failed");
        syslog(LOG_ERR, PQerrorMessage(ctx->conn));
        PQclear(res);
    }

    if (PQresultStatus(res = PQprepare(ctx->conn, STMT_FLOW7HDR_T_INSERT,
                         ctx->flow7hdr_t_insert, STMT_FLOW7HDR_T_NPARAMS,
                         NULL)) != PGRES_COMMAND_OK)
    {
        syslog(LOG_ERR, "postgres: PQprepare(flow7hdr_t_insert) failed");
        syslog(LOG_ERR, PQerrorMessage(ctx->conn));
        PQclear(res);
    }

    if (PQresultStatus(res = PQprepare(ctx->conn, STMT_FLOW1_T_INSERT,
                         ctx->flow1_t_insert, STMT_FLOW1_T_NPARAMS,
                         NULL)) != PGRES_COMMAND_OK)
    {
        syslog(LOG_ERR, "postgres: PQprepare(flow1_t_insert) failed");
        syslog(LOG_ERR, PQerrorMessage(ctx->conn));
        PQclear(res);
    }

    if (PQresultStatus(res = PQprepare(ctx->conn, STMT_FLOW5_T_INSERT,
                         ctx->flow5_t_insert, STMT_FLOW5_T_NPARAMS,
                         NULL)) != PGRES_COMMAND_OK)
    {
        syslog(LOG_ERR, "postgres: PQprepare(flow5_t_insert) failed");
        syslog(LOG_ERR, PQerrorMessage(ctx->conn));
        PQclear(res);
    }

    if (PQresultStatus(res = PQprepare(ctx->conn, STMT_FLOW7_T_INSERT,
                         ctx->flow7_t_insert, STMT_FLOW7_T_NPARAMS,
                         NULL)) != PGRES_COMMAND_OK)
    {
        syslog(LOG_ERR, "postgres: PQprepare(flow7_t_insert) failed");
        syslog(LOG_ERR, PQerrorMessage(ctx->conn));
        PQclear(res);
    }

    /* Bind values for netflow v1 header */
    ctx->flow1hdr_t_values[0] = (char *) &ctx->flowhdr_t_packet.address;
    ctx->flow1hdr_t_values[1] = (char *) &ctx->flowhdr_t_packet.version;
    ctx->flow1hdr_t_values[2] = (char *) &ctx->flowhdr_t_packet.count;
    ctx->flow1hdr_t_values[3] = (char *) &ctx->flowhdr_t_packet.uptime;
    ctx->flow1hdr_t_values[4] = (char *) &ctx->flowhdr_t_packet.secs;
    ctx->flow1hdr_t_values[5] = (char *) &ctx->flowhdr_t_packet.nsecs;

    /* Bind values for netflow v5 header */
    ctx->flow5hdr_t_values[0] = (char *) &ctx->flowhdr_t_packet.address;
    ctx->flow5hdr_t_values[1] = (char *) &ctx->flowhdr_t_packet.version;
    ctx->flow5hdr_t_values[2] = (char *) &ctx->flowhdr_t_packet.count;
    ctx->flow5hdr_t_values[3] = (char *) &ctx->flowhdr_t_packet.uptime;
    ctx->flow5hdr_t_values[4] = (char *) &ctx->flowhdr_t_packet.secs;
    ctx->flow5hdr_t_values[5] = (char *) &ctx->flowhdr_t_packet.nsecs;
    ctx->flow5hdr_t_values[6] = (char *) &ctx->flowhdr_t_packet.flowseq;
    ctx->flow5hdr_t_values[7] = (char *) &ctx->flowhdr_t_packet.engine;
    ctx->flow5hdr_t_values[8] = (char *) &ctx->flowhdr_t_packet.eng_id;
    ctx->flow5hdr_t_values[9] = (char *) &ctx->flowhdr_t_packet.sampling;

    /* Bind values for netflow v7 header */
    ctx->flow7hdr_t_values[0] = (char *) &ctx->flowhdr_t_packet.address;
    ctx->flow7hdr_t_values[1] = (char *) &ctx->flowhdr_t_packet.version;
    ctx->flow7hdr_t_values[2] = (char *) &ctx->flowhdr_t_packet.count;
    ctx->flow7hdr_t_values[3] = (char *) &ctx->flowhdr_t_packet.uptime;
    ctx->flow7hdr_t_values[4] = (char *) &ctx->flowhdr_t_packet.secs;
    ctx->flow7hdr_t_values[5] = (char *) &ctx->flowhdr_t_packet.nsecs;
    ctx->flow7hdr_t_values[6] = (char *) &ctx->flowhdr_t_packet.flowseq;
    ctx->flow7hdr_t_values[7] = (char *) &ctx->flowhdr_t_packet.resv;

    /* Bind values for netflow v1 data */
    ctx->flow1_t_values[0] = (char *) &ctx->flow_t_packet.address;
    ctx->flow1_t_values[1] = (char *) &ctx->flow_t_packet.version;
    ctx->flow1_t_values[2] = (char *) &ctx->flow_t_packet.number;
    ctx->flow1_t_values[3] = (char *) &ctx->flow_t_packet.srcaddr;
    ctx->flow1_t_values[4] = (char *) &ctx->flow_t_packet.dstaddr;
    ctx->flow1_t_values[5] = (char *) &ctx->flow_t_packet.nexthop;
    ctx->flow1_t_values[6] = (char *) &ctx->flow_t_packet.input_if;
    ctx->flow1_t_values[7] = (char *) &ctx->flow_t_packet.output_if;
    ctx->flow1_t_values[8] = (char *) &ctx->flow_t_packet.dpkts;
    ctx->flow1_t_values[9] = (char *) &ctx->flow_t_packet.doctets;
    ctx->flow1_t_values[10] = (char *) &ctx->flow_t_packet.fuptime;
    ctx->flow1_t_values[11] = (char *) &ctx->flow_t_packet.luptime;
    ctx->flow1_t_values[12] = (char *) &ctx->flow_t_packet.srcport;
    ctx->flow1_t_values[13] = (char *) &ctx->flow_t_packet.dstport;
    ctx->flow1_t_values[14] = (char *) &ctx->flow_t_packet.resv1;
    ctx->flow1_t_values[15] = (char *) &ctx->flow_t_packet.flags;
    ctx->flow1_t_values[16] = (char *) &ctx->flow_t_packet.prot;
    ctx->flow1_t_values[17] = (char *) &ctx->flow_t_packet.tos;
    ctx->flow1_t_values[18] = (char *) &ctx->flow_t_packet.tcp_retx_cnt;
    ctx->flow1_t_values[19] = (char *) &ctx->flow_t_packet.tcp_retx_sec;
    ctx->flow1_t_values[20] = (char *) &ctx->flow_t_packet.tcp_misseq_cnt;
    ctx->flow1_t_values[21] = (char *) &ctx->flow_t_packet.resv2;

    /* Bind values for netflow v5 data */
    ctx->flow5_t_values[0] = (char *) &ctx->flow_t_packet.address;
    ctx->flow5_t_values[1] = (char *) &ctx->flow_t_packet.version;
    ctx->flow5_t_values[2] = (char *) &ctx->flow_t_packet.number;
    ctx->flow5_t_values[3] = (char *) &ctx->flow_t_packet.srcaddr;
    ctx->flow5_t_values[4] = (char *) &ctx->flow_t_packet.dstaddr;
    ctx->flow5_t_values[5] = (char *) &ctx->flow_t_packet.nexthop;
    ctx->flow5_t_values[6] = (char *) &ctx->flow_t_packet.input_if;
    ctx->flow5_t_values[7] = (char *) &ctx->flow_t_packet.output_if;
    ctx->flow5_t_values[8] = (char *) &ctx->flow_t_packet.dpkts;
    ctx->flow5_t_values[9] = (char *) &ctx->flow_t_packet.doctets;
    ctx->flow5_t_values[10] = (char *) &ctx->flow_t_packet.fuptime;
    ctx->flow5_t_values[11] = (char *) &ctx->flow_t_packet.luptime;
    ctx->flow5_t_values[12] = (char *) &ctx->flow_t_packet.srcport;
    ctx->flow5_t_values[13] = (char *) &ctx->flow_t_packet.dstport;
    ctx->flow5_t_values[14] = (char *) &ctx->flow_t_packet.resv1;
    ctx->flow5_t_values[15] = (char *) &ctx->flow_t_packet.flags;
    ctx->flow5_t_values[16] = (char *) &ctx->flow_t_packet.prot;
    ctx->flow5_t_values[17] = (char *) &ctx->flow_t_packet.tos;

    /* Bind values for netflow v7 data */
    ctx->flow7_t_values[0] = (char *) &ctx->flow_t_packet.address;
    ctx->flow7_t_values[1] = (char *) &ctx->flow_t_packet.version;
    ctx->flow7_t_values[2] = (char *) &ctx->flow_t_packet.number;
    ctx->flow7_t_values[3] = (char *) &ctx->flow_t_packet.srcaddr;
    ctx->flow7_t_values[4] = (char *) &ctx->flow_t_packet.dstaddr;
    ctx->flow7_t_values[5] = (char *) &ctx->flow_t_packet.nexthop;
    ctx->flow7_t_values[6] = (char *) &ctx->flow_t_packet.input_if;
    ctx->flow7_t_values[7] = (char *) &ctx->flow_t_packet.output_if;
    ctx->flow7_t_values[8] = (char *) &ctx->flow_t_packet.dpkts;
    ctx->flow7_t_values[9] = (char *) &ctx->flow_t_packet.doctets;
    ctx->flow7_t_values[10] = (char *) &ctx->flow_t_packet.fuptime;
    ctx->flow7_t_values[11] = (char *) &ctx->flow_t_packet.luptime;
    ctx->flow7_t_values[12] = (char *) &ctx->flow_t_packet.srcport;
    ctx->flow7_t_values[13] = (char *) &ctx->flow_t_packet.dstport;
    ctx->flow7_t_values[14] = (char *) &ctx->flow_t_packet.mode;
    ctx->flow7_t_values[15] = (char *) &ctx->flow_t_packet.flags;
    ctx->flow7_t_values[16] = (char *) &ctx->flow_t_packet.prot;
    ctx->flow7_t_values[17] = (char *) &ctx->flow_t_packet.tos;
    ctx->flow7_t_values[18] = (char *) &ctx->flow_t_packet.src_as;
    ctx->flow7_t_values[19] = (char *) &ctx->flow_t_packet.dst_as;
    ctx->flow7_t_values[20] = (char *) &ctx->flow_t_packet.resv2;
    ctx->flow7_t_values[21] = (char *) &ctx->flow_t_packet.router;

    /* Set lenght of the parameters */
    /* netflow v1 header */
    ctx->flow1hdr_t_paramlengths[0] = sizeof(ctx->flowhdr_t_packet.address);
    ctx->flow1hdr_t_paramlengths[1] = sizeof(ctx->flowhdr_t_packet.version);
    ctx->flow1hdr_t_paramlengths[2] = sizeof(ctx->flowhdr_t_packet.count);
    ctx->flow1hdr_t_paramlengths[3] = sizeof(ctx->flowhdr_t_packet.uptime);
    ctx->flow1hdr_t_paramlengths[4] = sizeof(ctx->flowhdr_t_packet.secs);
    ctx->flow1hdr_t_paramlengths[5] = sizeof(ctx->flowhdr_t_packet.nsecs);

    /* netflow v5 header */
    ctx->flow5hdr_t_paramlengths[0] = sizeof(ctx->flowhdr_t_packet.address);
    ctx->flow5hdr_t_paramlengths[1] = sizeof(ctx->flowhdr_t_packet.version);
    ctx->flow5hdr_t_paramlengths[2] = sizeof(ctx->flowhdr_t_packet.count);
    ctx->flow5hdr_t_paramlengths[3] = sizeof(ctx->flowhdr_t_packet.uptime);
    ctx->flow5hdr_t_paramlengths[4] = sizeof(ctx->flowhdr_t_packet.secs);
    ctx->flow5hdr_t_paramlengths[5] = sizeof(ctx->flowhdr_t_packet.nsecs);
    ctx->flow5hdr_t_paramlengths[6] = sizeof(ctx->flowhdr_t_packet.flowseq);
    ctx->flow5hdr_t_paramlengths[7] = sizeof(ctx->flowhdr_t_packet.engine);
    ctx->flow5hdr_t_paramlengths[8] = sizeof(ctx->flowhdr_t_packet.eng_id);
    ctx->flow5hdr_t_paramlengths[9] = sizeof(ctx->flowhdr_t_packet.sampling);

    /* netflow v7 header */
    ctx->flow7hdr_t_paramlengths[0] = sizeof(ctx->flowhdr_t_packet.address);
    ctx->flow7hdr_t_paramlengths[1] = sizeof(ctx->flowhdr_t_packet.version);
    ctx->flow7hdr_t_paramlengths[2] = sizeof(ctx->flowhdr_t_packet.count);
    ctx->flow7hdr_t_paramlengths[3] = sizeof(ctx->flowhdr_t_packet.uptime);
    ctx->flow7hdr_t_paramlengths[4] = sizeof(ctx->flowhdr_t_packet.secs);
    ctx->flow7hdr_t_paramlengths[5] = sizeof(ctx->flowhdr_t_packet.nsecs);
    ctx->flow7hdr_t_paramlengths[6] = sizeof(ctx->flowhdr_t_packet.flowseq);
    ctx->flow7hdr_t_paramlengths[7] = sizeof(ctx->flowhdr_t_packet.resv);

    /* netflow v1 data */
    ctx->flow1_t_paramlengths[0] = sizeof(ctx->flow_t_packet.address);
    ctx->flow1_t_paramlengths[1] = sizeof(ctx->flow_t_packet.version);
    ctx->flow1_t_paramlengths[2] = sizeof(ctx->flow_t_packet.number);
    ctx->flow1_t_paramlengths[3] = sizeof(ctx->flow_t_packet.srcaddr);
    ctx->flow1_t_paramlengths[4] = sizeof(ctx->flow_t_packet.dstaddr);
    ctx->flow1_t_paramlengths[5] = sizeof(ctx->flow_t_packet.nexthop);
    ctx->flow1_t_paramlengths[6] = sizeof(ctx->flow_t_packet.input_if);
    ctx->flow1_t_paramlengths[7] = sizeof(ctx->flow_t_packet.output_if);
    ctx->flow1_t_paramlengths[8] = sizeof(ctx->flow_t_packet.dpkts);
    ctx->flow1_t_paramlengths[9] = sizeof(ctx->flow_t_packet.doctets);
    ctx->flow1_t_paramlengths[10] = sizeof(ctx->flow_t_packet.fuptime);
    ctx->flow1_t_paramlengths[11] = sizeof(ctx->flow_t_packet.luptime);
    ctx->flow1_t_paramlengths[12] = sizeof(ctx->flow_t_packet.srcport);
    ctx->flow1_t_paramlengths[13] = sizeof(ctx->flow_t_packet.dstport);
    ctx->flow1_t_paramlengths[14] = sizeof(ctx->flow_t_packet.resv1);
    ctx->flow1_t_paramlengths[15] = sizeof(ctx->flow_t_packet.flags);
    ctx->flow1_t_paramlengths[16] = sizeof(ctx->flow_t_packet.prot);
    ctx->flow1_t_paramlengths[17] = sizeof(ctx->flow_t_packet.tos);
    ctx->flow1_t_paramlengths[18] = sizeof(ctx->flow_t_packet.tcp_retx_cnt);
    ctx->flow1_t_paramlengths[19] = sizeof(ctx->flow_t_packet.tcp_retx_sec);
    ctx->flow1_t_paramlengths[20] = sizeof(ctx->flow_t_packet.tcp_misseq_cnt);
    ctx->flow1_t_paramlengths[21] = sizeof(ctx->flow_t_packet.resv2);

    /* netflow v5 data */
    ctx->flow5_t_paramlengths[0] = sizeof(ctx->flow_t_packet.address);
    ctx->flow5_t_paramlengths[1] = sizeof(ctx->flow_t_packet.version);
    ctx->flow5_t_paramlengths[2] = sizeof(ctx->flow_t_packet.number);
    ctx->flow5_t_paramlengths[3] = sizeof(ctx->flow_t_packet.srcaddr);
    ctx->flow5_t_paramlengths[4] = sizeof(ctx->flow_t_packet.dstaddr);
    ctx->flow5_t_paramlengths[5] = sizeof(ctx->flow_t_packet.nexthop);
    ctx->flow5_t_paramlengths[6] = sizeof(ctx->flow_t_packet.input_if);
    ctx->flow5_t_paramlengths[7] = sizeof(ctx->flow_t_packet.output_if);
    ctx->flow5_t_paramlengths[8] = sizeof(ctx->flow_t_packet.dpkts);
    ctx->flow5_t_paramlengths[9] = sizeof(ctx->flow_t_packet.doctets);
    ctx->flow5_t_paramlengths[10] = sizeof(ctx->flow_t_packet.fuptime);
    ctx->flow5_t_paramlengths[11] = sizeof(ctx->flow_t_packet.luptime);
    ctx->flow5_t_paramlengths[12] = sizeof(ctx->flow_t_packet.srcport);
    ctx->flow5_t_paramlengths[13] = sizeof(ctx->flow_t_packet.dstport);
    ctx->flow5_t_paramlengths[14] = sizeof(ctx->flow_t_packet.resv1);
    ctx->flow5_t_paramlengths[15] = sizeof(ctx->flow_t_packet.flags);
    ctx->flow5_t_paramlengths[16] = sizeof(ctx->flow_t_packet.prot);
    ctx->flow5_t_paramlengths[17] = sizeof(ctx->flow_t_packet.tos);

    /* netflow v7 data */
    ctx->flow7_t_paramlengths[0] = sizeof(ctx->flow_t_packet.address);
    ctx->flow7_t_paramlengths[1] = sizeof(ctx->flow_t_packet.version);
    ctx->flow7_t_paramlengths[2] = sizeof(ctx->flow_t_packet.number);
    ctx->flow7_t_paramlengths[3] = sizeof(ctx->flow_t_packet.srcaddr);
    ctx->flow7_t_paramlengths[4] = sizeof(ctx->flow_t_packet.dstaddr);
    ctx->flow7_t_paramlengths[5] = sizeof(ctx->flow_t_packet.nexthop);
    ctx->flow7_t_paramlengths[6] = sizeof(ctx->flow_t_packet.input_if);
    ctx->flow7_t_paramlengths[7] = sizeof(ctx->flow_t_packet.output_if);
    ctx->flow7_t_paramlengths[8] = sizeof(ctx->flow_t_packet.dpkts);
    ctx->flow7_t_paramlengths[9] = sizeof(ctx->flow_t_packet.doctets);
    ctx->flow7_t_paramlengths[10] = sizeof(ctx->flow_t_packet.fuptime);
    ctx->flow7_t_paramlengths[11] = sizeof(ctx->flow_t_packet.luptime);
    ctx->flow7_t_paramlengths[12] = sizeof(ctx->flow_t_packet.srcport);
    ctx->flow7_t_paramlengths[13] = sizeof(ctx->flow_t_packet.dstport);
    ctx->flow7_t_paramlengths[14] = sizeof(ctx->flow_t_packet.mode);
    ctx->flow7_t_paramlengths[15] = sizeof(ctx->flow_t_packet.flags);
    ctx->flow7_t_paramlengths[16] = sizeof(ctx->flow_t_packet.prot);
    ctx->flow7_t_paramlengths[17] = sizeof(ctx->flow_t_packet.tos);
    ctx->flow7_t_paramlengths[18] = sizeof(ctx->flow_t_packet.src_as);
    ctx->flow7_t_paramlengths[19] = sizeof(ctx->flow_t_packet.dst_as);
    ctx->flow7_t_paramlengths[20] = sizeof(ctx->flow_t_packet.resv2);
    ctx->flow7_t_paramlengths[21] = sizeof(ctx->flow_t_packet.router);

    /* Set format for the parameters */
    /* netflow v1 header */
    ctx->flow1hdr_t_paramformats[0] = 0;
    ctx->flow1hdr_t_paramformats[1] = 1;
    ctx->flow1hdr_t_paramformats[2] = 1;
    ctx->flow1hdr_t_paramformats[3] = 0;
    ctx->flow1hdr_t_paramformats[4] = 0;
    ctx->flow1hdr_t_paramformats[5] = 0;

    /* netflow v5 header */
    ctx->flow5hdr_t_paramformats[0] = 0;
    ctx->flow5hdr_t_paramformats[1] = 1;
    ctx->flow5hdr_t_paramformats[2] = 1;
    ctx->flow5hdr_t_paramformats[3] = 0;
    ctx->flow5hdr_t_paramformats[4] = 0;
    ctx->flow5hdr_t_paramformats[5] = 0;
    ctx->flow5hdr_t_paramformats[6] = 0;
    ctx->flow5hdr_t_paramformats[7] = 1;
    ctx->flow5hdr_t_paramformats[8] = 1;
    ctx->flow5hdr_t_paramformats[9] = 1;

    /* netflow v7 header */
    ctx->flow7hdr_t_paramformats[0] = 0;
    ctx->flow7hdr_t_paramformats[1] = 1;
    ctx->flow7hdr_t_paramformats[2] = 1;
    ctx->flow7hdr_t_paramformats[3] = 0;
    ctx->flow7hdr_t_paramformats[4] = 0;
    ctx->flow7hdr_t_paramformats[5] = 0;
    ctx->flow7hdr_t_paramformats[6] = 0;
    ctx->flow7hdr_t_paramformats[7] = 0;

    /* netflow v1 data */
    ctx->flow1_t_paramformats[0] = 0;
    ctx->flow1_t_paramformats[1] = 1;
    ctx->flow1_t_paramformats[2] = 0;
    ctx->flow1_t_paramformats[3] = 0;
    ctx->flow1_t_paramformats[4] = 0;
    ctx->flow1_t_paramformats[5] = 0;
    ctx->flow1_t_paramformats[6] = 1;
    ctx->flow1_t_paramformats[7] = 1;
    ctx->flow1_t_paramformats[8] = 0;
    ctx->flow1_t_paramformats[9] = 0;
    ctx->flow1_t_paramformats[10] = 0;
    ctx->flow1_t_paramformats[11] = 0;
    ctx->flow1_t_paramformats[12] = 1;
    ctx->flow1_t_paramformats[13] = 1;
    ctx->flow1_t_paramformats[14] = 1;
    ctx->flow1_t_paramformats[15] = 1;
    ctx->flow1_t_paramformats[16] = 1;
    ctx->flow1_t_paramformats[17] = 1;
    ctx->flow1_t_paramformats[18] = 1;
    ctx->flow1_t_paramformats[19] = 1;
    ctx->flow1_t_paramformats[20] = 1;
    ctx->flow1_t_paramformats[21] = 1;

    /* netflow v5 data */
    ctx->flow5_t_paramformats[0] = 0;
    ctx->flow5_t_paramformats[1] = 1;
    ctx->flow5_t_paramformats[2] = 0;
    ctx->flow5_t_paramformats[3] = 0;
    ctx->flow5_t_paramformats[4] = 0;
    ctx->flow5_t_paramformats[5] = 0;
    ctx->flow5_t_paramformats[6] = 1;
    ctx->flow5_t_paramformats[7] = 1;
    ctx->flow5_t_paramformats[8] = 0;
    ctx->flow5_t_paramformats[9] = 0;
    ctx->flow5_t_paramformats[10] = 0;
    ctx->flow5_t_paramformats[11] = 0;
    ctx->flow5_t_paramformats[12] = 1;
    ctx->flow5_t_paramformats[13] = 1;
    ctx->flow5_t_paramformats[14] = 1;
    ctx->flow5_t_paramformats[15] = 1;
    ctx->flow5_t_paramformats[16] = 1;
    ctx->flow5_t_paramformats[17] = 1;

    /* netflow v7 data */
    ctx->flow7_t_paramformats[0] = 0;
    ctx->flow7_t_paramformats[1] = 1;
    ctx->flow7_t_paramformats[2] = 0;
    ctx->flow7_t_paramformats[3] = 0;
    ctx->flow7_t_paramformats[4] = 0;
    ctx->flow7_t_paramformats[5] = 0;
    ctx->flow7_t_paramformats[6] = 1;
    ctx->flow7_t_paramformats[7] = 1;
    ctx->flow7_t_paramformats[8] = 0;
    ctx->flow7_t_paramformats[9] = 0;
    ctx->flow7_t_paramformats[10] = 0;
    ctx->flow7_t_paramformats[11] = 0;
    ctx->flow7_t_paramformats[12] = 1;
    ctx->flow7_t_paramformats[13] = 1;
    ctx->flow7_t_paramformats[14] = 1;
    ctx->flow7_t_paramformats[15] = 1;
    ctx->flow7_t_paramformats[16] = 1;
    ctx->flow7_t_paramformats[17] = 1;
    ctx->flow7_t_paramformats[18] = 1;
    ctx->flow7_t_paramformats[19] = 1;
    ctx->flow7_t_paramformats[20] = 0;
    ctx->flow7_t_paramformats[21] = 1;
}

/* begins transaction */
void begin(void *dbparam)
{
    ExecStatusType              est;
    PGresult                    *res;

    thread_context *ctx = (thread_context *) dbparam;
    if (!ctx->conn)
        return;

    /*if (ctx->nonblock) {
        if (!PQsendQuery(ctx->conn, "BEGIN"))
        {
            syslog(LOG_ERR, "postgres: cannot send BEGIN");
            syslog(LOG_ERR, PQerrorMessage(ctx->conn));
        }
    } else {
      */  if (!(res = PQexec(ctx->conn, "BEGIN")) || 
            ((est = PQresultStatus(res)) != PGRES_COMMAND_OK)) {

            syslog(LOG_ERR, "postgres: %s", PQerrorMessage(ctx->conn));
        }
        PQclear(res);
//    }
}

/* ends transaction */
void commit(void *dbparam)
{
    PGresult        *res;
    ExecStatusType  est;

    thread_context *ctx = (thread_context *) dbparam;
    if (!ctx->conn)
        return;

    if (ctx->nonblock) {
        while ((res = PQgetResult(ctx->conn)) != 0);
        PQclear(res);
        if (!PQsendQuery(ctx->conn, "COMMIT"))
            syslog(LOG_ERR, "postgres: cannot send COMMIT");
    } else {
        if (!(res = PQexec(ctx->conn, "COMMIT")) || 
            ((est = PQresultStatus(res)) != PGRES_COMMAND_OK)) {

            syslog(LOG_ERR, "postgres: %s", PQerrorMessage(ctx->conn));
        }
        PQclear(res);
    }
}

/* inserts netflow packet's header into database */
void inserthdr(uint32_t address, void *flowhdr, void *dbparam)
{
    struct in_addr  in;
    char            str[256];
    char            addr[128];
    short           version = ntohs(*((short *)flowhdr));
    struct tm       *stm;
    time_t          ttm;
    PGresult        *rc;
    ExecStatusType  est;

    thread_context *ctx = (thread_context *) dbparam;
    if (!ctx->conn)
        return;

    if (!flowhdr) {
        syslog(LOG_ERR, "postgres: NULL packet received");
        return;
    }

    in.s_addr = address;
    inet_ntop(AF_INET, &in, addr, sizeof(addr));

    ttm = time(NULL);
    stm = localtime(&ttm);

    switch (version) {
    case 1: {           /* netflow v1 */
        struct flowhdr1_t   *hdr = (struct flowhdr1_t *)flowhdr;

        strcpy(ctx->flowhdr_t_packet.address, addr);
        ctx->flowhdr_t_packet.version = htonl((uint32_t) version);
        ctx->flowhdr_t_packet.count = htonl(ntohs(hdr->count));
        sprintf(ctx->flowhdr_t_packet.uptime, "%u", (unsigned long)ntohl(hdr->uptime));
        sprintf(ctx->flowhdr_t_packet.secs, "%u", (unsigned long)ntohl(hdr->secs));
        sprintf(ctx->flowhdr_t_packet.nsecs, "%u", (unsigned long)ntohl(hdr->nsecs));

        if (PQresultStatus(rc = PQexecPrepared(ctx->conn, STMT_FLOW1HDR_T_INSERT,
                                 STMT_FLOW1HDR_T_NPARAMS, ctx->flow1hdr_t_values,
                                 ctx->flow1hdr_t_paramlengths,
                                 ctx->flow1hdr_t_paramformats,
                                 0)) != PGRES_COMMAND_OK)
        {
            syslog(LOG_ERR, "postgres: PQexecPrepared(flow1hdr_t_insert) failed");
            syslog(LOG_ERR, PQerrorMessage(ctx->conn));
            PQclear(rc);
        }

        break;
    }
	case 5: {           /* netflow v5 */
        struct flowhdr5_t   *hdr = (struct flowhdr5_t *)flowhdr;

        strcpy(ctx->flowhdr_t_packet.address, addr);
        ctx->flowhdr_t_packet.version = htonl((uint32_t) version);
        ctx->flowhdr_t_packet.count = htonl(ntohs(hdr->count));
        sprintf(ctx->flowhdr_t_packet.uptime, "%u", (unsigned long)ntohl(hdr->uptime));
        sprintf(ctx->flowhdr_t_packet.secs, "%u", (unsigned long)ntohl(hdr->secs));
        sprintf(ctx->flowhdr_t_packet.nsecs, "%u", (unsigned long)ntohl(hdr->nsecs));
        sprintf(ctx->flowhdr_t_packet.flowseq, "%u", (unsigned long)ntohl(hdr->flowseq));
        ctx->flowhdr_t_packet.engine = htons((uint16_t) hdr->engine);
        ctx->flowhdr_t_packet.eng_id = htons((uint16_t) hdr->eng_id);
        ctx->flowhdr_t_packet.sampling = htonl((uint32_t) hdr->sampling);

        if (PQresultStatus(rc = PQexecPrepared(ctx->conn, STMT_FLOW5HDR_T_INSERT,
                                 STMT_FLOW5HDR_T_NPARAMS, ctx->flow5hdr_t_values,
                                 ctx->flow5hdr_t_paramlengths,
                                 ctx->flow5hdr_t_paramformats,
                                 0)) != PGRES_COMMAND_OK)
        {
            syslog(LOG_ERR, "postgres: PQexecPrepared(flow5hdr_t_insert) failed");
            syslog(LOG_ERR, PQerrorMessage(ctx->conn));
            PQclear(rc);
        }

        break;
    }
	case 7: {          /* netflow v7 */
        struct flowhdr7_t	*hdr = (struct flowhdr7_t *)flowhdr;

        strcpy(ctx->flowhdr_t_packet.address, addr);
        ctx->flowhdr_t_packet.version = htonl((uint32_t) version);
        ctx->flowhdr_t_packet.count = htonl(ntohs(hdr->count));
        sprintf(ctx->flowhdr_t_packet.uptime, "%u", (unsigned long)ntohl(hdr->uptime));
        sprintf(ctx->flowhdr_t_packet.secs, "%u", (unsigned long)ntohl(hdr->secs));
        sprintf(ctx->flowhdr_t_packet.nsecs, "%u", (unsigned long)ntohl(hdr->nsecs));
        sprintf(ctx->flowhdr_t_packet.flowseq, "%u", (unsigned long)ntohl(hdr->flowseq));
        sprintf(ctx->flowhdr_t_packet.resv, "%u", (unsigned long)ntohl(hdr->resv));

        if (PQresultStatus(rc = PQexecPrepared(ctx->conn, STMT_FLOW7HDR_T_INSERT,
                                 STMT_FLOW7HDR_T_NPARAMS, ctx->flow7hdr_t_values,
                                 ctx->flow7hdr_t_paramlengths,
                                 ctx->flow7hdr_t_paramformats,
                                 0)) != PGRES_COMMAND_OK)
        {
            syslog(LOG_ERR, "postgres: PQexecPrepared(flow7hdr_t_insert) failed");
            syslog(LOG_ERR, PQerrorMessage(ctx->conn));
            PQclear(rc);
        }

        break;
    }
    default:
        syslog(LOG_ERR, "postgres: unknown version of netflow packet");
        return;
    }
/*    if (ctx->nonblock) {
        while ((rc = PQgetResult(ctx->conn)) != 0);
        PQclear(rc);
        if (!PQsendQuery(ctx->conn, str)) {
            syslog(LOG_ERR, "postgres: %s", PQerrorMessage(ctx->conn));
        }
    } else {
        if (!(rc = PQexec(ctx->conn, str)) || 
            ((est = PQresultStatus(rc)) != PGRES_COMMAND_OK)) {

            syslog(LOG_ERR, "postgres: %s", PQerrorMessage(ctx->conn));
        }
        PQclear(rc);
    }
    */
}

/* inserts flow into database */
void insertflow(uint32_t address, void *flowhdr, void *flow, int num, void *dbparam)
{
    struct in_addr  in;
    char            addr[128];
    char            str[512];
    short           version = ntohs(*((short *)flowhdr));
    struct tm       *stm;
    time_t          ttm;
    unsigned int    uptime;
    unsigned int    luptime;
    PGresult        *rc;
    ExecStatusType  est;
    struct flowhdr5_t   *hdr = (struct flowhdr5_t *)flowhdr;
    struct flow1_t      *fl1 = (struct flow1_t *)flow;
    struct flow5_t      *fl5 = (struct flow5_t *)flow;
    struct flow7_t      *fl7 = (struct flow7_t *)flow;
    uint32_t        address_for_db;

    thread_context *ctx = (thread_context *) dbparam;
    if (!ctx->conn)
        return;

    if (!flowhdr || !flow) {
        syslog(LOG_ERR, "postgres: NULL packet received");
        return;
    }

    in.s_addr = address;
    inet_ntop(AF_INET, &in, addr, sizeof(addr));

    if (ctx->ltime) 
        ttm = time(NULL);
    else {
        uptime = htonl((unsigned int)hdr->uptime);
        luptime = htonl((unsigned int)fl5->luptime);

        if (uptime > luptime) {
            ttm = htonl(hdr->secs) - (uptime - luptime)/1000;
        } else {
            ttm = htonl(hdr->secs) + (luptime - uptime)/1000;
        }
    }
    stm = localtime(&ttm);

    switch (version) {
    case 1:             /* netflow v1 */
        strcpy(ctx->flow_t_packet.address, addr);
        ctx->flow_t_packet.version = htonl((uint32_t) version);
        sprintf(ctx->flow_t_packet.number, "%u", num);
        sprintf(ctx->flow_t_packet.srcaddr, "%u.%u.%u.%u",
               fl1->srcaddr[0], fl1->srcaddr[1],
               fl1->srcaddr[2], fl1->srcaddr[3]);
        sprintf(ctx->flow_t_packet.dstaddr, "%u.%u.%u.%u",
               fl1->dstaddr[0], fl1->dstaddr[1],
               fl1->dstaddr[2], fl1->dstaddr[3]);
        sprintf(ctx->flow_t_packet.nexthop, "%u.%u.%u.%u",
               fl1->nexthop[0], fl1->nexthop[1],
               fl1->nexthop[2], fl1->nexthop[3]);
        ctx->flow_t_packet.input_if = fl1->input_if;
        ctx->flow_t_packet.output_if = fl1->output_if;
        sprintf(ctx->flow_t_packet.doctets, "%u", (unsigned long) htonl(fl1->doctets));
        sprintf(ctx->flow_t_packet.dpkts, "%u", (unsigned long) htonl(fl1->dpkts));
        sprintf(ctx->flow_t_packet.fuptime, "%u", (unsigned long) htonl(fl1->fuptime));
        sprintf(ctx->flow_t_packet.luptime, "%u", (unsigned long) htonl(fl1->luptime));
        ctx->flow_t_packet.srcport = htonl(htons(fl1->srcport));
        ctx->flow_t_packet.dstport = htonl(htons(fl1->dstport));
        ctx->flow_t_packet.resv1 = htonl((unsigned char)fl1->resv1);
        ctx->flow_t_packet.flags = htons((unsigned char)fl1->flags);
        ctx->flow_t_packet.prot = htons((unsigned char)fl1->prot);
        ctx->flow_t_packet.tos = htons((unsigned char)fl1->tos);
        ctx->flow_t_packet.tcp_retx_cnt = htons((unsigned char)fl1->tcp_retx_cnt);
        ctx->flow_t_packet.tcp_retx_sec = htons((unsigned char)fl1->tcp_retx_sec);
        ctx->flow_t_packet.tcp_misseq_cnt = htons((unsigned char)fl1->tcp_misseq_cnt);
        ctx->flow_t_packet.resv2 = (uint32_t)fl1->resv2;

        if (PQresultStatus(rc = PQexecPrepared(ctx->conn, STMT_FLOW1_T_INSERT,
                                 STMT_FLOW1_T_NPARAMS, ctx->flow1_t_values,
                                 ctx->flow1_t_paramlengths,
                                 ctx->flow1_t_paramformats,
                                 0)) != PGRES_COMMAND_OK)
        {
            syslog(LOG_ERR, "postgres: PQexecPrepared(flow1_t_insert) failed");
            syslog(LOG_ERR, PQerrorMessage(ctx->conn));
            PQclear(rc);
        }

        break;
    case 5:                 /* netflow v5 */
        strcpy(ctx->flow_t_packet.address, addr);
        ctx->flow_t_packet.version = htonl((uint32_t) version);
        sprintf(ctx->flow_t_packet.number, "%u", (unsigned long)(num+htonl(hdr->flowseq)));
        sprintf(ctx->flow_t_packet.srcaddr, "%u.%u.%u.%u/%u",
               fl5->srcaddr[0], fl5->srcaddr[1],
               fl5->srcaddr[2], fl5->srcaddr[3], fl5->srcmask);
        sprintf(ctx->flow_t_packet.dstaddr, "%u.%u.%u.%u/%u",
               fl5->dstaddr[0], fl5->dstaddr[1],
               fl5->dstaddr[2], fl5->dstaddr[3], fl5->dstmask);
        sprintf(ctx->flow_t_packet.nexthop, "%u.%u.%u.%u",
               fl5->nexthop[0], fl5->nexthop[1],
               fl5->nexthop[2], fl5->nexthop[3]);
        ctx->flow_t_packet.input_if = fl5->input_if;
        ctx->flow_t_packet.output_if = fl5->output_if;
        sprintf(ctx->flow_t_packet.doctets, "%u", (unsigned long) htonl(fl5->doctets));
        sprintf(ctx->flow_t_packet.dpkts, "%u", (unsigned long) htonl(fl5->dpkts));
        sprintf(ctx->flow_t_packet.fuptime, "%u", (unsigned long) htonl(fl5->fuptime));
        sprintf(ctx->flow_t_packet.luptime, "%u", (unsigned long) htonl(fl5->luptime));
        ctx->flow_t_packet.srcport = htonl(htons(fl5->srcport));
        ctx->flow_t_packet.dstport = htonl(htons(fl5->dstport));
        ctx->flow_t_packet.resv1 = htonl((unsigned char)fl5->resv1);
        ctx->flow_t_packet.flags = htons((unsigned char)fl5->flags);
        ctx->flow_t_packet.prot = htons((unsigned char)fl5->prot);
        ctx->flow_t_packet.tos = htons((unsigned char)fl5->tos);
            
        if (PQresultStatus(rc = PQexecPrepared(ctx->conn, STMT_FLOW5_T_INSERT,
                                 STMT_FLOW5_T_NPARAMS, ctx->flow5_t_values,
                                 ctx->flow5_t_paramlengths,
                                 ctx->flow5_t_paramformats,
                                 0)) != PGRES_COMMAND_OK)
        {
            syslog(LOG_ERR, "postgres: PQexecPrepared(flow5_t_insert) failed");
            syslog(LOG_ERR, PQerrorMessage(ctx->conn));
            PQclear(rc);
        }

        break;
    case 7:             /* netflow v7 */
        strcpy(ctx->flow_t_packet.address, addr);
        ctx->flow_t_packet.version = htonl((uint32_t) version);
        sprintf(ctx->flow_t_packet.number, "%u", (unsigned long)(num+htonl(hdr->flowseq)));
        sprintf(ctx->flow_t_packet.srcaddr, "%u.%u.%u.%u/%u",
               fl7->srcaddr[0], fl7->srcaddr[1],
               fl7->srcaddr[2], fl7->srcaddr[3], fl7->srcmask);
        sprintf(ctx->flow_t_packet.dstaddr, "%u.%u.%u.%u/%u",
               fl7->dstaddr[0], fl7->dstaddr[1],
               fl7->dstaddr[2], fl7->dstaddr[3], fl7->dstmask);
        sprintf(ctx->flow_t_packet.nexthop, "%u.%u.%u.%u",
               fl7->nexthop[0], fl7->nexthop[1],
               fl7->nexthop[2], fl7->nexthop[3]);
        ctx->flow_t_packet.input_if = fl7->input_if;
        ctx->flow_t_packet.output_if = fl7->output_if;
        sprintf(ctx->flow_t_packet.doctets, "%u", (unsigned long) htonl(fl7->doctets));
        sprintf(ctx->flow_t_packet.dpkts, "%u", (unsigned long) htonl(fl7->dpkts));
        sprintf(ctx->flow_t_packet.fuptime, "%u", (unsigned long) htonl(fl7->fuptime));
        sprintf(ctx->flow_t_packet.luptime, "%u", (unsigned long) htonl(fl7->luptime));
        ctx->flow_t_packet.srcport = htonl(htons(fl7->srcport));
        ctx->flow_t_packet.dstport = htonl(htons(fl7->dstport));
        ctx->flow_t_packet.mode = htons((unsigned char)fl7->mode);
        ctx->flow_t_packet.flags = htons((unsigned char)fl7->flags);
        ctx->flow_t_packet.prot = htons((unsigned char)fl7->prot);
        ctx->flow_t_packet.tos = htons((unsigned char)fl7->tos);
        sprintf(ctx->flow_t_packet.router, "%u.%u.%u.%u",
               fl7->router[0], fl7->router[1],
               fl7->router[2], fl7->router[3]);
        ctx->flow_t_packet.resv2 = htonl(htons(fl7->resv2));

        if (PQresultStatus(rc = PQexecPrepared(ctx->conn, STMT_FLOW7_T_INSERT,
                                 STMT_FLOW7_T_NPARAMS, ctx->flow7_t_values,
                                 ctx->flow7_t_paramlengths,
                                 ctx->flow7_t_paramformats,
                                 0)) != PGRES_COMMAND_OK)
        {
            syslog(LOG_ERR, "postgres: PQexecPrepared(flow7_t_insert) failed");
            syslog(LOG_ERR, PQerrorMessage(ctx->conn));
            PQclear(rc);
        }

        break;
	default:
        syslog(LOG_ERR, "postgres: unknown version of netflow packet");
        return;
    }
    /*
    if (ctx->nonblock) {
        while ((rc = PQgetResult(ctx->conn)) != 0);
        PQclear(rc);
        if (!PQsendQuery(ctx->conn, str)) {
            syslog(LOG_ERR, "postgres: %s", PQerrorMessage(ctx->conn));
        }
    } else {
        if (!(rc = PQexec(ctx->conn, str)) ||
            ((est = PQresultStatus(rc)) != PGRES_COMMAND_OK)) {

            syslog(LOG_ERR, "postgres: %s", PQerrorMessage(ctx->conn));
        }
        PQclear(rc);
    }
    */
}

void insertpkt(uint32_t address, void *packet, void *dbparam)
{
    struct flowhdr1_t   *hdr1 = (struct flowhdr1_t *)packet;
    short               version = ntohs(*((short *)packet));
    int                 i;
    void                *flow;

    thread_context *ctx = (thread_context *) dbparam;

    if (!packet) return;

    if (!ctx->conn) return;

    inserthdr(address, packet, dbparam);

    for (i = 0; i < htons(hdr1->count); i++) {
        switch (version) {
        case 1:
            flow = packet + sizeof(struct flowhdr1_t) +
                sizeof(struct flow1_t)*i;
            break;
        case 5:
            flow = packet + sizeof(struct flowhdr5_t) +
                sizeof(struct flow5_t)*i;
            break;
        case 7:
            flow = packet + sizeof(struct flowhdr7_t) +
                sizeof(struct flow7_t)*i;
            break;
        }
        insertflow(address, packet, flow, i, dbparam);
    }
}
