/**************************************************************************
 * Copyright (c) 2004-2006, Andrew Klyachkin <andrewk@region73.ru>        *
 * All rights reserved.                                                   *
 *                                                                        *
 * Redistribution and use in source and binary forms, with or without     *
 * modification, are permitted provided that the following conditions     *
 * are met:                                                               *
 *                                                                        *
 * 1. Redistributions of source code must retain the above copyright      *
 *    notice, this list of conditions and the following disclaimer.       *
 * 2. Redistributions in binary form must reproduce the above copyright   *
 *    notice, this list of conditions and the following disclaimer in the *
 *    documentation and/or other materials provided with the distribution.*
 * 3. Neither the name of the Andrew Klyachkin nor the names of his       *
 *    contributors may be used to endorse or promote products derived     *
 *    from this software without specific prior written permission.       *
 *                                                                        *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS    *
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT      *
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR  *
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT   *
 * AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, *
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       *
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  *
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  *
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    *
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  *
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   *
 **************************************************************************/

#include <sys/types.h>
#include <sys/socket.h>
#include <syslog.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/time.h>
#include <time.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <netdb.h>
#include <errno.h>

#include <libnnfc.h>

/*
 *  Parameters:
 *
 *  receivers - IP addreses of netflow collectors, which will receive
 *      redirected flows
 */

uint32_t            *rcv = 0;
int                 rcv_count = 0;
int                 *sock = 0;
int                 port = 9996;
struct sockaddr_in  *sa = 0;

void * _create_thread_context()
{
}

void _connect(char *fname, void *param)
{
    struct command  *commands = 0;
    int             n = 0;
    struct in_addr  addr;
    struct hostent  *he = 0;
    struct value    *v = 0;

	if (!fname) return;

    if (readcfg(fname, &commands)) {
        syslog(LOG_ERR, "redirect: parse error!");
        return;
    }

    CFG_GETI(port, "port", commands);

    v = getallvalues("receivers", commands);
    rcv_count = 0;
    while (v) {
        if (inet_pton(AF_INET, v->value, &addr) <= 0) {
            if ((he = gethostbyname(v->value)) == 0) {
                syslog(LOG_ERR, "redirect: invalid address "
                    "in parameter receivers");
                return;
            }
            addr = *(struct in_addr *)(he->h_addr);
        }
        if (!rcv) {
            rcv = (uint32_t *)calloc(1, sizeof(uint32_t));
        } else {
            rcv = (uint32_t *)realloc(rcv, sizeof(uint32_t)*(rcv_count+1));
        }
        rcv[rcv_count] = addr.s_addr;
        rcv_count++;
        v = v->next;
    }

    freecmds(commands);

    if (!rcv) {
        syslog(LOG_ERR, "redirect: no receivers specified in config");
        return;
    }

    sock = calloc(rcv_count, sizeof(int));
    sa = calloc(rcv_count, sizeof(struct sockaddr_in));

    for (n = 0; n < rcv_count; n++) {
        sa[n].sin_family = AF_INET;
        sa[n].sin_addr.s_addr = rcv[n];
        sa[n].sin_port = htons(port);

        sock[n] = socket(AF_INET, SOCK_DGRAM, 0);
        if (sock[n] < 0) {
            syslog(LOG_ERR, "redirect: %s", strerror(errno));
        }
    }
}

void _close(void *param)
{
    if (rcv) free(rcv);
}

void begin(void *param)
{
}

void commit(void *param)
{
}

void insertpkt(uint32_t address, void *packet, void *param)
{
    int                 len = 0, i = 0, n = 0;
    short               version = ntohs(*((short *)packet));
    struct flowhdr1_t   *hdr = (struct flowhdr1_t *)packet;

    if (!rcv) return;
    if (!packet) return;

    switch (version) {
    case 1:
        len = sizeof(struct flowhdr1_t) + 
            ntohs(hdr->count) * sizeof(struct flow1_t);
        break;
    case 5:
        len = sizeof(struct flowhdr5_t) + 
            ntohs(hdr->count) * sizeof(struct flow5_t);
        break;
    case 7:
        len = sizeof(struct flowhdr7_t) + 
            ntohs(hdr->count) * sizeof(struct flow7_t);
        break;
    default:
        syslog(LOG_ERR, "redirect: unknown netflow version %d", version);
        return;
    }

    for (i = 0; i < rcv_count; i++) {
        struct sockaddr_in  sa1 = sa[i];
        int                 r = sizeof(sa1);
		int                 s = sock[i];
        n = sendto(s, packet, len, 0, (struct sockaddr *)&sa1, r);
        if (n < len) {
            syslog(LOG_ERR, "redirect: %s", strerror(errno));
        }
    }
}
