/**************************************************************************
 * Copyright (c) 2004-2006, Andrew Klyachkin <andrewk@region73.ru>        *
 * All rights reserved.                                                   *
 *                                                                        *
 * Redistribution and use in source and binary forms, with or without     *
 * modification, are permitted provided that the following conditions     *
 * are met:                                                               *
 *                                                                        *
 * 1. Redistributions of source code must retain the above copyright      *
 *    notice, this list of conditions and the following disclaimer.       *
 * 2. Redistributions in binary form must reproduce the above copyright   *
 *    notice, this list of conditions and the following disclaimer in the *
 *    documentation and/or other materials provided with the distribution.*
 * 3. Neither the name of the Andrew Klyachkin nor the names of his       *
 *    contributors may be used to endorse or promote products derived     *
 *    from this software without specific prior written permission.       *
 *                                                                        *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS    *
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT      *
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR  *
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT   *
 * AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, *
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       *
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  *
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  *
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    *
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  *
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   *
 **************************************************************************/

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/param.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <string.h>

#include "nnfc.h"

struct options options = { port: 9996, 
    verbose: 0,
    uid: "nobody",
    gid: "nogroup",
    host: 0,
    config: "nnfc.cfg",
    localtime: 0,
    daemon: 0,
    pid: "nnfc.pid",
    stat: "nnfc.stat"
};

void usage()
{
    printf("new netflow collector\n");
    printf("Usage:\n");
    printf("\tnnfc [arguments]\n");
    printf("Arguments are:\n");
    printf("-h | -?\t\tthis help\n");
    printf("-v\t\tincrease verbosity level\n");
    printf("-p <port>\tport to bind (default: %d)\n", options.port);
    printf("-u <username>\tset username (default: %s)\n", options.uid);
    printf("-g <username>\tset groupname (default: %s)\n", options.gid);
    printf("-b <ip-addr>\taddress to bind (default: any)\n");
    printf("-f <config>\tconfig file (default: %s)\n", options.config);
    printf("-d\t\tdaemonize process\n");
    exit(-1);
}
 
void parse_arg(int argc, char *argv[])
{
    const char      *optstring = "h?p:vu:g:b:f:i:d";
    int             ch;
    struct in_addr  addr;
    struct hostent  *he;

    while ((ch = getopt(argc, argv, optstring)) != -1) {
        switch (ch) {
        case 'd': 
            options.daemon = 1;
            DEBUG(1, "daemon set to 1\n");
            break;
        case 'v': 
            options.verbose++;
            DEBUG(1, "verbosity level set to %d\n", options.verbose);
            break;
        case 'p': 
            options.port = atoi(optarg);
            DEBUG(1, "port set to %d\n", options.port);
            break;
		case 'u': 
            if (strlen(optarg) > 29) {
                fprintf(stderr, "ERROR: Too long username\n");
                exit(-1);
            } else {
                strncpy(options.uid, optarg, 30);
                DEBUG(1, "uid set to %s\n", options.uid);
            }
            break;
		case 'g': 
            if (strlen(optarg) > 29) {
                fprintf(stderr, "ERROR: Too long groupname\n");
                exit(-1);
            } else {
                strncpy(options.gid, optarg, 30);
                DEBUG(1, "gid set to %s\n", options.gid);
            }
            break;
		case 'b': 
            if (inet_pton(AF_INET, optarg, &addr) <= 0) {
                DEBUG(2, "may be it is hostname?\n");
                if ((he = gethostbyname(optarg)) == NULL) {
                    fprintf(stderr, "ERROR: Invalid address specified\n");
                    exit(-1);
                }
                addr = *(struct in_addr *)(he->h_addr);
            }
            options.host = addr.s_addr;
            if (options.verbose > 1) {
                bzero(&addr, sizeof(addr));
                addr.s_addr = options.host;
                fprintf(stderr, "host set to %s\n", 
                    inet_ntop(AF_INET, &addr, optarg, strlen(optarg)));
            }
            break;
        case 'f': 
            if (strlen(optarg) > MAXPATHLEN) {
                fprintf(stderr, "ERROR: Too long config file name\n");
                exit(-1);
            }
            strncpy(options.config, optarg, MAXPATHLEN);
            DEBUG(1, "config file set to %s\n", options.config);
            break;
        case 'h':
        case '?':
        default:
            usage();
        }
    }
    argc -= optind;
    argv += optind;
}
