/**************************************************************************
 * Copyright (c) 2004-2006, Andrew Klyachkin <andrewk@region73.ru>        *
 * All rights reserved.                                                   *
 *                                                                        *
 * Redistribution and use in source and binary forms, with or without     *
 * modification, are permitted provided that the following conditions     *
 * are met:                                                               *
 *                                                                        *
 * 1. Redistributions of source code must retain the above copyright      *
 *    notice, this list of conditions and the following disclaimer.       *
 * 2. Redistributions in binary form must reproduce the above copyright   *
 *    notice, this list of conditions and the following disclaimer in the *
 *    documentation and/or other materials provided with the distribution.*
 * 3. Neither the name of the Andrew Klyachkin nor the names of his       *
 *    contributors may be used to endorse or promote products derived     *
 *    from this software without specific prior written permission.       *
 *                                                                        *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS    *
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT      *
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR  *
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT   *
 * AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, *
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       *
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  *
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  *
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    *
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  *
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   *
 **************************************************************************/

#include <sys/types.h>
#include <sys/socket.h>
#include <stdlib.h>
#include <stdio.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <string.h>

#include "nnfc.h"

uint32_t    *addresses = NULL;
int         addr_count = 0;

void readconfig()
{
    struct command      *commands = 0;
    struct value        *v = 0;
    struct in_addr      addr;
    struct hostent      *he;
    struct nnfc_module  *m = NULL;
    int                 i;

    if (readcfg(options.config, &commands)) {
        fprintf(stderr, "Parse error!\n");
        exit(-1);
    }

    if (options.verbose > 4)
        printconfig(commands);
	
    if (getvalue("daemonize", commands)) {
        options.daemon = atoi(getvalue("daemonize", commands));
        DEBUG(1, "daemon set to %d\n", options.daemon);
    }
    if (getvalue("verbose", commands)) {
        options.verbose = atoi(getvalue("verbose", commands));
        DEBUG(1, "verbose level set to %d\n", options.verbose);
    }
    if (getvalue("port", commands)) {
        options.port = atoi(getvalue("port", commands));
        DEBUG(1, "port set to %d\n", options.port);
    }
    if (getvalue("pidfile", commands)) {
        strncpy(options.pid, getvalue("pidfile", commands), MAXPATHLEN);
        DEBUG(1, "pid set to %s\n", options.pid);
    }
    if (getvalue("statfile", commands)) {
        strncpy(options.stat, getvalue("statfile", commands), MAXPATHLEN);
        DEBUG(1, "stat set to %s\n", options.stat);
    }
    if (getvalue("uid", commands)) {
        strncpy(options.uid, getvalue("uid", commands), 30);
        DEBUG(1, "uid set to %s\n", options.uid);
    }
    if (getvalue("gid", commands)) {
        strncpy(options.gid, getvalue("gid", commands), 30);
        DEBUG(1, "gid set to %s\n", options.gid);
    }
    if (getvalue("localtime", commands)) {
        options.localtime = atoi(getvalue("localtime", commands));
        DEBUG(1, "localtime set to %d\n", options.localtime);
    }
    /* options.host */
    if (getvalue("host", commands)) {
        if (atoi(getvalue("host", commands)) == 0)
            options.host = 0;
        else {
            if (inet_pton(AF_INET, getvalue("host", commands), &addr) <= 0) {
                DEBUG(2, "may be it is hostname?\n");
                if ((he = gethostbyname(getvalue("host", commands))) == NULL) {
                    fprintf(stderr, "ERROR: Invalid address specified\n");
                    exit(-1);
                }
                addr = *(struct in_addr *)(he->h_addr);
            }
            options.host = addr.s_addr;
        }
    }

    DEBUG(4, "Loading modules\n");
    /* options.modules */
    v = getallvalues("modules", commands);

    DEBUG(4, "Initialized temporary modules\n");
    modules = 0;
    while (v) {
        /* finding the last element */
        if (modules) {
            m = modules;
            while (m->next) {
                m = m->next;
            }
            m->next = (struct nnfc_module *)calloc(1, 
                    sizeof(struct nnfc_module));
            strncpy(m->next->name, v->value, MAXPATHLEN);
        } else {
            modules = (struct nnfc_module *)calloc(1,
                    sizeof(struct nnfc_module));
            strncpy(modules->name, v->value, MAXPATHLEN);
        }
        v = v->next;
    }
    DEBUG(4, "modules filled\n");
    if (options.verbose > 1) {
        DEBUG(1, "Modules loaded:\n");
        m = modules;
        while (m) {
            fprintf(stderr, "%s\n", m->name);
            m = m->next;
        }
    }

    /* options.ips */
    v = getallvalues("ip", commands);
    addr_count = 0;
    while (v) {
        if (inet_pton(AF_INET, v->value, &addr) <= 0) {
            DEBUG(2, "may be it is hostname?\n");
            if ((he = gethostbyname(v->value)) == NULL) {
                fprintf(stderr, "ERROR: Invalid address in parameter ip\n");
                exit(-1);
            }
            addr = *(struct in_addr *)(he->h_addr);
        }
        if (addresses == NULL) {
            addresses = (uint32_t *)calloc(1, sizeof(uint32_t));
        } else {
            addresses = (uint32_t *)realloc(addresses, 
                sizeof(uint32_t)*(addr_count+1));
        }
        addresses[addr_count] = addr.s_addr;
        /* adding addr.s_addr to host list */
        addr_count++;
        /* adding exporter to statistics */
        nfstat.expstat = (struct estat *)realloc(nfstat.expstat,
            (nfstat.src_cnt+1) * sizeof(struct estat));
        nfstat.expstat[nfstat.src_cnt].addr = addr.s_addr;
        nfstat.expstat[nfstat.src_cnt].ver = 0;
        nfstat.expstat[nfstat.src_cnt].seq = 0;
        nfstat.expstat[nfstat.src_cnt].uptime = 0;
        nfstat.expstat[nfstat.src_cnt].p_count = 0;
        nfstat.expstat[nfstat.src_cnt].f_count = 0;
        nfstat.expstat[nfstat.src_cnt].m_count = 0;
        nfstat.src_cnt++;
        v = v->next;
    }

    if (options.verbose > 1)
        for (i = 0; i < addr_count; i++) {
            char    str[128];
            addr.s_addr = addresses[i];
            inet_ntop(AF_INET, &addr, str, sizeof(str));
            fprintf(stderr, "%s\n", str);
        }

    freecmds(commands);
}

