/**************************************************************************
 * Copyright (c) 2004-2006, Andrew Klyachkin <andrewk@region73.ru>        *
 * All rights reserved.                                                   *
 *                                                                        *
 * Redistribution and use in source and binary forms, with or without     *
 * modification, are permitted provided that the following conditions     *
 * are met:                                                               *
 *                                                                        *
 * 1. Redistributions of source code must retain the above copyright      *
 *    notice, this list of conditions and the following disclaimer.       *
 * 2. Redistributions in binary form must reproduce the above copyright   *
 *    notice, this list of conditions and the following disclaimer in the *
 *    documentation and/or other materials provided with the distribution.*
 * 3. Neither the name of the Andrew Klyachkin nor the names of his       *
 *    contributors may be used to endorse or promote products derived     *
 *    from this software without specific prior written permission.       *
 *                                                                        *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS    *
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT      *
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR  *
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT   *
 * AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, *
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       *
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  *
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  *
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    *
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  *
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   *
 **************************************************************************/

#include <sys/stat.h>
#include <sys/types.h>
#include <stdlib.h>
#include <stdio.h>
#include <netinet/in.h>
#include <signal.h>
#include <syslog.h>
#include <pwd.h>
#include <grp.h>
#include <unistd.h>
#include <pthread.h>
#include <string.h>
#include <time.h>

#include "nnfc.h"

int                 sock;
pthread_mutex_t     mutex = PTHREAD_MUTEX_INITIALIZER;
pthread_cond_t      cond = PTHREAD_COND_INITIALIZER;
pthread_t           reader_id[PTHREADS_NUMBER];
struct nnfc_module  *modules = 0;
struct pool         *POOL = 0;

int main(int argc, char *argv[])
{
    struct sockaddr_in  name;
    struct passwd       *passwd;
    struct group        *group;
	fd_set			rset;
	int			maxfd;
	pid_t			pid;
	FILE			*fp;
	pthread_attr_t		p_attr;
	char			s[MAXPATHLEN];
	struct sockaddr_in	clientname;
	unsigned char		packet[MAXMSG];
	size_t			size;
	int			nbytes, i, j;
	struct pool		*pool, *tail;
	sigset_t		sigset;
	time_t			tm;
	
	/* collector's startup */
	tm = time(NULL);
	nfstat.time = tm;
	
	/* command-line arguments */
	parse_arg(argc, argv);
	
	/* load config file */
	DEBUG(2, "Loading config file...\n");
	readconfig();
	
	/* syslog */
	DEBUG(2, "Opening syslog...\n");
	openlog("nnfc", LOG_NDELAY | LOG_PID, LOG_DAEMON);
	
	/* socket binding */
	DEBUG(2, "Opening socket...\n");
	sock = n_socket(AF_INET, SOCK_DGRAM, 0);
	
	bzero(&name, sizeof(name));
	name.sin_family = AF_INET;
	name.sin_port = htons(options.port);
	name.sin_addr.s_addr = options.host;

	DEBUG(2, "Binding socket on port %d...\n", options.port);
	n_bind(sock, (struct sockaddr *)&name, sizeof(name));
	
	/* signal handlers */
	DEBUG(2, "Installing signal handlers...\n");
	n_signal();
	
	/* set uid and gid */
	DEBUG(2, "Changing uid to %s and gid to %s...\n", 
		options.uid, options.gid);
	passwd = getpwnam(options.uid);
	group = getgrnam(options.gid);
	setgid(group->gr_gid);
	setuid(passwd->pw_uid);
	
	/* fork must be here */
	if (options.daemon) {
		n_daemon();
		setsid();
	}
	
	/* loading reader */
    printf("Create threads\n");
	pthread_attr_init(&p_attr);
	pthread_attr_setdetachstate(&p_attr, PTHREAD_CREATE_JOINABLE);
    for (j = 0; j < PTHREADS_NUMBER; j++) {
        //printf("reader_id[%d] = %d\n", j, &reader_id[j]);
        pthread_create(&reader_id[j], &p_attr, &(reader), NULL);
    printf("Thread %d created\n", j);
    }

	/* creating pid file */
	pid = getpid();
	snprintf(s, MAXPATHLEN, "%s", options.pid);
	fp = fopen(s, "w");
	if (fp != NULL) {
		fprintf(fp, "%d\n", pid);
		fclose(fp);
	}
	
	/* main cycle */
	sigemptyset(&sigset);
	sigaddset(&sigset, SIGHUP);
	sigaddset(&sigset, SIGINT);
	sigaddset(&sigset, SIGQUIT);
	sigaddset(&sigset, SIGTERM);
	sigaddset(&sigset, SIGSEGV);
	sigaddset(&sigset, SIGUSR1);
	sigaddset(&sigset, SIGCHLD);
	FD_ZERO(&rset);
	for (;;) {
		FD_SET(sock, &rset);
		maxfd = sock + 1;
		sigprocmask(SIG_BLOCK, &sigset, NULL);
		n_select(maxfd, &rset, NULL, NULL, NULL);
		sigprocmask(SIG_UNBLOCK, &sigset, NULL);
		if (FD_ISSET(sock, &rset)) {	/* socket is ready to read */
			/* reading socket v3 */
			size = sizeof(clientname);
			if ((nbytes = n_recvfrom(sock, packet, MAXMSG, 0,
					(struct sockaddr *)&clientname, 
					&size))) {
				DEBUG(2, "got packet from %s\n", n_sock_ntop(
					(struct sockaddr *)&clientname, size));
				
				/* verify sender's ip address */
				for (i = 0; i < addr_count; i++)
					if (addresses[i] == 
						clientname.sin_addr.s_addr)
						break;
				if (addresses[i] != 
					clientname.sin_addr.s_addr) {
					DEBUG(2, "unknown ip address\n");
					continue;
				}

				/* writing data in a pool */
				pool = (struct pool *)calloc(1, 
					sizeof(struct pool));
				pool->flags = 0;
				pool->src = clientname.sin_addr.s_addr;
				pool->size = nbytes;
				pool->packet = calloc(1, MAXMSG);
				if (nbytes > MAXMSG)
					memcpy(pool->packet, (void *)&packet,
						MAXMSG);
				else
					memcpy(pool->packet, (void *)&packet, 
						nbytes);
				pool->next = NULL;
				pthread_mutex_lock(&mutex);
				if (!POOL) {
					tail = pool;
					POOL = tail;
				} else {
					tail->next = pool;
					tail = tail->next;
				}
				pthread_mutex_unlock(&mutex);
				pthread_cond_signal(&cond);
            }
        }
    }
    return EXIT_SUCCESS;
}
