/**************************************************************************
 * Copyright (c) 2004-2006, Andrew Klyachkin <andrewk@region73.ru>        *
 * All rights reserved.                                                   *
 *                                                                        *
 * Redistribution and use in source and binary forms, with or without     *
 * modification, are permitted provided that the following conditions     *
 * are met:                                                               *
 *                                                                        *
 * 1. Redistributions of source code must retain the above copyright      *
 *    notice, this list of conditions and the following disclaimer.       *
 * 2. Redistributions in binary form must reproduce the above copyright   *
 *    notice, this list of conditions and the following disclaimer in the *
 *    documentation and/or other materials provided with the distribution.*
 * 3. Neither the name of the Andrew Klyachkin nor the names of his       *
 *    contributors may be used to endorse or promote products derived     *
 *    from this software without specific prior written permission.       *
 *                                                                        *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS    *
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT      *
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR  *
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT   *
 * AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, *
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       *
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  *
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  *
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    *
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  *
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   *
 **************************************************************************/

#ifndef NNFC_H
#define NNFC_H

#include <config.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/param.h>
#include <netinet/in.h>
#include <stdio.h>
#include <sys/time.h>
#include <unistd.h>
#include <pthread.h>

#include <libnnfc.h>

#define	DEBUG(level, fmt...)	{ if (options.verbose > level) \
                    fprintf(stderr, fmt); \
                }

#define DEBUGs(level, fmt...)	{ if (options.verbose > level) \
                    syslog(LOG_INFO, fmt); \
                }

#define PTHREADS_NUMBER 1

/* Internal nnfc structures */

struct options {
    int         port;			    /* port to bind */
    int         verbose;		    /* verbosity level */
    char        uid[30];		    /* username */
    char        gid[30];		    /* groupname */
    uint32_t    host;			    /* ip address to bind */
    char        config[MAXPATHLEN];	/* config file */
    int         localtime;	    	/* what time to log - */
                                    /* local or remote */
    int         daemon;			    /* daemonize process */
    char        pid[MAXPATHLEN];	/* pid file */
    char        stat[MAXPATHLEN];	/* statistics file */
};

/* packet queue */

struct pool {
    int         flags;          /* packet flags */
    uint32_t    src;            /* packet originator */
    int         size;           /* packet size */
    void        *packet;        /* netflow packet */
    struct pool *next;          /* next packet in pool */
};

/* statistics */
struct estat {
    uint32_t    addr;           /* IP address */
    int         ver;            /* last packet version */
    int         seq;            /* last packet seq number */
    int         uptime;         /* exporter's uptime */
    int         p_count;        /* received packet count */
    int         f_count;        /* received flow count */
    int         m_count;        /* missed flow count */
};

struct nfstat {
    time_t      time;           /* startup time */
    int         p_count;        /* received packet count */
    int         f_count;        /* received flow count */
    int         m_count;        /* missed flow count */
    int         src_cnt;        /* number of NF exporters */
    struct estat    *expstat;   /* per exporter statistics */
};

/* functions in signals.c */
void n_signal();

/* functions in cli.c */
void usage();
void parse_arg(int, char **);

/* function in config.c */
void readconfig();

/* functions in reader.c */
void *reader(void *);

/* functions in module */
extern void * _create_plugin_context();

extern int                  sock;       /* main socket */
extern struct options       options;    /* collector options */
extern struct nfstat        nfstat;     /* nnfc statistics */
extern struct nnfc_module   *modules;   /* modules */
extern struct pool          *POOL;      /* pool of packets */
extern uint32_t             *addresses; /* IP addresses of routers */
extern int                  addr_count; /* number of addresses */
extern pthread_mutex_t      mutex;      /* queue mutex */
extern pthread_cond_t       cond;       /* queue condition */
extern pthread_t            reader_id[PTHREADS_NUMBER];  /* reader() thread id */

#endif	/* NNFC_H */
