/**************************************************************************
 * Copyright (c) 2004-2006, Andrew Klyachkin <andrewk@region73.ru>        *
 * All rights reserved.                                                   *
 *                                                                        *
 * Redistribution and use in source and binary forms, with or without     *
 * modification, are permitted provided that the following conditions     *
 * are met:                                                               *
 *                                                                        *
 * 1. Redistributions of source code must retain the above copyright      *
 *    notice, this list of conditions and the following disclaimer.       *
 * 2. Redistributions in binary form must reproduce the above copyright   *
 *    notice, this list of conditions and the following disclaimer in the *
 *    documentation and/or other materials provided with the distribution.*
 * 3. Neither the name of the Andrew Klyachkin nor the names of his       *
 *    contributors may be used to endorse or promote products derived     *
 *    from this software without specific prior written permission.       *
 *                                                                        *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS    *
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT      *
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR  *
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT   *
 * AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, *
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       *
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  *
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  *
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    *
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  *
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   *
 **************************************************************************/

#include <stdio.h>
#include <dlfcn.h>
#include <pthread.h>
#include <stdlib.h>
#include <signal.h>

#include "nnfc.h"

/* We should have separate structures for module
 * and its context because the module is a part 
 * of the main thread */
struct module_context
{
    void *context;
    struct module_context *next;
};

typedef struct module_context mod_context_el;

void *reader(void *param)
{
    struct pool         *pool, *p;
    struct nnfc_module  *m;
    int                 flag = 0, i = 0;
    sigset_t            sigset;
    short               version;
    short               count;
    int                 flowseq, uptime, diff;
    mod_context_el      *mcontext = NULL, *mctx, *t = NULL;

	/* Our thread will not handle and receive signals */
    sigemptyset(&sigset);
    sigaddset(&sigset, SIGHUP);
    sigaddset(&sigset, SIGINT);
    sigaddset(&sigset, SIGQUIT);
    sigaddset(&sigset, SIGTERM);
    sigaddset(&sigset, SIGSEGV);
    sigaddset(&sigset, SIGUSR1);
    sigaddset(&sigset, SIGUSR2);
    sigaddset(&sigset, SIGCHLD);
    pthread_sigmask(SIG_BLOCK, &sigset, NULL);

    /* prepare modules */
    m = modules;
    mctx = (mod_context_el *)malloc(sizeof(mod_context_el *));
    mcontext = mctx;
    t = mctx;
    while (m) {
        /* loading module */
        if (!m->name) {
            DEBUG(1, "reader: unknown module name\n");
            pthread_exit(NULL);
        }
        if (module_open(m) < 0) {
            DEBUG(1, "reader: cannot open module %s\n", m->name);
            pthread_exit(NULL);
        }

        /* initialize context structure for current thread */
        if (!mctx)
           mctx = (mod_context_el *)malloc(sizeof(mod_context_el *));
        mctx->context = (*m->create_thread_context)();

        (*m->connect)(options.config, mctx->context);
        m = m->next;
        t->next = mctx;
        t = mctx;
        mctx = NULL;
    }

    /* eternal cycle */
    while (!flag) {
        /* waiting for data in pool */
        pthread_mutex_lock(&mutex);
        pthread_cond_wait(&cond, &mutex);
        pool = POOL;
        POOL = 0;
        pthread_mutex_unlock(&mutex);

        /* working with pool */
        if (!pool) continue;

        /* begin transaction */
        m = modules;
        mctx = mcontext;
        while (m) {
            (*m->begin)(mctx->context);
            m = m->next;
            mctx = mctx->next;
        }

        p = pool;
        while (p) {
            if (p->flags) {		/* finish processing !!! */
                DEBUG(4, "\nReceived flag - finisihing\n");
                flag = p->flags;
                break;
            }

            /* calculating statistics */
            nfstat.p_count++;
            version = ntohs(*((short *)p->packet));
            count = ntohs(((struct flowhdr1_t *)p->packet)->count);
            uptime = ntohs((
                (struct flowhdr1_t *)p->packet)->uptime);
            nfstat.f_count += count;
            for (i = 0; i < nfstat.src_cnt; i++) {
                if (nfstat.expstat[i].addr == p->src) {
                    nfstat.expstat[i].p_count++;
                    nfstat.expstat[i].f_count += count;
                    nfstat.expstat[i].ver = version;
                    nfstat.expstat[i].uptime = uptime;
                    if (version != 1) {
                        flowseq = ntohl(((struct flowhdr5_t *)
                            p->packet)->flowseq);
                        if (nfstat.expstat[i].seq) {
                            diff = flowseq - nfstat.expstat[i].seq;
                            if (diff) {
                                nfstat.expstat[i].m_count += diff;
                                nfstat.m_count += diff;
                            }
                        }
                        nfstat.expstat[i].seq = flowseq + count;
                    }
                }
            }

            m = modules;
            mctx = mcontext;
            while (m) {
                (*m->insertpkt)(p->src, p->packet, mctx->context);
                m = m->next;
                mctx = mctx->next;
            }

            /* process next */
            p = p->next;
        }

        m = modules;
        mctx = mcontext;
        while (m) {
            (*m->commit)(mctx->context);
            m = m->next;
            mctx = mctx->next;
        }

        /* freeing pool */
        while (pool) {
            p = pool->next;
            if (pool->packet) 
                free(pool->packet);
            else
                if (!pool->flags) {
                    DEBUG(4, "pool->packet is NULL?\n");
                }
            free(pool);
            pool = p;
        }
    }

    /* finishing modules */
    m = modules;
    mctx = mcontext;
    while (m) {
        printf("Closing module ...\n");
        (*m->close)(mctx->context);
        printf("module closed\n");
        module_close(m);
        m = m->next;
        mctx = mctx->next;
    }

    pthread_exit(NULL);
    return NULL;
}
