/**************************************************************************
 * Copyright (c) 2004-2006, Andrew Klyachkin <andrewk@region73.ru>        *
 * All rights reserved.                                                   *
 *                                                                        *
 * Redistribution and use in source and binary forms, with or without     *
 * modification, are permitted provided that the following conditions     *
 * are met:                                                               *
 *                                                                        *
 * 1. Redistributions of source code must retain the above copyright      *
 *    notice, this list of conditions and the following disclaimer.       *
 * 2. Redistributions in binary form must reproduce the above copyright   *
 *    notice, this list of conditions and the following disclaimer in the *
 *    documentation and/or other materials provided with the distribution.*
 * 3. Neither the name of the Andrew Klyachkin nor the names of his       *
 *    contributors may be used to endorse or promote products derived     *
 *    from this software without specific prior written permission.       *
 *                                                                        *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS    *
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT      *
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR  *
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT   *
 * AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, *
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       *
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  *
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  *
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    *
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  *
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   *
 **************************************************************************/

#include <sys/types.h>
#include <sys/wait.h>
#include <syslog.h>
#include <stdlib.h>
#include <signal.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <time.h>

#include "nnfc.h"

struct nfstat	nfstat = {
    time:       0,
    p_count:    0,
    f_count:    0,
    m_count:    0,
    src_cnt:    0,
    expstat:    0
};

void signal_interrupt(int sig)
{
    struct pool         *pool, *p;
    struct nnfc_module  *m1, *m2;
    void                *rc;
    int                 j;

    switch (sig) {
    case SIGINT:
        DEBUGs(1, "SIGINT received - exiting...");
        break;
    case SIGQUIT:
        DEBUGs(1, "SIGQUIT received - exiting...");
        break;
    case SIGSEGV:
        DEBUGs(1, "SIGSEGV received - exiting...");
        break;
    case SIGTERM:
        DEBUGs(1, "SIGTERM received - exiting...");
        break;
    default:
        DEBUGs(1, "Signal %d received - exiting...", sig);
	}
    DEBUGs(4, "shutdowning socket");
    n_shutdown(sock, SHUT_RDWR);
    DEBUGs(4, "sending flag to queue");
    pool = (struct pool *)calloc(1, sizeof(struct pool));
    pool->flags = 1;
    pool->packet = pool->next = 0;
    pool->src = pool->size = 0;
    if (POOL) {
        p = POOL;
        while (p->next)
            p = p->next;
    }
    pthread_mutex_lock(&mutex);
    if (!POOL) {
        POOL = pool;
    } else {
        p->next = pool;
    }
    pthread_mutex_unlock(&mutex);
    pthread_cond_signal(&cond);
    DEBUGs(4, "waiting for threads");
    for (j = 0; j < PTHREADS_NUMBER; j++) {
        //printf("reader_id[%d] = %d\n", j, &reader_id[j]);
        printf("join thread %d\n", j);
        //if (pthread_join(reader_id[j], &rc)) {
        if (pthread_join(reader_id[j], NULL)) {
            DEBUGs(4, "pthread_join: %s", strerror(errno));
        }
    }
    DEBUGs(4, "waiting for children");
    wait((int *)0);
    DEBUGs(4, "freeing modules");
    m1 = modules;
    while (m1) {
        m2 = m1->next;
        free(m1);
        m1 = m2;
    }
    DEBUGs(4, "unlinking pid file");
    unlink(options.pid);
    DEBUGs(4, "closing syslog");
    closelog();
    DEBUGs(4, "exiting");
    exit(0);
}

void signal_hup(int param)
{
    struct pool         *pool, *p;
    struct sockaddr_in  name;
    struct nnfc_module  *m1, *m2;
    pthread_attr_t      p_attr;
    void                *rc;
    int                 j;

    /* FULL config re-reading */
    /* 1. Stop all services */
    DEBUGs(0, "SIGHUP received - re-reading config");

    DEBUGs(2, "stopping all services");
    n_shutdown(sock, SHUT_RDWR);
    pool = (struct pool *)calloc(1, sizeof(struct pool));
    pool->flags = 1;
    pool->packet = pool->next = 0;
    pool->src = pool->size = 0;
    if (POOL) {
        p = POOL;
        while (p->next)
            p = p->next;
    }
    pthread_mutex_lock(&mutex);
    if (!POOL) {
        POOL = pool;
    } else {
        p->next = pool;
    }
    pthread_mutex_unlock(&mutex);
    pthread_cond_signal(&cond);
    for (j = 0; j < PTHREADS_NUMBER; j++) {
        pthread_join(reader_id[j], &rc);
    }

    DEBUGs(4, "freeing modules");
    m1 = modules;
    while (m1) {
        m2 = m1->next;
        free(m1);
        m1 = m2;
    }

    DEBUGs(2, "reading config file %s", options.config);
    readconfig();

    DEBUGs(2, "rebinding socket");
    sock = n_socket(AF_INET, SOCK_DGRAM, 0);

    memset(&name, 0, sizeof(name));
    name.sin_family = AF_INET;
    name.sin_port = htons(options.port);
    name.sin_addr.s_addr = options.host;

    n_bind(sock, (struct sockaddr *)&name, sizeof(name));

    DEBUGs(2, "relaunching reader");
    pthread_attr_init(&p_attr);
    pthread_attr_setdetachstate(&p_attr, PTHREAD_CREATE_JOINABLE);
    for (j = 0; j < PTHREADS_NUMBER; j++) {
        pthread_create(&reader_id[j], &p_attr, &reader, NULL);
    }

    DEBUGs(2, "rereading finished");
}

void signal_usr1(int p)
{
    FILE            *f;
    time_t          tm;
    struct tm       *stm;
    int             i;
    char            tmp[128];
    struct in_addr  a;

    /* Writes statistics in file */
    DEBUGs(1, "SIGUSR1 received - writing statistics...");
    f = fopen(options.stat, "w");
    if (!f) {
        DEBUGs(1, "Cannot create %s", options.stat);
    } else {
        tm = time(NULL);
        stm = localtime(&tm);
        fprintf(f, "New Netflow Collector Statistics\n");
        fprintf(f, "Current time: %04d-%02d-%02d %02d:%02d:%02d\n", 
            stm->tm_year+1900, stm->tm_mon+1, stm->tm_mday,
            stm->tm_hour, stm->tm_min, stm->tm_sec);
        stm = localtime(&nfstat.time);
        fprintf(f, "Collector started %04d-%02d-%02d %02d:%02d:%02d\n",
            stm->tm_year+1900, stm->tm_mon+1, stm->tm_mday,
            stm->tm_hour, stm->tm_min, stm->tm_sec);
        fprintf(f, "Uptime: %d sec\n", tm - nfstat.time);
        fprintf(f, "Packets received: %d\n", nfstat.p_count);
        fprintf(f, "Flows received: %d\n", nfstat.f_count);
        fprintf(f, "Missed flows: %d\n", nfstat.m_count);
        fprintf(f, "Packets/sec: %0.2f\n", 
            nfstat.p_count / (float)(tm - nfstat.time) );
        fprintf(f, "Flows/sec: %0.2f\n",
            nfstat.f_count / (float)(tm - nfstat.time) );
        fprintf(f, "\nPer exporter statistics\n\n");
        for (i = 0; i < nfstat.src_cnt; i++) {
            a.s_addr = nfstat.expstat[i].addr;
            inet_ntop(AF_INET, &a, tmp, 128);
            fprintf(f, "Exporter's IP: %s\n", tmp);
            fprintf(f, "Exporter's uptime: %lu\n", 
                nfstat.expstat[i].uptime);
            fprintf(f, "Last packet version: %d\n", 
                nfstat.expstat[i].ver);
            fprintf(f, "Last flow sequence number: %lu\n", 
                nfstat.expstat[i].seq);
            fprintf(f, "Packets received: %d\n",
                nfstat.expstat[i].p_count);
            fprintf(f, "Flows received: %d\n",
                nfstat.expstat[i].f_count);
            fprintf(f, "Missed flows: %d\n",
                nfstat.expstat[i].m_count);
            fprintf(f, "Packets/sec: %0.2f\n",
                nfstat.expstat[i].p_count / 
                (float)(tm - nfstat.time));
            fprintf(f, "Flows/sec: %0.2f\n",
                nfstat.expstat[i].f_count / 
                (float)(tm - nfstat.time));
        }
        fclose(f);
    }
}

void signal_usr2(int p)
{
    /* What will we do here? */
    DEBUGs(1, "SIGUSR2 received - ignoring...");
}

void signal_chld(int p)
{
    pid_t   pid;
    int     stat;

    pid = wait(&stat);
    return;
}

void n_signal()
{
    signal(SIGHUP, (sig_t)signal_hup);
    signal(SIGINT, (sig_t)signal_interrupt);
    signal(SIGQUIT, (sig_t)signal_interrupt);
    signal(SIGTERM, (sig_t)signal_interrupt);
    signal(SIGSEGV, (sig_t)signal_interrupt);
    signal(SIGUSR1, (sig_t)signal_usr1);
    signal(SIGUSR2, (sig_t)signal_usr2);
    signal(SIGCHLD, (sig_t)signal_chld);
#ifdef SIGNAL_SIGILL
    signal(SIGILL, SIG_IGN);
#endif
#ifdef SIGNAL_SIGABRT
    signal(SIGABRT, SIG_IGN);
#endif
#ifdef SIGNAL_SIGFPE
    signal(SIGFPE, SIG_IGN);
#endif
#ifdef SIGNAL_SIGPIPE
    signal(SIGPIPE, SIG_IGN);
#endif
#ifdef SIGNAL_SIGALRM
    signal(SIGALRM, SIG_IGN);
#endif
#ifdef SIGNAL_SIGBUS
    signal(SIGBUS, SIG_IGN);
#endif
#ifdef SIGNAL_SIGPOLL
    signal(SIGPOLL, SIG_IGN);
#endif
#ifdef SIGNAL_SIGPROF
    signal(SIGPROF, SIG_IGN);
#endif
#ifdef SIGNAL_SIGSYS
    signal(SIGSYS, SIG_IGN);
#endif
#ifdef SIGNAL_SIGTRAP
    signal(SIGTRAP, SIG_IGN);
#endif
#ifdef SIGNAL_SIGVTALRM
    signal(SIGVTALRM, SIG_IGN);
#endif
#ifdef SIGNAL_SIGXCPU
    signal(SIGXCPU, SIG_IGN);
#endif
#ifdef SIGNAL_SIGXFSZ
    signal(SIGXFSZ, SIG_IGN);
#endif
#ifdef SIGNAL_SIGEMT
    signal(SIGEMT, SIG_IGN);
#endif
#ifdef SIGNAL_SIGTSTP
    signal(SIGTSTP, SIG_IGN);
#endif
#ifdef SIGNAL_SIGTTIN
    signal(SIGTTIN, SIG_IGN);
#endif
#ifdef SIGNAL_SIGTTOU
    signal(SIGTTOU, SIG_IGN);
#endif
}
